package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchRecipeException;
import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.models.*;
import be.ac.ulb.infof307.g02.utils.CryptoSingleton;
import be.ac.ulb.infof307.g02.views.HelpPage;
import javafx.scene.Scene;

/**
 * This class if responsible for the navigation between the views. It keeps
 * track of a current scene, current Controller of the scene and session data
 * that is shared among all controllers.
 */
public class ControllerMaster {

    /**
     * The main scene of the application
     */
    private Scene scene;
    private Controller currentController;
    private HelpController helpController;
    /**
     * Contains the information about the logged in user
     */
    private Session session;

    public ControllerMaster(Scene scene) {
        this.scene = scene;
        this.scene.getStylesheets().add(Constants.CSS_FILE_PATH);
        this.showLogin();

    }

    public Scene getScene() {
        return this.scene;
    }

    /**
     * Closes the help, reset the session and the cryptography
     * Then, shows the login screen
     */
    public void onLogout() {
        this.closeHelp();
        this.setSession(null);
        CryptoSingleton.resetCrypto();
        this.onShowLogin();
    }

    /**
     * Sets the session to newSession and then shows the map
     *
     * @param newSession
     */
    public void onLogin(Session newSession) {
        this.session = newSession;
        this.onShowMap();
    }

    private void setSession(Session session) {
        this.session = session;
    }

    /**
     * Show Login
     */
    public void onRegister() {
        this.onShowLogin();
    }

    /**
     * Shows the login view
     */
    public void onShowLogin() {
        this.showLogin();

    }

    private void showLogin() {
        LoginController controller = new LoginController(this);
        this.setCurrentController(controller);
    }

    private void setCurrentController(Controller controller) {
        this.currentController = controller;
    }

    /**
     * Shows the register view
     */
    public void onShowRegister() {
        RegisterController controller = new RegisterController(this);
        this.setCurrentController(controller);
    }

    /**
     * Shows the map view.
     */
    public void onShowMap() {
        MapController controller = new MapController(this);
        this.setCurrentController(controller);
    }

    /**
     * Shows the edit profile view
     */
    public void onShowEditProfile() {
        EditAccountController controller = new EditAccountController(this);
        this.setCurrentController(controller);
    }

    /**
     * Shows the add product view
     */
    public void onShowAddProduct(ProductCreationObserver observer) {
        new CreateProductController(observer);
    }

    /**
     * Shows the result view
     *
     * @param groceryList
     * @param position
     */
    public void showResultsScreen(GroceryList groceryList, Position position) {
        ShowResultController controller = new ShowResultController(this, groceryList);
        this.setCurrentController(controller);
    }

    /**
     * Shows all the grocery lists of the connected user
     */
    public void onShowAllGroceryLists() {
        GroceryListSelectionController controller = new GroceryListSelectionController(this);
        this.setCurrentController(controller);
    }

    /**
     * Shows the help view in a new stage
     *
     * @param helpPage
     */
    public void onShowHelp(HelpPage helpPage) {
        this.helpController = new HelpController(this, helpPage);
        this.currentController.onHelpViewStatusChanged(true);
    }

    /**
     * Closes the Help view
     */
    private void closeHelp() {
        if (this.helpController != null) {
            this.helpController.close();
        }
        this.helpController = null;
    }


    /**
     * Called when the user close the help screen
     */
    public void onUserClosedHelp() {
        this.currentController.onHelpViewStatusChanged(false);
        this.helpController = null;
    }

    /**
     * Checks if the help is screen is opened
     *
     * @return True if the screen is opened, else false
     */
    public boolean isHelpOpen() {
        return this.helpController != null;
    }

    /**
     * Shows all the recipes of the connected user
     */
    public void onShowAllRecipes() {
        RecipeSelectionController controller = new RecipeSelectionController(this);
        this.setCurrentController(controller);
    }

    /**
     * Shows the recipe view
     *
     * @throws RecipeCreationException
     * @throws NoSuchRecipeException
     */
    public void onShowRecipe(String recipeName) throws NoSuchRecipeException, RecipeCreationException {
        RecipeController controller = new RecipeController(this, recipeName);
        this.setCurrentController(controller);

    }

    /**
     * Called when a grocery List has been generated from a recipe
     *
     * @param recipeName The name of the recipe
     */
    public void onGroceryListGenerated(String recipeName) {
        this.onShowGroceryList(recipeName);
    }

    /**
     * Shows the groceryList view
     *
     * @param listName
     */
    public void onShowGroceryList(String listName) {
        GroceryListController controller = new GroceryListController(this, listName);
        this.setCurrentController(controller);

    }

    /**
     * Shows the view to select a grocery list for a Best Tour.
     */
    public void onShowGroceryListSelector() {
        new SelectBestTourGroceryListController(this);

    }

    /**
     * Shows the best route.
     *
     * @param list The grocery list used for the Best Tour
     */
    public void onBestRouteAlgoReady(GroceryList list) {
        MapController mapController = (MapController) this.currentController;
        mapController.findBestRoute(list);
    }

    /**
     * Removes start and destination pin of best tour.
     * Used if the user closes the grocery list selection Popup
     */
    public void onCloseSelectGroceryListView() {
        MapController mapController = (MapController) this.currentController;
        mapController.removeBestTourPositions();
    }

    /**
     * Shows the Importation view
     *
     * @param targetShop The shop in which the articles will be imported
     */
    public void onArticlesImport(Shop targetShop) {
        ImportArticlesController controller = new ImportArticlesController(this, targetShop);
        this.setCurrentController(controller);
    }

    /**
     * Updates the position of the user in the Session
     *
     * @param position The new Position
     */
    public void updateUserPosition(Position position) {
        this.session.setPosition(position);
    }

    public Position getUserPosition() {
        return this.session.getPosition();
    }

    public User getUser() {
        return this.session.getUser();
    }

    public void setUser(User user) {
        this.session.setUser(user);
    }

    public String getUserName() {
        return this.session.getUserName();
    }
}
