package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.UserDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchUserFoundException;
import be.ac.ulb.infof307.g02.exceptions.database.UserNameTakenException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.User;
import be.ac.ulb.infof307.g02.utils.CryptoSingleton;
import be.ac.ulb.infof307.g02.views.EditAccountView;

/**
 * Controller of the edit profile Window.
 */
public class EditAccountController extends ControllerWithMenu {
    private UserDB database;
    private EditAccountView view;

    public EditAccountController(ControllerMaster master) {
        this.master = master;
        try {
            this.view = new EditAccountView(this, master.getScene());
            this.database = new UserDB();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        }
    }

    /**
     * Checks if the new informations are correct, then update the account informations
     *
     * @param newUserName
     * @param currentPassword
     * @param newPassword
     * @param newPasswordConfirm
     */
    public void updateAccountInformation(String newUserName, String currentPassword, String newPassword,
                                         String newPasswordConfirm) {
        try {
            if (!this.isPasswordCorrect(currentPassword)) {
                this.view.showInvalidCurrentPasswordError();
                return;
            }

            if (!newPassword.isEmpty() && !newPassword.equals(newPasswordConfirm)) {
                this.view.showPasswordsDontMatchError();
                return;
            }

            this.view.hideAllLabels();

            User user = this.master.getUser();
            String finalUserName = user.getUserName();
            String finalPassword = user.getHashedPassword();
            if (!newUserName.isEmpty()) {
                this.updateUsername(newUserName);
                finalUserName = newUserName;
            }
            if (!newPassword.isEmpty()) {
                finalPassword = this.updatePassword(newPassword);
            }

            if (!newUserName.isEmpty() || !newPassword.isEmpty()) {
                // There's no need to re-hash the password. That's the reason why we use getBytes
                User updatedUser = new User(finalUserName, finalPassword.getBytes());
                CryptoSingleton.updateUserKeys(user, updatedUser);
            }

            this.view.showProfileEditedPopup();
            this.master.onShowMap();
        } catch (UserNameTakenException e) {
            this.view.showErrorUsername();
        } catch (UserCreationException e) {
            this.view.showErrorPassword();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        }
    }

    /**
     * Updates the user name in the database and in the session.
     *
     * @param newUserName String
     * @throws UserNameTakenException
     */
    private void updateUsername(String newUserName) throws UserNameTakenException {
        boolean newUserNameIsDifferent = !newUserName.equals(this.master.getUserName());
        if (newUserNameIsDifferent) {
            User oldUser = this.master.getUser();
            User updatedUser = new User(newUserName, oldUser.getHashedPasswordBytes());

            this.database.updateUserName(oldUser.getUserName(), newUserName);
            this.master.setUser(updatedUser);
        }
    }

    /**
     * Checks if the password match with the one of the logged in user
     *
     * @throws DatabaseException
     */
    private boolean isPasswordCorrect(String password) throws DatabaseException {
        try {
            User user = this.database.getUserByName(this.master.getUserName());
            return user.passwordMatches(password);
        } catch (NoSuchUserFoundException e) {
            // There will always be a User with this userName, so we don't have to worry
            return true;
        }
    }

    /**
     * Updates the password of the logged in User.
     *
     * @param newPassword String
     * @throws UserCreationException
     * @throws DatabaseException
     */
    private String updatePassword(String newPassword) throws UserCreationException, DatabaseException {
        User user = this.master.getUser();
        User updatedUser = new User(user.getUserName(), newPassword);
        this.database.updatePassword(updatedUser.getUserName(), updatedUser.getHashedPassword());
        this.master.setUser(updatedUser);
        return updatedUser.getHashedPassword();
    }
}