package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.GroceryListDB;
import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.GroceryListInsertionException;
import be.ac.ulb.infof307.g02.exceptions.database.GroceryListItemInsertException;
import be.ac.ulb.infof307.g02.models.GroceryList;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.views.GroceryListView;
import be.ac.ulb.infof307.g02.views.HelpPage;

import java.util.List;

/**
 * Controller of the grocery list view.
 */
public class GroceryListController extends ControllerWithMenuThatSaves implements ProductCreationObserver {

    private GroceryListView view;
    private GroceryListDB groceryListDB;
    private GroceryList groceryList;

    public GroceryListController(ControllerMaster master, String listName) {
        super(HelpPage.GROCERY_LIST);
        this.master = master;

        try {
            this.groceryListDB = new GroceryListDB();
            this.groceryList = this.groceryListDB.getGroceryList(this.master.getUser(), listName);

            ProductDB productDB = new ProductDB();
            List<Product> allProductsList = productDB.getAllProducts();
            this.view = new GroceryListView(this, this.master.getScene(), this.groceryList, allProductsList);

        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();

        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }

    /**
     * Creates a new Product
     */
    public void createProduct() {
        this.master.onShowAddProduct(this);
    }

    /**
     * Shows shops containing a product
     */
    public void showShopsWithProduct(ProductAmount productAmount) {
        this.saveGroceryList(this.view.getProductAmountsList());

        GroceryList groceryListOneProduct = new GroceryList(this.groceryList.getGroceryListName());
        groceryListOneProduct.addProductOrIncreaseAmount(productAmount);
        this.master.showResultsScreen(groceryListOneProduct, this.master.getUserPosition());
    }

    /**
     * Creates a new GroceryList, fill it with all the items retrieved from the
     * view and saves it in the database.
     */
    public void saveGroceryList(List<ProductAmount> productAmountsList) {
        GroceryList listToSave = new GroceryList(this.groceryList.getGroceryListName(), productAmountsList);
        this.saveListInDB(listToSave);
    }

    /**
     * Saves the GroceryList in the database. To save it, it deletes all the
     * list entries associated to it from the database and creates new entries
     * for each element of the Grocery List.
     *
     * @param listToSave GroceryList
     */
    private void saveListInDB(GroceryList listToSave) {
        try {
            this.groceryListDB.deleteGroceryList(this.master.getUser(), this.groceryList.getGroceryListName());
            this.groceryListDB.addGroceryList(this.master.getUser(), listToSave);
            this.groceryList = listToSave;
            this.view.showSuccessPopup();
        } catch (GroceryListInsertionException | GroceryListItemInsertException e) {
            this.view.showErrorSavingList();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }

    /**
     * Finds all shops whose contain every item from the list
     */
    public void showShopsWithAllProducts() {
        this.saveGroceryList(this.view.getProductAmountsList());
        this.master.showResultsScreen(this.groceryList, this.master.getUserPosition());
    }

    public String getGroceryListName() {
        return this.groceryList.getGroceryListName();
    }


    /**
     * Asks the user whether he wants to save the grocery list changes and saves it if he does.
     */
    @Override
    protected void askSaveConfirmation() {
        if (this.view.isListModified() && this.view.showSaveMessageBox()) {
            this.saveGroceryList(this.view.getProductAmountsList());
        }
    }

    /**
     * Asks the view to add a newly created product to the choices
     *
     * @param product The newly created product
     */
    @Override
    public void onProductCreated(Product product) {
        this.view.addNewProduct(product);
    }

    /**
     * Asks the view to enable the addProduct button
     */
    @Override
    public void onCreateProductClosed() {
        this.view.enableCreateProductButton();
    }

}
