package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.database.ShopDB;
import be.ac.ulb.infof307.g02.exceptions.database.ArticleAlreadyInShopException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.import_export.CSVNotConsistentException;
import be.ac.ulb.infof307.g02.exceptions.import_export.NoFileSelectedException;
import be.ac.ulb.infof307.g02.models.Article;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.Shop;
import be.ac.ulb.infof307.g02.utils.ArticleImporterFromFile;
import be.ac.ulb.infof307.g02.views.ImportArticlesView;

import java.util.List;

/**
 * Controller of the ArticleProductView. This class is responsible for the manual and from a csv importation of articles
 * into a shop
 */
public class ImportArticlesController extends ControllerWithMenu implements ProductCreationObserver, ArticleImportationObserver {

    private ProductDB productDB;
    private ImportArticlesView view;
    /**
     * The shop in which the articles need to be imported
     */
    private Shop targetShop;

    public ImportArticlesController(ControllerMaster master, Shop targetShop) {
        this.master = master;
        this.targetShop = targetShop;

        try {
            this.productDB = new ProductDB();
            List<Product> importableProducts = this.getImportableProducts();
            this.view = new ImportArticlesView(this, this.master.getScene(), importableProducts);

        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }

    /**
     * Tells the view that a new product has been created
     *
     * @param product
     */
    @Override
    public void onProductCreated(Product product) {
        this.view.addNewProduct(product);
    }

    /**
     * Tells the view that new products can be created again
     */
    @Override
    public void onCreateProductClosed() {
        this.view.enableCreateProduct();

    }

    /**
     * Tells the view that a new article has been created
     *
     * @param article
     */
    @Override
    public void articleImported(Article article) {
        this.view.removeProduct(article.getProduct());
    }

    /**
     * @return All the products that can be imported in the Shop (i.e the ones that are not already in the stock)
     */
    private List<Product> getImportableProducts() throws DatabaseException {
        List<Product> importableProducts = this.productDB.getAllProducts();

        for (Product product : this.targetShop.getProducts()) {
            importableProducts.remove(product);
        }

        return importableProducts;
    }


    /**
     * Import articles from a csv using the ArticleImporterFromFile
     */
    public void importFromCSV() {
        try {
            ArticleImporterFromFile articleImporterFromFile = new ArticleImporterFromFile(this);
            articleImporterFromFile.importFrom(this.targetShop);
            this.view.showImportSuccessful();

        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();

        } catch (CSVNotConsistentException e) {
            this.view.showImportCSVConsistencyError(e.getMessage());

        } catch (NoFileSelectedException ignored) {
            //Will never appear
        } catch (ArticleAlreadyInShopException e) {
            this.view.showImportFailedForSomeProducts();
        }
    }

    /**
     * Imports the products added manually be the user
     */
    public void importArticles(List<Article> articlesToImport) {
        try {
            ShopDB shopDB = new ShopDB();
            for (Article article : articlesToImport) {
                shopDB.insertArticleInShop(this.targetShop.getId(), article);
            }
            this.view.showImportSuccessful();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (ArticleAlreadyInShopException e) {
            this.view.showImportFailedForSomeProducts();
        }


    }

    /**
     * Tells the ControllerMaster that the user wants to create a product
     */
    public void onCreateNewProduct() {
        this.master.onShowAddProduct(this);
    }

}
