package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.UserDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchUserFoundException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.Session;
import be.ac.ulb.infof307.g02.models.User;
import be.ac.ulb.infof307.g02.utils.CryptoSingleton;
import be.ac.ulb.infof307.g02.views.LoginView;

/**
 * Controller of the login View.
 */
public class LoginController extends Controller {

    private UserDB database;
    private LoginView view;

    public LoginController(ControllerMaster master) {
        this.master = master;
        try {
            this.view = new LoginView(this, this.master.getScene());
            this.database = new UserDB();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        }
    }

    /**
     * Checks if the credentials for a user are correct
     * If so, connect the user to the app.
     *
     * @param userName
     * @param password
     */
    public void login(String userName, String password) {
        try {
            User newUser = new User(userName, password);
            if (this.areCredentialsCorrect(newUser)) {
                CryptoSingleton.setupCryptoForUser(newUser);
                this.connectUser(newUser);
            } else {
                this.view.showIncorrectCredentialsError();
            }
        } catch (UserCreationException e) {
            this.view.showInvalidCredentialsError();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        }
    }

    /**
     * Connects a user to the application
     *
     * @param user The user to connect
     */
    private void connectUser(User user) {
        Session newSession = new Session(user);
        this.master.onLogin(newSession);
    }

    /**
     * Consults the database to see if the credentials of a user are
     * correct.
     *
     * @param newUser
     * @return a boolean indicating whether the credentials are correct
     * @throws DatabaseException
     */
    private boolean areCredentialsCorrect(User newUser) throws DatabaseException {
        try {
            User dbUser = this.database.getUserByName(newUser.getUserName());
            return dbUser.passwordMatches(newUser);
        } catch (NoSuchUserFoundException e) {
            return false;
        }
    }

    /**
     * Shows the register view
     */
    public void showRegister() {
        this.master.onShowRegister();
    }

}
