package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.database.ShopDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoShopHasThatProductException;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchShopException;
import be.ac.ulb.infof307.g02.models.*;
import be.ac.ulb.infof307.g02.utils.ArticlesExporter;
import be.ac.ulb.infof307.g02.views.HelpPage;
import be.ac.ulb.infof307.g02.views.MapView;
import org.json.JSONException;
import org.json.JSONObject;

import javax.ws.rs.ProcessingException;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.ArrayList;
import java.util.List;

/**
 * Controller of the MapView
 */
public class MapController extends ControllerWithMenu {

    /**
     * URL of the service used to geolocalise a user
     */
    private static final String LOCATION_SERVICE_URL = "https://api.ipdata.co";
    private MapView view;

    /**
     * All the shops that can be displayed on the map
     */
    private List<Shop> shops;

    /**
     * The shops currently displayed on the map
     */
    private List<Shop> displayedShops;

    /**
     * The starting position of a bestTour
     */
    private Position startingPos;

    /**
     * The starting position of a bestTour
     */
    private Position destinationPos;

    /**
     * Class constructor that retrieve the shops from the database and
     * instantiate the view
     *
     * @param master ControllerMaster
     */
    public MapController(ControllerMaster master) {
        super(HelpPage.MAP);
        this.master = master;

        // Geolocalize the user only if he has not set his position yet
        if (this.master.getUserPosition() == null) {
            this.geoLocalise();
        }

        try {
            ShopDB shopDB = new ShopDB();
            this.shops = shopDB.getAllShops();
            this.displayedShops = this.shops;
            this.view = new MapView(this, this.master.getScene());
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }

    /**
     * GeoLocalises the user using his IP Address
     */
    private void geoLocalise() {
        Position position;
        try {
            Client client = ClientBuilder.newClient();
            Response response = client.target(LOCATION_SERVICE_URL)
                    .request(MediaType.APPLICATION_JSON_TYPE)
                    .header("Accept", "application/json")
                    .get();
            //Need to extract the data (lng and lat) from a JSON data structure
            String JSONString = response.readEntity(String.class);
            JSONObject obj = new JSONObject(JSONString);
            String lat = obj.get("latitude").toString();
            Double latitude = Double.parseDouble(lat);
            String lng = obj.get("longitude").toString();
            Double longitude = Double.parseDouble(lng);
            position = new Position(latitude, longitude);

        } catch (ProcessingException | JSONException e) {
            // It would be appear only if no internet connection or error in the json string content
            // Set the position to the grand place by default
            position = new Position(Constants.DEFAULT_LAT, Constants.DEFAULT_LNG);
        }
        this.updateUserPosition(position);
    }

    /**
     * Updates the user position.
     *
     * @param position user position (latitude, longitude).
     */
    public final void updateUserPosition(Position position) {
        this.master.updateUserPosition(position);
    }

    /**
     * Called when the user clicks on the export button of a shop.
     * First, it gets the shop with correct id from the shop list.
     * Then, it asks the ArticlesExporter to process the export.
     *
     * @param shopId
     */
    public void onExportArticles(int shopId) {
        try {
            Shop shopToExport = this.getShopByID(shopId);
            ArticlesExporter.exportFrom(shopToExport);
        } catch (NoSuchShopException e) {
            this.view.showUnexpectedError();
        }

    }

    /**
     * Called when the user clicks on the import button of a shop.
     * First, it gets the shop with correct id from the shop list.
     * Then, it asks the ArticleImporterFromFile to process the export.
     *
     * @param shopId
     */
    public void importShopArticles(int shopId) {

        try {
            Shop targetShop = this.getShopByID(shopId);
            this.master.onArticlesImport(targetShop);
        } catch (NoSuchShopException e) {
            this.view.showUnexpectedError();
        }

    }

    /**
     * Returns the shop corresponding to a shopID,
     * if no such shop is found it throws an NoSuchException
     *
     * @param shopId
     * @return Shop
     * @throws NoSuchShopException
     */
    private Shop getShopByID(int shopId) throws NoSuchShopException {
        Shop res = null;
        for (Shop shop : this.shops) {
            if (shop.getId() == shopId) {
                res = shop;
            }
        }
        if (res == null) {
            throw new NoSuchShopException();
        }
        return res;
    }

    /**
     * Returns the user position.
     *
     * @return Position object
     */
    public Position getUserPosition() {
        return this.master.getUserPosition();
    }

    /**
     * Called when the starting and the destination position have been
     * clicked on the map. It stores the starting and the destination position and
     * call the showGroceryListSelectorView so the user can select the grocery list
     * that he wants. Later the best tour for this grocery list will be found
     *
     * @param start
     * @param end
     */
    public void onPositionsForTourReady(Position start, Position end) {
        this.startingPos = start;
        this.destinationPos = end;
        this.master.onShowGroceryListSelector();

    }

    /**
     * Finds the best route from the start position to the end position to buy all the
     * products in the grocery list, then shows it.
     *
     * @param groceryList
     */
    public void findBestRoute(GroceryList groceryList) {
        this.view.hideFilterMenu();
        BestTourFinder finder;
        try {
            finder = new BestTourFinder(this.startingPos, this.destinationPos, this.displayedShops, groceryList);
            BestTour bestTour = finder.getBestTour();
            this.view.showBestTour(bestTour);
        } catch (NoShopHasThatProductException e) {
            this.view.showNoBestTourError();
            this.view.removeBestTourPositions();
        }
    }

    public void removeBestTourPositions() {
        this.view.removeBestTourPositions();
    }


    /**
     * Returns a list of all the product available in the database.
     * Used for the auto-completion in the filter menu
     *
     * @return List<Product>
     */
    public List<Product> getProducts() {
        try {
            ProductDB productDB = new ProductDB();
            return productDB.getAllProducts();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
        return null;
    }

    /**
     * Filter the shops of displayedShops with de filterData given by the view.
     *
     * @param data A name, a distance, opening days, an opening time and an product list
     */
    public void filterShops(FilterData data) {
        this.view.clearFiltersOnMap();
        ShopFilterer shopFilterer = new ShopFilterer(new ArrayList<>(this.shops), this.master.getUserPosition());
        this.displayedShops = shopFilterer.filterShops(data);
        this.view.filterShops(this.displayedShops);


    }

    /**
     * Reset the displayedShops list to all the shops/
     */
    public void clearFilters() {
        this.displayedShops = new ArrayList<>(this.shops);
    }

    public List<Shop> getShops() {
        return new ArrayList<>(this.shops);
    }
}
