package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.GroceryListDB;
import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.database.RecipeDB;
import be.ac.ulb.infof307.g02.exceptions.database.*;
import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.models.GroceryList;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.Recipe;
import be.ac.ulb.infof307.g02.views.HelpPage;
import be.ac.ulb.infof307.g02.views.RecipeView;

import java.util.List;

/**
 * Controller of Recipe view.
 */
public class RecipeController extends ControllerWithMenuThatSaves {

    private RecipeView view;
    private RecipeDB recipeDB;
    private Recipe recipe;
    private boolean isRecipeInfoModified = false;

    /**
     * Constructor: Fetches the recipe from the DB and creates the view.
     *
     * @param master     ControllerMaster
     * @param recipeName
     * @throws NoSuchRecipeException
     * @throws RecipeCreationException
     */
    public RecipeController(ControllerMaster master, String recipeName) throws NoSuchRecipeException, RecipeCreationException {
        super(HelpPage.RECIPE);
        this.master = master;

        try {
            this.recipeDB = new RecipeDB();
            this.recipe = this.recipeDB.getRecipe(this.master.getUser(), recipeName);

            ProductDB productDB = new ProductDB();
            List<Product> allProductsList = productDB.getAllProducts();
            this.view = new RecipeView(this, this.master.getScene(), this.recipe, allProductsList);
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (DatabaseException e) {
            this.view.showErrorGettingRecipe();
        }
    }

    public String getRecipeName() {
        return this.recipe.getName();
    }

    /**
     * Saves the recipe, generates the grocery list from the recipe, saves it in the DB
     * and then shows it.
     */
    public void generateGroceryList() {
        try {
            this.saveRecipe();
            GroceryList groceryList = this.getGroceryListFromView();
            this.saveOrOverWriteGroceryListInDb(groceryList);
            this.master.onGroceryListGenerated(this.recipe.getName());
        } catch (RecipeCreationException e) {
            this.view.showErrorSavingRecipe();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        }
    }

    /**
     * Creates a new Recipe, fills it with all the items retrieved from the
     * view and saves it in the database.
     *
     * @throws RecipeCreationException
     */
    public void saveRecipe() throws RecipeCreationException {
        int numPeople = this.view.getNumPeopleValue();
        String instructions = this.view.getInstructionsText();
        List<ProductAmount> productAmountList = this.view.getProductAmountsList();
        Recipe recipeToSave = new Recipe(this.recipe.getName(), numPeople, productAmountList, instructions);
        this.saveRecipeInDB(recipeToSave);
    }

    /**
     * Asks the user whether he wants to save the recipe changes and saves it if he does.
     */
    @Override
    protected void askSaveConfirmation() {
        if ((this.view.isListModified() || this.isRecipeInfoModified) && this.view.showSaveMessageBox()) {
            try {
                this.saveRecipe();
            } catch (RecipeCreationException e) {
                this.view.showErrorSavingRecipe();
            }
        }
    }

    /**
     * Whenever a change is made, the boolean variable isRecipeInfoModified is set to true.
     */
    public void acknowledgeChange() {
        this.isRecipeInfoModified = true;
    }

    /**
     * Creates a GroceryList bases on the content of the recipe view
     */
    private GroceryList getGroceryListFromView() {
        return new GroceryList(this.recipe.getName(), this.view.getProductAmountsList());
    }

    /**
     * Saves the grocery list to the database. If it already exists, we overwrite it.
     *
     * @param groceryList
     * @throws DatabaseConnectionException
     * @throws DatabaseException
     */
    private void saveOrOverWriteGroceryListInDb(GroceryList groceryList) throws DatabaseConnectionException, DatabaseException {
        GroceryListDB groceryListDB = new GroceryListDB();
        try {
            groceryListDB.addGroceryList(this.master.getUser(), groceryList);
        } catch (GroceryListInsertionException | GroceryListItemInsertException e) {
            // The recipe already exists in the database. The user must confirm the overwrite
            if (this.view.askUserOverWriteConfirmation()) {
                try {
                    groceryListDB.deleteGroceryList(this.master.getUser(), this.recipe.getName());
                    groceryListDB.addGroceryList(this.master.getUser(), groceryList);
                } catch (GroceryListInsertionException | GroceryListItemInsertException e2) {
                    // These exceptions were already handled on the lines above
                }
            }
        }
    }

    /**
     * Saves the GroceryList in the database. In order to save it, it deletes all
     * the entries associated with the list in the database and creates a new entry for
     * each element of the Grocery List.
     *
     * @param recipeToSave
     */
    private void saveRecipeInDB(Recipe recipeToSave) {
        try {
            this.recipeDB.deleteRecipe(this.master.getUser(), this.recipe.getName());
            this.recipeDB.addRecipe(this.master.getUser(), recipeToSave);
            this.recipe = recipeToSave;
            this.view.showSuccessPopup();
            this.isRecipeInfoModified = false;
            this.view.setIsListModified(false);
        } catch (RecipeInsertionException | RecipeItemInsertException e) {
            this.view.showErrorSavingRecipe();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }


    /**
     * Returns the Instructions of the recipe
     */
    public String getInstructions() {
        return this.recipe.getInstructions();
    }


    /**
     * Returns the number of people for which the recipe is made
     *
     * @return
     */
    public int getNumberPeople() {
        return this.recipe.getNumberOfPeople();
    }

}