package be.ac.ulb.infof307.g02.controllers;

import be.ac.ulb.infof307.g02.database.ShopDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.models.AutoChefScoreCalculator;
import be.ac.ulb.infof307.g02.models.FindShopsResult;
import be.ac.ulb.infof307.g02.models.GroceryList;
import be.ac.ulb.infof307.g02.models.Shop;
import be.ac.ulb.infof307.g02.utils.DistanceCalculator;
import be.ac.ulb.infof307.g02.utils.Utils;
import be.ac.ulb.infof307.g02.views.ShowResultView;

import java.util.ArrayList;
import java.util.List;

/**
 * Controller of the ShowResultView. This view displays all the shops that contain all the product of a grocery list
 */
public class ShowResultController extends ControllerWithMenu {

    /**
     * The grocery list that contains all the products that need to be sold by the shops
     */
    private GroceryList groceryList;

    /**
     * All the shops contained in the database
     */
    private List<Shop> allShops;

    private ShowResultView view;

    /**
     * Class Constructor
     *
     * @param master
     * @param groceryList
     */
    public ShowResultController(ControllerMaster master, GroceryList groceryList) {
        this.master = master;
        this.groceryList = groceryList;
        try {
            this.view = new ShowResultView(this, this.master.getScene());
            ShopDB shopDB = new ShopDB();
            this.allShops = shopDB.getAllShops();
            this.computeResults();
        } catch (DatabaseConnectionException e) {
            this.view.showDatabaseConnectionError();
        } catch (DatabaseException e) {
            this.view.showDataBaseInternalError();
        }
    }

    /**
     * Finds all the shops that contain all the products of a grocery list
     *
     * @return A list of shops
     */
    private Shop[] findShopsWithAllProducts() {
        ArrayList<Shop> result = new ArrayList<>();
        for (Shop shop : this.allShops) {
            if (shop.sellAllProducts(this.groceryList.getProductsList())) {
                result.add(shop);
            }
        }
        Shop[] res = new Shop[result.size()];
        res = result.toArray(res);
        return res;
    }

    /**
     * Finds the shops with all the products of a grocery list. Computes the price
     * of the grocery list for each shop. Computes the distances between the user
     * and all the shops. Computes the autochef score of each shop, this score is used to rank the shops using
     * both distance and price. Asks the view to display the results.
     */
    private void computeResults() {
        Shop[] shopsWithAllProducts = this.findShopsWithAllProducts();

        List<Double> pricesList = new ArrayList<>();
        List<Double> distancesList = new ArrayList<>();

        for (Shop shop : shopsWithAllProducts) {
            double price = shop.getPriceOfList(this.groceryList);
            double distance = DistanceCalculator.getRealDistance(shop.getPosition(), this.master.getUserPosition());
            pricesList.add(price);
            distancesList.add(distance);
        }
        List<Double> autochefScores = AutoChefScoreCalculator.computeScore(pricesList, distancesList);
        this.showResultsInView(shopsWithAllProducts, pricesList, distancesList, autochefScores);
    }

    /**
     * Tells the view to display the results.
     *
     * @param shopsWithAllProducts
     * @param pricesList
     * @param distancesList
     * @param autochefScores
     */
    private void showResultsInView(Shop[] shopsWithAllProducts, List<Double> pricesList,
                                   List<Double> distancesList, List<Double> autochefScores) {
        for (int i = 0; i < shopsWithAllProducts.length; i++) {
            Shop shop = shopsWithAllProducts[i];
            double price = pricesList.get(i);
            double distance = distancesList.get(i);
            double autochefScore = autochefScores.get(i);

            FindShopsResult res = new FindShopsResult(shop.getName(), Utils.round(distance, 2), Utils.round(price, 2), Utils.round(autochefScore, 2));
            this.view.addResult(res);

        }
    }

    /**
     * Go back to the groceryLists view from which we came from
     */
    public void goBackToGroceryList() {
        this.master.onShowGroceryList(this.groceryList.getGroceryListName());
    }
}
