package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.utils.CryptoSingleton;

import java.sql.*;

/**
 * Connection to the database.
 */
public class DatabaseConnection {
    private Connection connection;

    /**
     * Establishes a connection to the database at the given path. Throws an
     * exception if the connection is not successful.
     *
     * @param databasePath the path to the database file
     * @throws DatabaseConnectionException
     */
    public DatabaseConnection(String databasePath) throws DatabaseConnectionException {
        // We must tell JDBC and SQLite to respect foreign key constraints
        org.sqlite.SQLiteConfig config = new org.sqlite.SQLiteConfig();
        config.enforceForeignKeys(true);

        String connectionPath = "jdbc:sqlite::resource:" + databasePath;
        this.connection = null;
        try {
            Class.forName("org.sqlite.JDBC");
            this.connection = DriverManager.getConnection(connectionPath, config.toProperties());
        } catch (SQLException e) {
            throw new DatabaseConnectionException("An error happened during the database connection "
                    + e.getMessage(), e);
        } catch (Exception e) {
            // If the exception is not a SQL exception we don't need to throw it.
            // It's ok to ignore it.
        }
    }

    /**
     * Empty all the tables of the connected database
     *
     * @throws SQLException
     */
    public void empty() throws SQLException {
        PreparedStatement emptyUser = this.connection.prepareStatement("DELETE FROM User;");
        emptyUser.execute();
        PreparedStatement emptyShop = this.connection.prepareStatement("DELETE FROM Shop;");
        emptyShop.execute();
        PreparedStatement emptyProduct = this.connection.prepareStatement("DELETE FROM Product;");
        emptyProduct.execute();
    }

    /**
     * Executes the SQL query and returns the ResultSet.
     *
     * @param query
     * @throws SQLException
     */
    public ResultSet executeQuery(String query) throws SQLException {
        Statement statement = this.connection.createStatement();
        return statement.executeQuery(query);
    }

    /**
     * Creates a prepared statement
     *
     * @param statement
     * @throws SQLException
     */
    public PreparedStatement createPreparedStatement(String statement) throws SQLException {
        return this.connection.prepareStatement(statement);
    }

    /**
     * Creates a prepared statement that returns the keys generated after the statement is executed
     * (ex: after inserting a Shop, returns the ID of the newly-inserted shop)
     *
     * @param statement
     * @throws SQLException
     */
    public PreparedStatement createPreparedStatementWithGeneratedKeys(String statement) throws SQLException {
        return this.connection.prepareStatement(statement, Statement.RETURN_GENERATED_KEYS);
    }

    /**
     * Creates an encrypted prepared statement. By default, all the values used in this statement
     * are encrypted. To set an un-encrypted values, special methods must be used.
     *
     * @param statement
     * @throws SQLException
     */
    public EncryptedPreparedStatement createEncryptedPreparedStatement(String statement) throws SQLException {
        return new EncryptedPreparedStatement(
                this.createPreparedStatement(statement),
                CryptoSingleton.getCrypto()
        );
    }
}
