package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.*;
import be.ac.ulb.infof307.g02.exceptions.model.InvalidProductAmountException;
import be.ac.ulb.infof307.g02.models.GroceryList;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.User;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class responsible to get and update info on the grocery lists in the database.
 */
public final class GroceryListDB extends ObjectDB {

    /**
     * Class constructor that fetches the connection to the production database. Throws an exception if
     * the connection is not successful.
     *
     * @throws DatabaseConnectionException
     */
    public GroceryListDB() throws DatabaseConnectionException {
        super();
    }

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public GroceryListDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }

    /**
     * Insert a new GroceryList in the database
     *
     * @param listOwner The owner of the list
     * @param list      the GroceryList
     */
    public void addGroceryList(User listOwner, GroceryList list)
            throws GroceryListInsertionException, GroceryListItemInsertException {
        String listName = list.getGroceryListName();

        try {
            // Insert the list in the GroceryList Table
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("INSERT INTO GroceryList(ListOwner, ListName) VALUES(?,?)");
            statement.setUnencryptedString(1, listOwner.getUserName());
            statement.setString(2, listName);
            statement.execute();

            // Insert all the items of the list in the
            // GroceryListContainsProduct Table
            for (ProductAmount item : list.getProductAmountList()) {
                this.addItemToGroceryList(listOwner, listName, item);
            }

        } catch (SQLException e) {
            throw new GroceryListInsertionException(
                    "The GroceryList couldn't be inserted in the database" + e.getMessage(), e);
        }
    }

    /**
     * Get all the names of the GroceryLists of a user from the database
     *
     * @param owner The owner of the list
     * @return a list containing the names of all GroceryLists
     */
    public List<String> getAllGroceryListsName(User owner) {
        List<String> nameList = new ArrayList<>();
        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("SELECT ListName FROM GroceryList WHERE ListOwner=?" +
                            "GROUP BY ListName");
            statement.setUnencryptedString(1, owner.getUserName());
            statement.execute();
            EncryptedResultSet result = statement.getEncryptedResultSet();
            while (result.next()) {
                nameList.add(result.getString("ListName"));
            }
        } catch (SQLException e) {
            // Don't need to throw it
        }
        return nameList;
    }

    /**
     * Get a GroceryList from the database
     *
     * @param listOwner The owner of the list
     * @param listName  The name of the list
     * @return the Grocery list
     */
    public GroceryList getGroceryList(User listOwner, String listName)
            throws DatabaseException {
        GroceryList resultGroceryList = new GroceryList(listName);

        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("SELECT GroceryListContainsProduct.Product, Quantity"
                            + " FROM GroceryListContainsProduct"
                            + " WHERE ListOwner=? and ListName=?");
            statement.setUnencryptedString(1, listOwner.getUserName());
            statement.setString(2, listName);
            statement.execute();

            ProductDB productDB = new ProductDB(this.getDatabaseConnection());

            EncryptedResultSet result = statement.getEncryptedResultSet();
            while (result.next()) {
                String productName = result.getString("Product");
                Product product = new Product(productDB.getProduct(productName));
                double quantity = result.getDouble("Quantity");
                ProductAmount item = new ProductAmount(product, quantity);
                resultGroceryList.addProductAmount(item);
            }
        } catch (SQLException | InvalidProductAmountException e) {
            throw new DatabaseException("Error getting the GroceryList", e);
        } catch (NoProductFoundException e) {
            // Will Never Happen
        }

        return resultGroceryList;
    }

    /**
     * Delete a GroceryList from the database
     *
     * @param listOwner the user whose the list belongs to
     * @param listName  the name of the list
     */
    public void deleteGroceryList(User listOwner, String listName) throws DatabaseException {
        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("DELETE FROM GroceryList WHERE ListOwner = ? and ListName = ?");
            statement.setUnencryptedString(1, listOwner.getUserName());
            statement.setString(2, listName);
            statement.execute();
        } catch (SQLException e) {
            throw new DatabaseException("An error has happened in the database", e);
        }
    }

    /**
     * Add a GroceryListItem to a given GroceryList
     *
     * @param listOwner the user whose the list belongs to
     * @param listName  the name of the list the item should be inserted in
     * @param item      the item we want to insert
     */
    public void addItemToGroceryList(User listOwner, String listName, ProductAmount item)
            throws GroceryListItemInsertException {
        String productName = item.getProduct().getName();
        double quantity = item.getAmount();

        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection().createEncryptedPreparedStatement(
                    "INSERT INTO GroceryListContainsProduct(ListOwner, ListName, Product, Quantity) VALUES(?,?,?,?)");
            statement.setUnencryptedString(1, listOwner.getUserName());
            statement.setString(2, listName);
            statement.setString(3, productName);
            statement.setDouble(4, quantity);
            statement.execute();
        } catch (SQLException e) {
            throw new GroceryListItemInsertException(
                    "Couldn't insert " + item.toString() + "in the database " + e.getMessage(), e);
        }
    }
}
