package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.crypto.NoSuchKeysFoundForUserException;
import be.ac.ulb.infof307.g02.exceptions.crypto.UserAlreadyHasKeysException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.models.User;
import sun.security.rsa.RSAPublicKeyImpl;

import java.security.InvalidKeyException;
import java.security.PublicKey;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;


/**
 * Used to store, update and fetch key information (RSA public key, encrypted AES key, salt) in the database.
 */
public class KeyDB extends ObjectDB {

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public KeyDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }


    /**
     * Return the user's salt.
     *
     * @param userName
     * @throws NoSuchKeysFoundForUserException
     * @throws DatabaseException
     */
    public byte[] getSalt(String userName) throws NoSuchKeysFoundForUserException, DatabaseException {
        return this.getBytes(userName, "Salt");
    }

    /**
     * Return user's encrypted AES key
     *
     * @param userName
     * @throws DatabaseException
     * @throws NoSuchKeysFoundForUserException
     */
    public byte[] getEncryptedAESKey(String userName) throws DatabaseException, NoSuchKeysFoundForUserException {
        return this.getBytes(userName, "EncryptedAESKey");
    }


    /**
     * Return user's RSA public key
     *
     * @param userName
     * @throws DatabaseException
     * @throws NoSuchKeysFoundForUserException
     */
    public PublicKey getPublicKey(String userName) throws DatabaseException, NoSuchKeysFoundForUserException {
        byte[] publicKeyBytes = this.getBytes(userName, "PublicKey");
        PublicKey publicKey = null;
        try {
            publicKey = new RSAPublicKeyImpl(publicKeyBytes);
        } catch (InvalidKeyException e) {
            // Will never happen
        }
        return publicKey;
    }

    /**
     * Fetches the bytes in the given column of the Key table for the given user
     *
     * @param userName
     * @param columnName
     * @throws DatabaseException
     * @throws NoSuchKeysFoundForUserException
     */
    private byte[] getBytes(String userName, String columnName) throws DatabaseException, NoSuchKeysFoundForUserException {
        try {
            PreparedStatement statement = this.getDatabaseConnection().createPreparedStatement("SELECT " + columnName + " FROM Key WHERE UserName=?");
            statement.setString(1, userName);
            statement.execute();
            ResultSet result = statement.getResultSet();

            if (result.next()) {
                return result.getBytes(columnName);
            } else {
                throw new NoSuchKeysFoundForUserException("No " + columnName + " found for user " + userName);
            }
        } catch (SQLException e) {
            throw new DatabaseException("Error fetching the " + columnName + " of user " + userName, e);
        }
    }

    /**
     * Store user's public data (RSA public key, encrypted AES key and salt) in the key table
     *
     * @param username
     * @param publicKey
     * @param encryptedAESKey
     * @param salt
     * @throws UserAlreadyHasKeysException
     */
    public void storePublicData(String username, PublicKey publicKey, byte[] encryptedAESKey, byte[] salt) throws UserAlreadyHasKeysException {
        try {
            PreparedStatement statement = this.getDatabaseConnection()
                    .createPreparedStatement("INSERT INTO Key (UserName, PublicKey, EncryptedAESKey, Salt) VALUES(?,?,?,?)");
            statement.setString(1, username);
            statement.setBytes(2, publicKey.getEncoded());
            statement.setBytes(3, encryptedAESKey);
            statement.setBytes(4, salt);
            statement.executeUpdate();
            statement.close();
        } catch (SQLException e) {
            throw new UserAlreadyHasKeysException("User " + username + " already has public key, encrypted AES key and salt", e);
        }
    }

    /**
     * Delete the user's public data (RSA public key, encrypted AES key, salt) from the key table
     *
     * @param updatedUser
     * @throws NoSuchKeysFoundForUserException
     */
    public void deletePublicData(User updatedUser) throws NoSuchKeysFoundForUserException {
        try {
            // We delete the key for using the new username, because old user name was changed
            // to new username because of CASCADE
            PreparedStatement statement = this.getDatabaseConnection()
                    .createPreparedStatement("DELETE FROM Key WHERE UserName=?");
            statement.setString(1, updatedUser.getUserName());
            statement.execute();
        } catch (SQLException e) {
            throw new NoSuchKeysFoundForUserException("Couldn't delete public data. User " + updatedUser.getUserName() + "not found.", e);
        }
    }
}
