package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.exceptions.crypto.NoKeyPairPresentException;
import be.ac.ulb.infof307.g02.exceptions.crypto.NoSuchKeysFoundForUserException;
import be.ac.ulb.infof307.g02.exceptions.crypto.UserAlreadyHasKeysException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.models.User;
import sun.security.rsa.RSAPrivateCrtKeyImpl;

import javax.crypto.SecretKey;
import javax.crypto.spec.PBEParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.*;
import java.security.*;
import java.security.cert.CertificateException;
import java.util.Collections;
import java.util.Enumeration;

/**
 * Allows storing and fetching cryptographic keys for a given user (both RSA and AES keys) as well as updating them.
 * These keys are stored in the keystore.jks file (RSA private key) and the Key table (everything else)
 */
public class KeyStorer {

    /**
     * Salt Size. Each user has a salt, which is used as an input for encryption and decryption,
     * making dictionary/rainbow table much harder.
     */
    private static final int SALT_SIZE = 32;

    /**
     * Parameters used to configure the Password Based Encryption of the keystore
     */
    private static final int ITERATIONS = 100_000;
    private static final String KEYSTORE_FORMAT = "PKCS12";
    private static final String ALGORITHM = "AES";
    private static final String PRIVATE_KEY_ALIAS = "privateKey";
    private static final String PASSWORD_PROTECTION_ALGORITHM = "PBEWithHmacSHA512AndAES_128";
    private static final String KEYSTORE_PASSWORD = "password";

    private KeyStore keyStore;
    private KeyDB keyDB;

    /**
     * Loads the keystore file and does setup using the database connection.
     *
     * @param databaseConnection
     */
    public KeyStorer(DatabaseConnection databaseConnection) {
        try {
            this.keyDB = new KeyDB(databaseConnection);
            this.keyStore = KeyStore.getInstance(KEYSTORE_FORMAT);
            InputStream stream = KeyStorer.getKeystoreFileInputStream();
            if (stream == null) {
                this.keyStore.load(null, KEYSTORE_PASSWORD.toCharArray());
            } else {
                this.keyStore.load(stream, KEYSTORE_PASSWORD.toCharArray());
                stream.close();
            }

        } catch (NoSuchAlgorithmException | CertificateException | KeyStoreException | IOException e) {
            // Will not happen
        }
    }

    private static InputStream getKeystoreFileInputStream() {
        try {
            return new FileInputStream(new File(Constants.KEYSTORE_FILE_PATH));
        } catch (FileNotFoundException e) {
            // Will never happen;
            return null;
        }
    }

    /**
     * Return the user's RSA key pair
     *
     * @param user
     * @throws NoKeyPairPresentException
     * @throws DatabaseException
     */
    public KeyPair getKeyPair(User user) throws NoKeyPairPresentException, DatabaseException {
        try {
            String userName = user.getUserName();
            PublicKey publicKey = this.keyDB.getPublicKey(userName);
            PrivateKey privateKey = this.getPrivateKey(user);
            return new KeyPair(publicKey, privateKey);
        } catch (NoSuchAlgorithmException e) {
            // Will never happen
            return null;
        } catch (UnrecoverableEntryException | NoSuchKeysFoundForUserException | KeyStoreException e) {
            throw new NoKeyPairPresentException();
        }
    }

    /**
     * Return user's RSA private key
     *
     * @param user
     * @throws NoSuchKeysFoundForUserException
     * @throws DatabaseException
     * @throws NoSuchAlgorithmException
     * @throws UnrecoverableEntryException
     * @throws KeyStoreException
     */
    private PrivateKey getPrivateKey(User user) throws NoSuchKeysFoundForUserException, DatabaseException, NoSuchAlgorithmException, UnrecoverableEntryException, KeyStoreException {
        byte[] salt = this.keyDB.getSalt(user.getUserName());
        KeyStore.PasswordProtection passwordProtectionPolicy = this.getPasswordProtection(user.getHashedPassword(), salt);
        KeyStore.SecretKeyEntry entry = (KeyStore.SecretKeyEntry) this.keyStore.getEntry(getAliasForPrivateKey(user), passwordProtectionPolicy);
        byte[] privateKeyBytes = entry.getSecretKey().getEncoded();
        PrivateKey privateKey = null;
        try {
            privateKey = RSAPrivateCrtKeyImpl.newKey(privateKeyBytes);
        } catch (InvalidKeyException e) {
            // Will never happen
        }
        return privateKey;
    }

    /**
     * Returns the alias for the user's private key - the alias is used to store/fetch the key in the keystore.jks file
     *
     * @param user
     */
    private String getAliasForPrivateKey(User user) {
        return PRIVATE_KEY_ALIAS + user.getUserName();
    }

    /**
     * Return user's encrypted AES key
     *
     * @param user
     * @throws DatabaseException
     * @throws NoSuchKeysFoundForUserException
     */
    public byte[] getEncryptedAESKey(User user) throws DatabaseException, NoSuchKeysFoundForUserException {
        return this.keyDB.getEncryptedAESKey(user.getUserName());
    }


    /**
     * Return password protection used for the user in the keystore
     *
     * @param password
     * @param salt
     */
    private KeyStore.PasswordProtection getPasswordProtection(String password, byte[] salt) {
        return new KeyStore.PasswordProtection(
                password.toCharArray(),
                PASSWORD_PROTECTION_ALGORITHM,
                new PBEParameterSpec(salt, ITERATIONS));
    }

    /**
     * Gets an InputStream to the keystore resource
     */
    public static InputStream getKeystoreResourceInputStream() {
        return KeyStorer.class.getClassLoader().getResourceAsStream(Constants.KEYSTORE_RESOURCE_PATH);
    }

    /**
     * Returns an output stream to the keystore.jks file
     */
    private OutputStream getKeystoreOutputStream() {
        OutputStream output = null;
        try {
            File file = new File(Constants.KEYSTORE_FILE_PATH);
            output = new FileOutputStream(file);
        } catch (FileNotFoundException e) {
            // No need to handle: always exists
        }
        return output;
    }

    /**
     * Store the user's key chain (RSA keypair, encrypted AES key and salt)
     *
     * @param keyPair
     * @param encryptedAESKey
     * @param user
     * @throws UserAlreadyHasKeysException
     */
    public void storeAllKeys(KeyPair keyPair, byte[] encryptedAESKey, User user) throws UserAlreadyHasKeysException {
        try {
            byte[] salt = this.createSalt();
            this.storePrivateKey(keyPair.getPrivate(), user, salt);
            this.keyDB.storePublicData(user.getUserName(), keyPair.getPublic(), encryptedAESKey, salt);
            OutputStream outputStream = this.getKeystoreOutputStream();
            this.keyStore.store(outputStream, KEYSTORE_PASSWORD.toCharArray());
            outputStream.close();
        } catch (KeyStoreException | NoSuchAlgorithmException | IOException | CertificateException e) {
            // Will never appear
            int a = 5;
        }
    }

    /**
     * Store user's private key in the keystore.jks file
     *
     * @param privateKey
     * @param user
     * @param salt
     * @throws KeyStoreException
     */
    private void storePrivateKey(PrivateKey privateKey, User user, byte[] salt) throws KeyStoreException {
        SecretKey secretKey = new SecretKeySpec(privateKey.getEncoded(), ALGORITHM);
        KeyStore.PasswordProtection passwordProtectionPolicy = this.getPasswordProtection(user.getHashedPassword(), salt);
        this.keyStore.setEntry(getAliasForPrivateKey(user), new KeyStore.SecretKeyEntry(secretKey), passwordProtectionPolicy);
    }

    /**
     * Create a salt that is used to protect the data from rainbow/dictionary attacks
     */
    private byte[] createSalt() {
        byte[] salt = new byte[SALT_SIZE];
        new SecureRandom().nextBytes(salt);
        return salt;
    }

    /**
     * Update user's keys (required when they change their password or username).
     *
     * @param keyPair
     * @param encryptedAesKey
     * @param oldUser
     * @param updatedUser
     */
    public void updateAllKeys(KeyPair keyPair, byte[] encryptedAesKey, User oldUser, User updatedUser) {
        try {
            this.deletePrivateKey(oldUser);
            this.deleteAllPublicInfo(updatedUser);
            this.storeAllKeys(keyPair, encryptedAesKey, updatedUser);
        } catch (KeyStoreException | UserAlreadyHasKeysException | NoSuchKeysFoundForUserException e) {
            // Won't happen
        }
    }

    /**
     * Delete user's salt, RSA public key and encrypted AES key .
     *
     * @param updatedUser
     * @throws NoSuchKeysFoundForUserException
     */
    private void deleteAllPublicInfo(User updatedUser) throws NoSuchKeysFoundForUserException {
        this.keyDB.deletePublicData(updatedUser);
    }

    /**
     * Delete the user's RSA private key
     *
     * @param oldUser
     * @throws KeyStoreException
     */
    private void deletePrivateKey(User oldUser) throws KeyStoreException {
        this.keyStore.deleteEntry(getAliasForPrivateKey(oldUser));
    }

    public void emptyKeyStore() {
        try {
            Enumeration<String> aliases = keyStore.aliases();
            for (String alias : Collections.list(aliases)) {
                keyStore.deleteEntry(alias);
            }
        } catch (KeyStoreException e) {
            e.printStackTrace();
        }
    }
}
