package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoProductFoundException;
import be.ac.ulb.infof307.g02.exceptions.database.ProductAlreadyExistsException;
import be.ac.ulb.infof307.g02.models.Product;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class responsible to get and update info on the users in the database.
 */
public class ProductDB extends ObjectDB {

    /**
     * Class constructor that fetches the connection to the production database. Throws an exception if
     * the connection is not successful.
     *
     * @throws DatabaseConnectionException
     */
    public ProductDB() throws DatabaseConnectionException {
        super();
    }

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public ProductDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }

    /**
     * Creates a product from the information in the given ResultSet
     *
     * @param result
     * @throws SQLException
     */
    public static Product getProductFromResultSet(ResultSet result) throws SQLException {
        String name = result.getString("Name");
        Product.Unit unit = Product.Unit.create(result.getString("Unit"));
        double proteins = result.getDouble("Proteins");
        double lipids = result.getDouble("Lipids");
        double carbohydrates = result.getDouble("Carbohydrates");
        return new Product(name, unit, lipids, proteins, carbohydrates);
    }

    /**
     * Add a product in the database
     *
     * @param product
     * @throws ProductAlreadyExistsException
     */
    public void addProduct(Product product) throws ProductAlreadyExistsException {
        try {
            PreparedStatement statement = this.getDatabaseConnection()
                    .createPreparedStatement(
                            "INSERT INTO Product (Name, Unit, Proteins, Lipids, Carbohydrates) VALUES(?,?,?,?,?)");
            statement.setString(1, product.getName());
            statement.setString(2, product.getUnitName());
            statement.setDouble(3, product.getProteinGramsPerUnit());
            statement.setDouble(4, product.getLipidGramsPerUnit());
            statement.setDouble(5, product.getCarbohydrateGramsPerUnit());
            statement.execute();
        } catch (SQLException e) {
            throw new ProductAlreadyExistsException("The product with name " + product.getName() + " already exists");
        }
    }

    /**
     * Retrieve the product "productName" from the database
     *
     * @param productName
     * @return a Product object
     */
    public Product getProduct(String productName) throws DatabaseException, NoProductFoundException {
        try {
            PreparedStatement statement = this.getDatabaseConnection()
                    .createPreparedStatement(
                            "SELECT Name, Unit, Proteins, Lipids, Carbohydrates FROM Product WHERE Name=?");
            statement.setString(1, productName);
            statement.execute();
            ResultSet result = statement.getResultSet();
            if (result.next()) {
                // Retrieve the shop information
                return getProductFromResultSet(result);
            } else {
                throw new NoProductFoundException("No product with name " + productName);
            }
        } catch (SQLException e) {
            throw new DatabaseException("Error fetching the product with name " + productName + e.getMessage(), e);
        }
    }

    /**
     * Retrieve all the products from the database
     *
     * @return a List of products
     */
    public List<Product> getAllProducts() throws DatabaseException {

        List<Product> allProducts = new ArrayList<>();

        try {
            PreparedStatement statement = this.getDatabaseConnection().
                    createPreparedStatement(
                            "SELECT Name, Unit, Proteins, Lipids, Carbohydrates FROM Product");
            statement.execute();
            ResultSet result = statement.getResultSet();

            String name;
            Product.Unit unit;
            double proteins;
            double lipids;
            double carbohydrates;
            while (result.next()) {
                name = result.getString("Name");
                unit = Product.Unit.create(result.getString("Unit"));
                proteins = result.getDouble("Proteins");
                lipids = result.getDouble("Lipids");
                carbohydrates = result.getDouble("Carbohydrates");
                Product resProduct = new Product(name, unit, lipids, proteins, carbohydrates);
                allProducts.add(resProduct);
            }
        } catch (SQLException e) {
            throw new DatabaseException("Couldn't get all products.", e);
        }

        return allProducts;
    }

}
