package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.*;
import be.ac.ulb.infof307.g02.exceptions.model.InvalidProductAmountException;
import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.Recipe;
import be.ac.ulb.infof307.g02.models.User;
import javafx.util.Pair;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class responsible to get and update info on the recipes in the database.
 */
public class RecipeDB extends ObjectDB {

    /**
     * Class constructor that fetches the connection to the production database. Throws an exception if
     * the connection is not successful.
     *
     * @throws DatabaseConnectionException
     */
    public RecipeDB() throws DatabaseConnectionException {
        super();
    }

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public RecipeDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }

    /**
     * Insert a new Recipe in the database
     *
     * @param recipeOwner The owner of the recipe
     * @param recipe      the Recipe
     * @throws RecipeInsertionException
     * @throws RecipeItemInsertException
     */
    public void addRecipe(User recipeOwner, Recipe recipe)
            throws RecipeInsertionException, RecipeItemInsertException {
        String recipeName = recipe.getName();
        String instructions = recipe.getInstructions();
        int numPeople = recipe.getNumberOfPeople();

        try {
            // Insert the recipe in the Recipe Table
            String query = "INSERT INTO Recipe(RecipeOwner, RecipeName, NumPeople, Instructions) VALUES(?,?,?,?)";
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement(query);
            statement.setUnencryptedString(1, recipeOwner.getUserName());
            statement.setString(2, recipeName);
            statement.setInt(3, numPeople);
            statement.setString(4, instructions);
            statement.execute();

            // Insert all the items of the recipe in the
            // RecipeContainsProduct Table
            for (ProductAmount item : recipe.getProductAmountList()) {
                this.addProductAmountToRecipe(recipeOwner, recipeName, item);
            }

        } catch (SQLException e) {
            throw new RecipeInsertionException(
                    "The Recipe couldn't be inserted in the database" + e.getMessage());
        }

    }


    /**
     * Get a Recipe from the database
     *
     * @param recipeOwner The owner of the recipe
     * @param recipeName  The name of the recipe
     * @return the recipe
     * @throws NoSuchRecipeException
     * @throws RecipeCreationException
     */
    public Recipe getRecipe(User recipeOwner, String recipeName)
            throws DatabaseException, NoSuchRecipeException, RecipeCreationException {

        try {

            Pair<Integer, String> generalRecipeInfos = this.getRecipeGeneralInfos(recipeName, recipeOwner);
            int numPeople = generalRecipeInfos.getKey();
            String instructions = generalRecipeInfos.getValue();

            List<ProductAmount> recipeProductAmounts = this.getRecipeProductAmounts(recipeName, recipeOwner);

            Recipe recipe = new Recipe(recipeName, numPeople, instructions);
            for (ProductAmount productAmount : recipeProductAmounts) {
                recipe.addProductAmount(productAmount);
            }

            return recipe;
        } catch (InvalidProductAmountException | ProductAmountsFetchException e) {
            throw new DatabaseException("Error getting the Recipe", e);
        }
    }

    /**
     * Gets the Instructions of a recipe and the number of people from the database
     *
     * @param recipeName  The name of the recipe
     * @param recipeOwner The owner of the recipe
     * @return A Pair containing the number of people and the instructions
     * @throws NoSuchRecipeException
     */
    private Pair<Integer, String> getRecipeGeneralInfos(String recipeName, User recipeOwner)
            throws NoSuchRecipeException {

        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("SELECT Instructions, NumPeople From Recipe WHERE RecipeOwner=?" +
                            "and RecipeName=?");
            statement.setUnencryptedString(1, recipeOwner.getUserName());
            statement.setString(2, recipeName);
            statement.execute();
            EncryptedResultSet result = statement.getEncryptedResultSet();

            String instructions;
            int numPeople;
            if (result.next()) {
                instructions = result.getString("Instructions");
                numPeople = result.getInt("NumPeople");
            } else {
                throw new NoSuchRecipeException("No recipe with name " + recipeName + " for owner " + recipeOwner);
            }

            return new Pair<>(numPeople, instructions);

        } catch (SQLException e) {
            throw new NoSuchRecipeException("No recipe with name " + recipeName + " for owner " + recipeOwner);
        }
    }

    /**
     * Gets all the ProductAmounts of a recipe
     *
     * @param recipeName  The name of the recipe
     * @param recipeOwner The owner of the recipe
     * @return A list of ProductAmounts
     * @throws ProductAmountsFetchException
     */
    private List<ProductAmount> getRecipeProductAmounts(String recipeName, User recipeOwner)
            throws ProductAmountsFetchException {

        List<ProductAmount> productAmounts = new ArrayList<>();
        ProductDB productDB = new ProductDB(this.getDatabaseConnection());

        try {
            // Fetch products and amounts
            EncryptedPreparedStatement statementProducts = this.getDatabaseConnection().createEncryptedPreparedStatement(
                    "SELECT RecipeContainsProduct.Product, Quantity" + " FROM RecipeContainsProduct"
                            + " WHERE RecipeOwner=? and RecipeName=?");
            statementProducts.setUnencryptedString(1, recipeOwner.getUserName());
            statementProducts.setString(2, recipeName);
            statementProducts.execute();

            EncryptedResultSet resultProducts = statementProducts.getEncryptedResultSet();
            while (resultProducts.next()) {
                Product product = productDB.getProduct(resultProducts.getString("Product"));
                double quantity = resultProducts.getDouble("Quantity");
                ProductAmount productAmount = new ProductAmount(product, quantity);
                productAmounts.add(productAmount);
            }

        } catch (SQLException | DatabaseException | NoProductFoundException e) {
            throw new ProductAmountsFetchException("Couldn't fetch all the ProductAmounts of the recipe from the" +
                    "database", e);
        }
        return productAmounts;
    }

    /**
     * Gets all the names of the Recipes of a user from the database
     *
     * @param owner The owner of the recipe
     * @return a list containing the names of all Recipes
     */
    public List<String> getAllRecipesName(User owner) throws DatabaseException {
        List<String> nameList = new ArrayList<>();
        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("SELECT RecipeName FROM Recipe WHERE RecipeOwner=? GROUP BY " +
                            "RecipeName");
            statement.setUnencryptedString(1, owner.getUserName());
            statement.execute();

            EncryptedResultSet result = statement.getEncryptedResultSet();
            while (result.next()) {
                nameList.add(result.getString("RecipeName"));
            }
        } catch (SQLException e) {
            throw new DatabaseException("Couldn't get all the recipes names", e);
        }

        return nameList;
    }

    /**
     * Deletes a Recipe from the database
     *
     * @param recipeOwner The user whose the recipe belongs to
     * @param recipeName  The name of the recipe
     * @throws DatabaseException
     */
    public void deleteRecipe(User recipeOwner, String recipeName) throws DatabaseException {
        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection()
                    .createEncryptedPreparedStatement("DELETE FROM Recipe WHERE RecipeOwner=? and RecipeName=?");
            statement.setUnencryptedString(1, recipeOwner.getUserName());
            statement.setString(2, recipeName);
            statement.execute();
        } catch (SQLException e) {
            throw new DatabaseException("An error has occurred in the database", e);
        }
    }

    /**
     * Add a ProductAmount to a given Recipe
     *
     * @param recipeOwner the user whose the recipe belongs to
     * @param recipeName  the name of the recipe the item should be inserted in
     * @param item        the item we want to insert
     * @throws RecipeItemInsertException
     */
    public void addProductAmountToRecipe(User recipeOwner, String recipeName, ProductAmount item)
            throws RecipeItemInsertException {

        String productName = item.getProduct().getName();
        double amount = item.getAmount();

        try {
            EncryptedPreparedStatement statement = this.getDatabaseConnection().createEncryptedPreparedStatement(
                    "INSERT INTO RecipeContainsProduct(RecipeOwner, RecipeName, Product, Quantity) VALUES(?,?,?,?)");
            statement.setUnencryptedString(1, recipeOwner.getUserName());
            statement.setString(2, recipeName);
            statement.setString(3, productName);
            statement.setDouble(4, amount);
            statement.execute();
        } catch (SQLException e) {
            throw new RecipeItemInsertException(
                    "Couldn't insert " + item.toString() + "in the database " + e.getMessage());
        }
    }
}
