package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.*;
import be.ac.ulb.infof307.g02.exceptions.model.SerializationException;
import be.ac.ulb.infof307.g02.models.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class responsible to get and update info on the shops in the database.
 */
public class ShopDB extends ObjectDB {

    /**
     * Class constructor that fetches the connection to the production database. Throws an exception if
     * the connection is not successful.
     *
     * @throws DatabaseConnectionException
     */
    public ShopDB() throws DatabaseConnectionException {
        super();
    }

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public ShopDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }

    /**
     * Insert a new entry in the table Shop as well as in the table Schedule from
     * the database. The entry for each shop contains a name and a position
     * (latitude and longitude).
     * <p>
     * ATTENTION: the shop ID is set to the ID given by the database.
     *
     * @param shop
     * @return the ID given by the database to the shop
     */
    public int addShop(Shop shop) throws ShopInsertionException {
        try {
            int shopID = this.insertEmptyShop(shop);
            this.insertSchedule(shop);
            this.insertStockForEmptyShop(shop);
            return shopID;

        } catch (EmptyShopInsertionException | ScheduleInsertionException e) {
            throw new ShopInsertionException("Couldn't insert the shop in the Database", e);
        }

    }

    /**
     * Inserts the basic information of a shop (no schedule or stock) in the database.
     *
     * @param shop
     * @throws EmptyShopInsertionException
     */
    private int insertEmptyShop(Shop shop) throws EmptyShopInsertionException {
        try {
            PreparedStatement shopStatement = this.getDatabaseConnection()
                    .createPreparedStatementWithGeneratedKeys(
                            "INSERT INTO Shop (Name, Latitude, Longitude) VALUES(?,?,?)");
            shopStatement.setString(1, shop.getName());
            shopStatement.setDouble(2, shop.getPosition().getLatitude());
            shopStatement.setDouble(3, shop.getPosition().getLongitude());
            shopStatement.execute();

            // Get the ID of the shop that has just been added.
            ResultSet shopIDs = shopStatement.getGeneratedKeys();
            shopIDs.next();
            int shopID = shopIDs.getInt(1);
            shop.setId(shopID);
            return shopID;

        } catch (SQLException e) {
            throw new EmptyShopInsertionException("Couldn't insert the empty shop in the Database", e);
        }
    }

    /**
     * Inserts the schedule of a shop in the database
     *
     * @param shop The shop that owns the schedule
     */
    private void insertSchedule(Shop shop) throws ScheduleInsertionException {
        try {
            PreparedStatement scheduleStatement = this.getDatabaseConnection()
                    .createPreparedStatement("INSERT INTO Schedule (ShopID, SerializedSchedule) VALUES(?,?)");
            scheduleStatement.setInt(1, shop.getId());
            Schedule schedule = shop.getSchedule();

            // Serialize the schedule
            byte[] serializedTimeSlots = Serializer.serialize(schedule.getDaySchedules());
            ByteArrayInputStream bais = new ByteArrayInputStream(serializedTimeSlots);
            scheduleStatement.setBinaryStream(2, bais, serializedTimeSlots.length);

            // Insert the schedule
            scheduleStatement.execute();

        } catch (SQLException | SerializationException e) {
            throw new ScheduleInsertionException("Couldn't insert the shop schedule in the database", e);
        }
    }

    /**
     * Inserts the stock of a shop in the database
     *
     * @param shop
     */
    private void insertStockForEmptyShop(Shop shop) {
        int shopID = shop.getId();
        for (Article article : shop.getStock().values()) {
            try {
                this.insertArticleInShop(shopID, article);
            } catch (ArticleAlreadyInShopException e) {
                // Will never happen because this is used to add the stock of a new shop
            }
        }
    }

    /**
     * Inserts an article into a shop in the database
     *
     * @param shopID
     * @param article
     * @throws ArticleAlreadyInShopException
     */
    public void insertArticleInShop(int shopID, Article article) throws ArticleAlreadyInShopException {
        double price = article.getPrice();
        String product = article.getProductName();

        try {
            PreparedStatement statement = this.getDatabaseConnection()
                    .createPreparedStatement("INSERT INTO ShopHasProduct(Shop, Product, Price) VALUES(?,?,?)");

            statement.setInt(1, shopID);
            statement.setString(2, product);
            statement.setDouble(3, price);

            statement.execute();
        } catch (SQLException e) {
            throw new ArticleAlreadyInShopException("The article with name " + product + " already exists");
        }
    }

    /**
     * Delete a shop from the database
     *
     * @param shopID The ID of the shop that should be deleted
     * @throws SQLException
     */
    public void deleteShop(int shopID) throws SQLException {
        PreparedStatement deleteStatement = this.getDatabaseConnection().
                createPreparedStatement(
                        "DELETE FROM Shop WHERE ID = ?");
        deleteStatement.setInt(1, shopID);
        deleteStatement.execute();
    }

    /**
     * Gets all the shops from the database
     *
     * @return A List of shops
     * @throws DatabaseException
     */
    public List<Shop> getAllShops() throws DatabaseException {
        List<Shop> shops = new ArrayList<>();
        String query = "SELECT * FROM Shop INNER JOIN Schedule ON SHOP.ID = Schedule.ShopID";

        try {
            ResultSet result = this.getDatabaseConnection().executeQuery(query);
            while (result.next()) {
                // Retrieve the shop information
                int shopID = result.getInt("ID");
                String name = result.getString("Name");
                Position position = new Position(result.getDouble("Latitude"), result.getDouble("Longitude"));
                Schedule schedule = this.getScheduleFromResultSet(result);
                Shop shop = new Shop(shopID, name, schedule, position);
                shops.add(shop);
            }
            fillStocks(shops);
            return shops;

        } catch (SQLException | StockFillingException | ScheduleFetchException e) {
            throw new DatabaseException("Couldn't get all shops.", e);
        }
    }

    /**
     * Gets the schedule of a shop from a ResultSet
     *
     * @param resultSet The resultSet
     * @return The Shop's schedule
     * @throws ScheduleFetchException
     */
    private Schedule getScheduleFromResultSet(ResultSet resultSet)
            throws ScheduleFetchException {
        try {
            // Deserialize the schedule of the shop
            byte[] serializedSchedule = (byte[]) resultSet.getObject("SerializedSchedule");
            ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(serializedSchedule);
            ObjectInputStream objectInputStream = new ObjectInputStream(byteArrayInputStream);

            List<DaySchedule> scheduleTimeSlots = (List<DaySchedule>) objectInputStream.readObject();
            return new Schedule(scheduleTimeSlots);

        } catch (IOException | SQLException | ClassNotFoundException e) {
            throw new ScheduleFetchException("Error while getting the Schedule from the database", e);
        }
    }

    /**
     * Fills the stocks of a list of shops
     *
     * @param shops The list of shops
     * @throws StockFillingException
     */
    private void fillStocks(List<Shop> shops) throws StockFillingException {
        try {
            String queryStock = "SELECT * FROM ShopHasProduct INNER JOIN Product on Product.Name = ShopHasProduct.Product";
            ResultSet result = getDatabaseConnection().executeQuery(queryStock);

            while (result.next()) {
                int shopID = result.getInt("Shop");
                double price = result.getDouble("Price");
                Product product = ProductDB.getProductFromResultSet(result);
                for (Shop shop : shops) {
                    if (shop.getId() == shopID) {
                        shop.addToStock(new Article(product, price));
                    }
                }
            }
        } catch (SQLException e) {
            throw new StockFillingException("Error while inserting the Articles in the Stock", e);
        }
    }

}
