package be.ac.ulb.infof307.g02.database;

import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchUserFoundException;
import be.ac.ulb.infof307.g02.exceptions.database.UserNameTakenException;
import be.ac.ulb.infof307.g02.models.User;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * Class responsible to get and update the info on the users in the database.
 */
public class UserDB extends ObjectDB {

    /**
     * Class constructor that fetches the connection to the production database. Throws an exception if
     * the connection is not successful.
     *
     * @throws DatabaseConnectionException
     */
    public UserDB() throws DatabaseConnectionException {
        super();
    }

    /**
     * Class constructor that stores the given database connection for later use.
     *
     * @param databaseConnection
     */
    public UserDB(DatabaseConnection databaseConnection) {
        super(databaseConnection);
    }

    /**
     * Allows updating the profile in the database by giving a new user name to an
     * account. Throws an IllegalArgumentException if the new user name is already
     * taken.
     *
     * @param oldUserName
     * @param newUserName
     */
    public void updateUserName(String oldUserName, String newUserName) throws UserNameTakenException {
        String query = "UPDATE USER SET UserName = ? WHERE UserName = ?";
        try {
            PreparedStatement statement = this.getDatabaseConnection().createPreparedStatement(query);
            statement.setString(1, newUserName);
            statement.setString(2, oldUserName);
            statement.executeUpdate();
        } catch (SQLException e) {
            throw new UserNameTakenException("Error updating the username " + e.getMessage());
        }
    }

    /**
     * Allows updating the password of a user in the database. Throws an
     * IllegalArgumentException if the new password is not valid (ex: "")
     *
     * @param userName
     * @param newHashedPassword
     */
    public void updatePassword(String userName, String newHashedPassword) throws DatabaseException {
        String query = "UPDATE USER SET Password = ? WHERE UserName = ?";
        try {
            PreparedStatement statement = this.getDatabaseConnection().createPreparedStatement(query);
            statement.setString(1, newHashedPassword);
            statement.setString(2, userName);
            statement.executeUpdate();
        } catch (SQLException e) {
            throw new DatabaseException("Error while updating Password" + e.getMessage(), e);
        }
    }

    /**
     * Return the User with the given user name. Throws an IllegalArgumentException
     * if the task was not be completed with success (ex: The user name does not
     * exist in the database)
     *
     * @param userName
     * @return Returns User with the given user name
     * @throws DatabaseException
     * @throws NoSuchUserFoundException
     */
    public User getUserByName(String userName) throws DatabaseException, NoSuchUserFoundException {
        String query = "SELECT UserName, Password FROM USER WHERE UserName = ?";
        try {
            PreparedStatement statement = this.getDatabaseConnection().createPreparedStatement(query);
            statement.setString(1, userName);
            statement.execute();

            ResultSet encryptedResult = statement.getResultSet();
            if (encryptedResult.next()) {
                String userNameFromDB = encryptedResult.getString("UserName");
                byte[] passwordFromDB = encryptedResult.getString("Password").getBytes();

                statement.close();
                return new User(userNameFromDB, passwordFromDB);
            } else {
                throw new NoSuchUserFoundException("There is no user with username " + userName);
            }
        } catch (SQLException e) {
            throw new DatabaseException("Error while getting the user", e);
        }
    }

    /**
     * Inserts a new entry in the database. A new user is created. Throws an
     * IllegalArgumentException if the action is not possible (ex: username already
     * exists).
     *
     * @param user
     * @throws UserNameTakenException
     */
    public void addUser(User user) throws UserNameTakenException {
        String userName = user.getUserName();
        String hashedPassword = user.getHashedPassword();
        String query = "INSERT INTO User(UserName, Password) VALUES(?,?)";
        try {
            PreparedStatement statement = this.getDatabaseConnection().createPreparedStatement(query);
            statement.setString(1, userName);
            statement.setString(2, hashedPassword);
            statement.executeUpdate();
            statement.close();
        } catch (SQLException e) {
            throw new UserNameTakenException("Username already taken " + e.getMessage());
        }
    }
}
