package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.exceptions.database.NoShopHasThatProductException;
import be.ac.ulb.infof307.g02.utils.DistanceCalculator;

import java.util.*;

/**
 * This class attempts to find the best tour using the Commodity Adding Heuristic
 * (CAH) algorithm. CAH starts with an initial solution containing only the
 * start position and the shop that minimizes the total cost for purchasing the
 * first product. At each iteration, a new product is considered and the
 * convenience to purchase the product in one of the shops already included in
 * the current solution or to add another shop is evaluated. The algorithm
 * terminates when all the products have been considered.
 */
public class BestTourFinder {
    private static final int NUM_SHOP_LIMIT_WHERE_WE_STOP_FINDING_OPTIMAL_TOUR = 8;
    private final GroceryList groceryList;
    private final List<Shop> shops;
    private final Position startPos;
    private final Position endPos;
    private final List<Position> bestTourPositions;
    private final Map<Shop, List<ProductAmount>> productsAtShop;
    private List<Shop> bestTourShops;
    private double bestDistance = Double.MAX_VALUE;

    /**
     * Constructor that finds the best tour to buy all the products in the grocery list
     * among the given shops that starts at the startPos and ends at endPos.
     * <p>
     * It uses the Commodity Adding Heuristic (CAH) algorithm to select the shops and then tries to
     * find the best route among those shops if the number of shops is small enough.
     * <p>
     * When the grocery list has a product that no shop has, NoShopHasThatProductException is thrown.
     *
     * @param startPos
     * @param endPos
     * @param shops
     * @param groceryList
     * @throws NoShopHasThatProductException
     */
    public BestTourFinder(Position startPos, Position endPos, List<Shop> shops, GroceryList groceryList)
            throws NoShopHasThatProductException {
        this.bestTourShops = new ArrayList<>();
        this.bestTourPositions = new ArrayList<>();
        this.startPos = startPos;
        this.endPos = endPos;
        // Make copies because we'll add a fake shop and a fake product and we don't
        // want to alter the input data
        this.shops = new ArrayList<>(shops);
        this.groceryList = new GroceryList(groceryList);
        this.productsAtShop = new HashMap<>();

        this.searchForBestTour();
    }

    /**
     * Find the shortest path starting at startPos, going through all the shops in the best tour
     * and ending at endPos, for small instances. Big instances are too expensive since this method
     * solves the travelling salesman problem, which is NP-hard.
     * <p>
     * We use backtracking to find the optimal route.
     *
     * @param solution
     */
    private void findBestRoute(Set<Shop> solution) {
        if (solution.size() > NUM_SHOP_LIMIT_WHERE_WE_STOP_FINDING_OPTIMAL_TOUR) {
            this.bestTourShops = new ArrayList<>(solution);
            this.bestTourPositions.add(this.startPos);
            for (Shop shop : solution) {
                this.bestTourPositions.add(shop.getPosition());
            }
            this.bestTourPositions.add(this.endPos);
            return;
        }

        List<Shop> route = new ArrayList<>();
        double distance = 0;
        ArrayList<Shop> shopList = new ArrayList<>(solution);
        for (int i = 0; i < shopList.size(); i++) {
            double saveDistance = this.distance(this.startPos, shopList.get(i).getPosition());
            Shop examinedShop = shopList.get(i);

            distance += saveDistance;
            route.add(examinedShop);
            shopList.remove(examinedShop);

            this.findBestRouteUsingBacktrack(shopList, route, distance);

            shopList.add(i, examinedShop);
            route.remove(examinedShop);
            distance -= saveDistance;
        }
    }

    /**
     * extension of the findBestRoute algorithm using backtracking.
     * Stop if the actual distance is bigger than the best one.
     *
     * @param shopList
     * @param route
     * @param distance
     */
    private void findBestRouteUsingBacktrack(List<Shop> shopList, List<Shop> route, double distance) {
        double currentDistance = distance;
        Position positionOfLastShopInRoute = route.get(route.size() - 1).getPosition();
        if (shopList.isEmpty()) {
            currentDistance += this.distance(positionOfLastShopInRoute, this.endPos);
            if (currentDistance < this.bestDistance) {
                this.bestDistance = currentDistance;
                // Save best tour
                this.bestTourShops.clear();
                this.bestTourShops.addAll(route);

                // Save best positions
                this.bestTourPositions.clear();
                this.bestTourPositions.add(this.startPos);
                for (Shop shop : this.bestTourShops) {
                    this.bestTourPositions.add(shop.getPosition());
                }
                this.bestTourPositions.add(this.endPos);
            }
        } else {
            for (int i = 0; i < shopList.size(); i++) {
                Shop examinedShop = shopList.get(i);
                double saveDistance = this.distance(positionOfLastShopInRoute, examinedShop.getPosition());
                if (currentDistance + saveDistance < this.bestDistance) {
                    currentDistance += saveDistance;
                    route.add(examinedShop);
                    shopList.remove(examinedShop);
                    this.findBestRouteUsingBacktrack(shopList, route, currentDistance);
                    shopList.add(i, examinedShop);
                    route.remove(examinedShop);
                    currentDistance -= saveDistance;
                }
            }
        }
    }

    /**
     * Finds the best shops to buy all the products in the grocery list. If the number of shops isn't too high,
     * it then finds the best route that goes through those shops.
     * <p>
     * If a product is present in no shop, NoShopHasThatProductException is thrown.
     *
     * @throws NoShopHasThatProductException
     */
    private void searchForBestTour() throws NoShopHasThatProductException {
        Set<Shop> solution = new HashSet<>();
        Set<Shop> remainingShops = new HashSet<>(this.shops);
        for (ProductAmount productAmount : this.groceryList.getProductAmountList()) {
            Shop bestShopForProduct = this.findBestShopForProduct(solution, remainingShops, productAmount);
            solution.add(bestShopForProduct);

            boolean firstSeeingGoingToShop = !this.productsAtShop.containsKey(bestShopForProduct);
            if (firstSeeingGoingToShop) {
                this.productsAtShop.put(bestShopForProduct, new ArrayList<>());
                remainingShops.remove(bestShopForProduct);
            }

            this.productsAtShop.get(bestShopForProduct).add(productAmount);
        }
        this.findBestRoute(solution);
    }


    /**
     * Finds the best shop to buy the given product. It computes to cost of buying it at all the shops we're
     * planning to visit (price cost only) and shops we haven't visited yet (price + distance cost), if they have
     * the product, and return the shop that minimizes that cost.
     * <p>
     * If no shop has the given product, NoShopHasThatProductException is thrown.
     *
     * @param solution
     * @param remainingShops
     * @param productAmount
     * @throws NoShopHasThatProductException
     */
    private Shop findBestShopForProduct(Set<Shop> solution, Set<Shop> remainingShops, ProductAmount productAmount)
            throws NoShopHasThatProductException {
        Shop bestShop = null;
        Double bestCost = Double.MAX_VALUE;

        // Current Solution
        for (Shop shop : solution) {
            Product product = productAmount.getProduct();
            if (shop.hasProduct(product)) {
                double cost = this.getCostForProductInVisitedShop(productAmount, shop);
                if (cost < bestCost) {
                    bestCost = cost;
                    bestShop = shop;
                }
            }
        }

        // Remaining shops
        for (Shop shop : remainingShops) {
            if (shop.hasProduct(productAmount.getProduct())) {
                double cost = this.getCostForProductInNewShop(productAmount, shop, solution);
                if (cost < bestCost) {
                    bestCost = cost;
                    bestShop = shop;
                }
            }
        }

        if (bestShop == null) {
            throw new NoShopHasThatProductException("No shop has the product " + productAmount.getProductName());
        }

        return bestShop;
    }

    /**
     * Gets the cost of buying the product in the shop, given that we already planned to visit the shop to buy another
     * product. Thus there is a price cost but no distance cost.
     *
     * @param productAmount
     * @param shop
     */
    private double getCostForProductInVisitedShop(ProductAmount productAmount, Shop shop) {
        return this.getProductPriceAtShop(productAmount, shop);
    }

    /**
     * Gets the cost of buying the product in the shop, given that we haven't planned yet to go to that shop.
     * Thus, there is a price cost and a distance cost.
     *
     * @param productAmount
     * @param shop
     * @return
     */
    private double getCostForProductInNewShop(ProductAmount productAmount, Shop shop, Set<Shop> solution) {
        return this.getProductPriceAtShop(productAmount, shop) + this.getDistanceCostOfAddingShop(shop, solution);
    }

    /**
     * Return the cost of adding a new shop to the actual tour, using an approximation.
     * <p>
     * For each pair of shops A and B, we find the cost of adding the new between in the tour shop A and B
     * and return the minimum value.
     *
     * @param shop
     * @param solution
     */
    private double getDistanceCostOfAddingShop(Shop shop, Set<Shop> solution) {
        // This is a very crude approximation, the true cost is much more complicated
        // Adding the shop may radically change the optimal path, but we don't consider
        // that. We simply find the 2 closest shop and assume that the shop will be between
        // both in the path.
        // Example - current shops: A -> B -> C -> D -> E
        // If the closest shop to a new shop S is B and C, we suppose that the new best
        // path with S is A -> B -> S -> C -> D -> E
        // We don't even consider the case were the closest shops aren't the closest on
        // the path such as when the closest shop are B and E

        Double bestDistance = Double.MAX_VALUE;
        Position shopPosition = shop.getPosition();
        List<Position> positions = new ArrayList<>();
        for (Shop solutionShop : solution) {
            positions.add(solutionShop.getPosition());
        }
        positions.add(this.startPos);

        for (Position positionA : positions) {
            for (Position positionB : positions) {
                double addShopDistance =
                        this.distance(positionA, shopPosition)
                                + this.distance(shopPosition, positionB)
                                - this.distance(positionA, positionB);
                if (addShopDistance < bestDistance) {
                    bestDistance = addShopDistance;
                }
            }
        }

        return bestDistance * Constants.PRICE_PER_KM;
    }

    /**
     * Give the linear distance between two points
     *
     * @param posA
     * @param posB
     */
    private double distance(Position posA, Position posB) {
        return DistanceCalculator.getLinearDistance(posA, posB);
    }

    /**
     * Returns the price of buying a product in a certain amount at the given shop.
     *
     * @param productAmount
     * @param shop
     * @return
     */
    private double getProductPriceAtShop(ProductAmount productAmount, Shop shop) {
        return productAmount.getAmount() * shop.getProductPrice(productAmount.getProduct());
    }

    /**
     * Returns the shops (in order) of the best tour.
     * Note: it doesn't return the start and end position (since they are not shops).
     */
    public List<Shop> getBestTourShops() {
        return this.bestTourShops;
    }

    /**
     * Returns all the positions of the tour: the start position, the shops (in order) and the end position.
     */
    public List<Position> getBestTourPositions() {
        return this.bestTourPositions;
    }

    /**
     * Returns all the products that the purchaser will buy in a given shop during the best tour.
     *
     * @param shop
     */
    private List<ProductAmount> getProductsAtShop(Shop shop) {
        return this.productsAtShop.get(shop);
    }

    /**
     * Create a BestTour object to store the result of the algorithm
     *
     * @return BestTour object
     */
    public BestTour getBestTour() {
        // Creates all the groceryLists needed for the best tour
        List<GroceryList> groceryLists = new ArrayList<>();
        for (Shop shop : this.getBestTourShops()) {

            GroceryList newList = new GroceryList(shop.getName(), this.getProductsAtShop(shop));
            groceryLists.add(newList);
        }

        return new BestTour(groceryLists, this.getBestTourPositions());
    }

    public Position getStartPos() {
        return this.startPos;
    }

    public Position getEndPos() {
        return this.endPos;
    }
}
