package be.ac.ulb.infof307.g02.models;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Represents a shop's schedule for a day.
 * It implements Serializable because we need to serialize the object to store it in the database.
 */
public class DaySchedule implements Serializable {

    /**
     * Attributes a unique ID to the class to make serialization work. This ID is used
     * to make sure that the deserialized object is of this class.
     */
    private static final long serialVersionUID = 7980881753543942355L;

    /**
     * A DaySchedule is divided into multiple TimeSlots
     */
    private final List<TimeSlot> timeSlots;
    private final Day day;

    /**
     * Default constructor.
     *
     * @param day
     */
    public DaySchedule(Day day) {
        this.day = day;
        this.timeSlots = new ArrayList<>();
    }

    /**
     * Copy constructor.
     *
     * @param daySchedule
     */
    public DaySchedule(DaySchedule daySchedule) {
        this.day = daySchedule.getDay();
        this.timeSlots = new ArrayList<>();
        for (TimeSlot timeSlot : daySchedule.getTimeSlots()) {
            addTimeSlot(timeSlot);
        }
    }

    private Day getDay() {
        return this.day;
    }

    public List<TimeSlot> getTimeSlots() {
        List<TimeSlot> timeSlots = new ArrayList<>();
        for (TimeSlot timeSlot : this.timeSlots) {
            timeSlots.add(new TimeSlot(timeSlot.getOpenTime(), timeSlot.getCloseTime()));
        }
        return timeSlots;
    }

    public String getDayName() {
        return this.day.getName();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        DaySchedule that = (DaySchedule) o;
        return this.getDay() == that.getDay() &&
                Objects.equals(this.timeSlots, that.timeSlots);
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.timeSlots, this.getDay());
    }

    /**
     * Adds a time slot to the day schedule, making sure it doesn't overlap any
     * existing time slot.
     *
     * @param newTimeSlot
     */
    public final void addTimeSlot(TimeSlot newTimeSlot) {
        for (TimeSlot timeSlot : this.timeSlots) {
            if (timeSlot.isTimeSlotOverlapping(newTimeSlot))
                throw new IllegalArgumentException();
        }

        this.timeSlots.add(newTimeSlot);
    }

    /**
     * @return true if there are no TimeSlots in the DaySchedule (the shop is closed for that day).
     */
    public boolean isEmpty() {
        return this.timeSlots.isEmpty();
    }
}
