package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.utils.Utils;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Represents a grocery list.
 */
public class GroceryList extends ProductAmountListOwner {

    private String groceryListName;

    public GroceryList(String listName) {
        this(listName, new ArrayList<>());
    }

    public GroceryList(GroceryList groceryList) {
        this(groceryList.getGroceryListName(), groceryList.getUnderlyingProductAmountList());
    }

    public GroceryList(String groceryListName, List<ProductAmount> productList) {
        super(productList);
        if (groceryListName.trim().equals("")) {
            throw new IllegalArgumentException("Name cannot be empty or contain only whitespaces");
        }
        this.groceryListName = groceryListName;
    }

    public String getGroceryListName() {
        return this.groceryListName;
    }

    /**
     * Deletes an item from a grocery list
     *
     * @param productAmount The item to be deleted
     */
    public void deleteItem(ProductAmount productAmount) {
        this.getUnderlyingProductAmountList().remove(productAmount);
    }

    /**
     * Gets a list of all the products of a grocery list with their amount
     */
    public List<ProductAmount> getProductAmountList() {
        return new ArrayList<>(this.getUnderlyingProductAmountList());
    }

    /**
     * Gets a list of all the products contained in the productAmountList.
     */
    public List<Product> getProductsList() {
        ArrayList<Product> products = new ArrayList<>();
        for (ProductAmount item : this.getUnderlyingProductAmountList()) {
            products.add(item.getProduct());
        }
        return products;
    }

    /**
     * Searches and returns a product in the productAmountList. If product is not
     * found return null (An item is a ProductAmount object).
     *
     * @param name the name of the product.
     * @return the item if found, else null.
     */
    public ProductAmount getProductAmountByProductName(String name) {
        for (ProductAmount item : this.getUnderlyingProductAmountList()) {
            if (item.getProductName().equals(name)) {
                return item;
            }
        }
        return null;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        GroceryList that = (GroceryList) o;
        return Objects.equals(this.groceryListName, that.groceryListName) &&
                Utils.listsHaveSameElements(this.getUnderlyingProductAmountList(), that.getUnderlyingProductAmountList());
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.groceryListName, this.getUnderlyingProductAmountList());
    }

    /**
     * Transforms a GroceryList into an ArrayList of strings in which each
     * string contains the information of a GroceryList element.
     *
     * @return The ArrayList of Strings.
     */
    public List<String> toList() {
        ArrayList<String> productList = new ArrayList<>();
        for (ProductAmount item : this.getUnderlyingProductAmountList()) {
            String info = item.toString();
            productList.add(info);
        }
        return productList;
    }


}
