package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.utils.Utils;

import java.util.List;

/**
 * Calculates the nutritional values (proteins, lipids, carbohydrates and calories)
 */
public class NutritionalValuesCalculator {

    private static final int DOUBLE_PRECISION_VALUE = 2;

    // Don't allow instantiation
    private NutritionalValuesCalculator() {
    }

    /**
     * Calculates the correct nutritional value (proteins, lipids, carbohydrates and calories) of a product with a certain quantity.
     *
     * @param productAmount product with quantity
     */
    public static NutritionalValues computeNutritionalValues(ProductAmount productAmount) {
        Product product = productAmount.getProduct();
        double lipidsGramsPerUnit = product.getLipidGramsPerUnit();
        double proteinsGramsPerUnit = product.getProteinGramsPerUnit();
        double carbohydratesGramsPerUnit = product.getCarbohydrateGramsPerUnit();
        double amountOfProduct = productAmount.getAmount();

        double proteins;
        double lipids;
        double carbohydrates;
        if (product.getUnit() == Product.Unit.KG || product.getUnit() == Product.Unit.LITER) {
            // For products measured in Kg or L we show the number of protein, lipids, carbohydrates and calories per 100g
            proteins = Product.computeMacroAmountPer100g(amountOfProduct, proteinsGramsPerUnit);
            lipids = Product.computeMacroAmountPer100g(amountOfProduct, lipidsGramsPerUnit);
            carbohydrates = Product.computeMacroAmountPer100g(amountOfProduct, carbohydratesGramsPerUnit);
        } else {
            proteins = proteinsGramsPerUnit * amountOfProduct;
            lipids = lipidsGramsPerUnit * amountOfProduct;
            carbohydrates = carbohydratesGramsPerUnit * amountOfProduct;
        }

        double calories = Product.computeNumberOfCalories(proteins, lipids, carbohydrates);
        return new NutritionalValues(Utils.round(proteins, DOUBLE_PRECISION_VALUE), Utils.round(lipids, DOUBLE_PRECISION_VALUE), Utils.round(carbohydrates, DOUBLE_PRECISION_VALUE), Utils.round(calories, DOUBLE_PRECISION_VALUE));
    }

    /**
     * Calculate all the nutritional value for the list of product
     * @param productAmounts
     */
    public static NutritionalValues computeAllNutritionalValues(List<ProductAmount> productAmounts) {
        double totalProteins = 0;
        double totalLipids = 0;
        double totalCarbohydrates = 0;
        double totalCalories = 0;
        for (ProductAmount productAmount : productAmounts) {
            NutritionalValues nutritionalValues = computeNutritionalValues(productAmount);
            totalProteins += nutritionalValues.getProteins();
            totalLipids += nutritionalValues.getLipids();
            totalCarbohydrates += nutritionalValues.getCarbohydrates();
            totalCalories += nutritionalValues.getCalories();
        }
        return new NutritionalValues(totalProteins, totalLipids, totalCarbohydrates, totalCalories);
    }
}
