package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.constants.Constants;

import java.util.Objects;

/**
 * Represents a Product. Stores the name and the unit. We trim the whitespace of
 * the name. The trimmed name can't be empty. The unit must be contained in
 * ["Kg","L","Unit"].
 */
public class Product {

    // Macros are given per 100g, we need to multiply by 10 to get macros per Kg
    private static final int TO_KILO_MULTIPLIER = 10;
    private static final int MIN_VALUE_ACCEPTED = 0;

    private final double lipidGramsPerUnit;
    private final double proteinsGramsPerUnit;
    private final double carbohydrateGramsPerUnit;
    private String name;
    private Unit unit;

    /**
     * Product constructor.
     *
     * @param name,          the name of the product
     * @param unit,          the unit of the product (either Kg, L or Unit)
     * @param lipids,        lipids per 100g (for Kg and L) or per unit (for Unit)
     * @param proteins,      proteins per 100g (for Kg and L) or per unit (for Unit)
     * @param carbohydrates, proteins per 100g (for Kg and L) or per unit (for Unit)
     */
    public Product(String name, Unit unit, double lipids, double proteins, double carbohydrates) {
        if (name.trim().equals("")) {
            throw new IllegalArgumentException("The name must not be empty or filled with only whitespace");
        }

        if (lipids < MIN_VALUE_ACCEPTED || proteins < MIN_VALUE_ACCEPTED || carbohydrates < MIN_VALUE_ACCEPTED) {
            throw new IllegalArgumentException("No negative nutritional values allowed.");
        }

        this.name = name;
        this.unit = unit;
        this.proteinsGramsPerUnit = proteins;
        this.lipidGramsPerUnit = lipids;
        this.carbohydrateGramsPerUnit = carbohydrates;
    }

    /**
     * Copy constructor.
     *
     * @param product
     */
    public Product(Product product) {
        this(product.getName(), product.getUnit(), product.getLipidGramsPerUnit(), product.getProteinGramsPerUnit(), product.getCarbohydrateGramsPerUnit());
    }

    /**
     * Compute the exact quantity of a nutritional value(lipids or proteins or carbohydrates).
     *
     * @param amountOfProduct    quantity of the product
     * @param macroAmountPerUnit the basic quantity of a nutritional value.
     * @return the exact quantity of a nutritional value
     */
    public static double computeMacroAmountPer100g(double amountOfProduct, double macroAmountPerUnit) {
        return amountOfProduct * (macroAmountPerUnit * TO_KILO_MULTIPLIER);
    }

    /**
     * Compute the exact quantity of calories of a product
     * by multiplying each parameter with a specific constant of the parameter and then we sum all of the values
     * to obtain the exact quantity of calories of a product.
     *
     * @param proteins      value in grams of proteins
     * @param lipids        value in grams of lipids
     * @param carbohydrates value in grams of carbohydrates
     * @return the exact quantity of calories of a product
     */
    public static double computeNumberOfCalories(double proteins, double lipids, double carbohydrates) {
        double proteinCalories = (proteins) * Constants.PROTEINS_CALORIES_PER_GRAM;
        double lipidsCalories = (lipids) * Constants.LIPIDS_CALORIES_PER_GRAM;
        double carbohydratesCalories = (carbohydrates) * Constants.CARBOHYDRATE_CALORIES_PER_GRAM;
        return proteinCalories + lipidsCalories + carbohydratesCalories;
    }

    public Unit getUnit() {
        return this.unit;
    }

    public String getName() {
        return this.name;
    }

    public String getUnitName() {
        return this.unit.getUnitName();
    }

    @Override
    public String toString() {
        return this.name + " (" + this.getUnitName() + ")";
    }

    public double getLipidGramsPerUnit() {
        return this.lipidGramsPerUnit;
    }

    public double getProteinGramsPerUnit() {
        return this.proteinsGramsPerUnit;
    }

    public double getCarbohydrateGramsPerUnit() {
        return this.carbohydrateGramsPerUnit;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Product product = (Product) o;
        return Objects.equals(this.name, product.name) &&
                Objects.equals(this.unit, product.unit);
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.name, this.unit);
    }

    /**
     * Enum representing all the possible units that may be used for a product.
     */
    public enum Unit {
        KG("Kg"),
        LITER("L"),
        UNIT("Unit");

        private final String unitName;

        Unit(String unitName) {
            this.unitName = unitName;
        }

        public static Unit create(String unitName) {
            switch (unitName) {
                case "Kg":
                    return KG;
                case "L":
                    return LITER;
                case "Unit":
                    return UNIT;
                default:
                    throw new IllegalArgumentException("That unit doesn't exists.");
            }
        }

        public String getUnitName() {
            return this.unitName;
        }

    }


}
