package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.exceptions.model.InvalidProductAmountException;

import java.util.ArrayList;
import java.util.List;

/**
 * Base abstract class (extended by Recipe and GroceryList) that contains a list of ProductAmounts
 * and that defines useful methods.
 */
public abstract class ProductAmountListOwner {

    private List<ProductAmount> productAmountList;

    /**
     * Constructor that creates a productAmountList from all the given ProductAmounts,
     * and handles the situation when there are 2 productAmounts with the same product
     *
     * @param productAmountList
     */
    public ProductAmountListOwner(List<ProductAmount> productAmountList) {
        this.productAmountList = new ArrayList<>();
        for (ProductAmount productAmount : productAmountList) {
            this.addProductOrIncreaseAmount(productAmount);
        }
    }

    /**
     * Returns a direct reference to the product amount list (not a copy).
     * Only subclasses may access it.
     */
    protected List<ProductAmount> getUnderlyingProductAmountList() {
        return this.productAmountList;
    }

    /**
     * Only used for testing purposes
     *
     * @param productAmountList
     */
    public void setUnderlyingProductAmountList(List<ProductAmount> productAmountList) {
        this.productAmountList = productAmountList;
    }

    /**
     * Adds the product to the list. If the list already contains the item,
     * it throws an exception.
     *
     * @param newProductAmount
     * @throws InvalidProductAmountException
     */
    public final void addProductAmount(ProductAmount newProductAmount) throws InvalidProductAmountException {
        for (ProductAmount prodAmount : this.productAmountList) {
            if (newProductAmount.getProduct().equals(prodAmount.getProduct())) {
                throw new InvalidProductAmountException("The list already contains the item " + newProductAmount.getProductName());
            }
        }
        this.productAmountList.add(newProductAmount);
    }

    /**
     * Tries to insert an item in the list. If the item is already in the
     * list, we increase the amount of the product.
     *
     * @param productAmount
     */
    public final void addProductOrIncreaseAmount(ProductAmount productAmount) {
        try {
            this.addProductAmount(productAmount);
        } catch (InvalidProductAmountException e) {
            // If the list already contains the item, update the quantity of the item
            ProductAmount itemToUpdate = this.getProductAmountByProductName(productAmount.getProduct().getName());
            itemToUpdate.setAmount(itemToUpdate.getAmount() + productAmount.getAmount());
        }
    }

    /**
     * Search and return a product in the productAmountList list. If product is not
     * found return null
     *
     * @param name the name of the product
     * @return the item if found, else null.
     */
    public ProductAmount getProductAmountByProductName(String name) {
        for (ProductAmount item : this.productAmountList) {
            if (item.getProductName().equals(name)) {
                return item;
            }
        }
        return null;
    }

    public boolean isEmpty() {
        return this.productAmountList.isEmpty();
    }
}
