package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.utils.Utils;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Represents a recipe, which contains a name, number of people,
 * a list of ingredients with quantities and some instructions.
 */
public class Recipe extends ProductAmountListOwner {
    private int numberOfPeople;
    private String name;
    private String instructions;

    /**
     * Creates an empty recipe for one person.
     *
     * @param name
     * @throws RecipeCreationException
     */
    public Recipe(String name) throws RecipeCreationException {
        this(name, 1, new ArrayList<>(), "");
    }

    public Recipe(String name, int numberOfPeople, List<ProductAmount> productAmountList, String instructions)
            throws RecipeCreationException {
        super(productAmountList);
        if (numberOfPeople < 0) {
            throw new RecipeCreationException(
                    "The number of people must be greater than or equal 0 but was " + Integer.toString(numberOfPeople));
        }
        if (name.trim().equals("")) {
            throw new RecipeCreationException("Name cannot be empty or contain only whitespaces");

        }
        this.name = name;
        this.numberOfPeople = numberOfPeople;
        this.instructions = instructions;
    }

    /**
     * Constructor used to make testing easier.
     *
     * @param name
     * @param numberOfPeople
     * @throws RecipeCreationException
     */
    public Recipe(String name, int numberOfPeople) throws RecipeCreationException {
        this(name, numberOfPeople, new ArrayList<>(), "");
    }

    /**
     * Constructor used to make testing easier.
     *
     * @param name
     * @param numberOfPeople
     * @param instructions
     * @throws RecipeCreationException
     */
    public Recipe(String name, int numberOfPeople, String instructions) throws RecipeCreationException {
        this(name, numberOfPeople, new ArrayList<>(), instructions);
    }

    public int getNumberOfPeople() {
        return this.numberOfPeople;
    }

    public String getName() {
        return this.name;
    }

    public String getInstructions() {
        return this.instructions;
    }

    /**
     * Gets a list of the products with their amounts
     */
    public List<ProductAmount> getProductAmountList() {
        return new ArrayList<>(this.getUnderlyingProductAmountList());
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Recipe recipe = (Recipe) o;
        return this.numberOfPeople == recipe.numberOfPeople &&
                Utils.listsHaveSameElements(this.getUnderlyingProductAmountList(), recipe.getUnderlyingProductAmountList()) &&
                Objects.equals(this.name, recipe.name) &&
                Objects.equals(this.instructions, recipe.instructions);
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.getUnderlyingProductAmountList(), this.numberOfPeople, this.name, this.instructions);
    }

}
