package be.ac.ulb.infof307.g02.models;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import static be.ac.ulb.infof307.g02.models.Day.NUM_DAYS;

/**
 * Represents the schedule for a shop. It has a list of DaySchedules, each having a list of TimeSlots.
 */
public class Schedule implements Serializable {
    /**
     * Attributes a unique ID to the class to make serialization work. This ID is used
     * to make sure that the deserialized object is of this class.
     */
    private static final long serialVersionUID = 1797473856212607718L;
    /**
     * The DaySchedules that make up the schedule
     */
    private List<DaySchedule> daySchedules;

    public Schedule() {
        this.daySchedules = new ArrayList<>();
        for (Day day : Day.values()) {
            this.daySchedules.add(new DaySchedule(day));
        }
    }

    public Schedule(List<DaySchedule> daySchedules) {
        if (daySchedules.size() != NUM_DAYS) {
            throw new IllegalArgumentException();
        }
        this.daySchedules = new ArrayList<>();
        for (DaySchedule daySchedule : daySchedules) {
            this.daySchedules.add(new DaySchedule(daySchedule));
        }
    }

    public Schedule(Schedule schedule) {
        this(schedule.getDaySchedules());
    }

    /**
     * Adds a TimeSlot to the schedule
     *
     * @param newTimeSlot TimeSlot to be added
     * @param day
     */
    public void addTimeSlot(TimeSlot newTimeSlot, Day day) {
        this.daySchedules.get(day.getIndex()).addTimeSlot(newTimeSlot);
    }

    /**
     * Gets all the TimeSlots for a day
     *
     * @param day
     * @return Copy of the list of TimeSlots for the given day
     */
    public List<TimeSlot> getDayTimeSlots(Day day) {
        return this.daySchedules.get(day.getIndex()).getTimeSlots();
    }

    /**
     * Returns a copy of the time slots of the schedule, organized by day.
     */
    public List<DaySchedule> getDaySchedules() {
        List<DaySchedule> dayScheduleList = new ArrayList<>();
        for (DaySchedule daySchedule : this.daySchedules) {
            dayScheduleList.add(new DaySchedule(daySchedule));
        }
        return dayScheduleList;
    }

    public boolean isEmpty() {
        for (DaySchedule daySchedule : this.daySchedules) {
            if (!daySchedule.isEmpty())
                return false;
        }

        return true;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Schedule schedule = (Schedule) o;
        return Objects.equals(this.daySchedules, schedule.daySchedules);
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.daySchedules);
    }
}
