package be.ac.ulb.infof307.g02.models;

import java.util.*;

/**
 * Represents a shop with a name, a schedule, a position and a stock of articles.
 */
public class Shop {

    private int id;
    private String name;
    private Schedule schedule;
    private Position position;
    private Map<Product, Article> stock;
    // We use a map <Product, Article> to make the search in the stock easier and more readable

    /**
     * Construct the Shop given the ID (from the database).
     *
     * @param id       Unique ID of the shop
     * @param name     Name of the shop
     * @param schedule Schedule of the shop
     * @param position The position (latitude, longitude)
     * @param stock    A list of articles
     */
    public Shop(int id, String name, Schedule schedule, Position position,
                Map<Product, Article> stock) {
        if (id < 0 || name.equals("") || schedule.isEmpty()) {
            throw new IllegalArgumentException();
        }
        this.id = id;
        this.name = name;
        this.schedule = schedule;
        this.position = position;
        this.stock = stock;
    }

    /**
     * Construct the Shop with an empty stock.
     *
     * @param id       ID of the shop
     * @param name     Name of the shop
     * @param schedule Schedule of the shop
     * @param position The position (latitude, longitude)
     */
    public Shop(int id, String name, Schedule schedule, Position position) {
        this(id, name, schedule, position, new HashMap<>());
    }

    /**
     * Construct the Shop without a real ID (indicated by having ID=1). This Shop object is not
     * in the database, and insertion in the database will replace ID with the correct ID.
     *
     * @param name     Name of the shop
     * @param schedule Schedule of the shop
     * @param position The position (latitude, longitude)
     * @param stock    A list of articles
     */
    public Shop(String name, Schedule schedule, Position position,
                Map<Product, Article> stock) {
        if (name.equals("") || schedule.isEmpty()) {
            throw new IllegalArgumentException("The shop must have name and a schedule");
        }
        this.id = -1;
        this.name = name;
        this.schedule = schedule;
        this.position = position;
        this.stock = stock;
    }

    public int getId() {
        return this.id;
    }

    public void setId(int id) {
        if (id < 0) {
            throw new IllegalArgumentException("The ID of a shop must be positive");
        }
        this.id = id;
    }

    public String getName() {
        return this.name;
    }

    public void setName(String name) {
        if (name.equals("")) {
            throw new IllegalArgumentException("The name of shop can't be empty");
        }
        this.name = name;
    }

    public Position getPosition() {
        return new Position(this.position);
    }

    public Schedule getSchedule() {
        return new Schedule(this.schedule);
    }

    public void setSchedule(Schedule schedule) {
        if (schedule.isEmpty()) {
            throw new IllegalArgumentException();
        }
        this.schedule = schedule;
    }

    public Map<Product, Article> getStock() {
        return new HashMap<>(this.stock);
    }

    public Set<Product> getProducts() {
        return this.stock.keySet();
    }


    /**
     * @param productsList
     * @return true if all products of the list are in the shop's stock, else false
     */
    public boolean sellAllProducts(List<Product> productsList) {
        for (Product product : productsList) {
            if (!this.stock.containsKey(product)) {
                return false;
            }
        }
        return true;
    }

    /**
     * @param groceryList
     * @return Return the total price of a grocery list
     */
    public double getPriceOfList(GroceryList groceryList) {
        double result = 0;
        for (ProductAmount item : groceryList.getProductAmountList()) {
            Article article = this.stock.get(item.getProduct());
            result += article.getPrice() * item.getAmount();
        }
        return result;
    }

    public double getProductPrice(Product product) {
        return this.stock.get(product).getPrice();
    }

    public List<DaySchedule> getDaySchedules() {
        return this.schedule.getDaySchedules();
    }

    /**
     * Adds an article to the stock
     *
     * @param article
     */
    public void addToStock(Article article) {
        this.stock.put(article.getProduct(), article);
    }

    public boolean hasProduct(Product product) {
        return this.stock.containsKey(product);
    }

    @Override
    public String toString() {
        return this.name;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Shop shop = (Shop) o;
        return this.id == shop.id &&
                Objects.equals(this.name, shop.name) &&
                Objects.equals(this.schedule, shop.schedule) &&
                Objects.equals(this.position, shop.position) &&
                Objects.equals(this.stock, shop.stock);
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.id, this.name, this.schedule, this.position, this.stock);
    }
}
