package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.utils.DistanceCalculator;

import java.util.ArrayList;
import java.util.List;

/**
 * Handles all the filters that can be applied on an set of shops.
 */
public class ShopFilterer {
    private List<Shop> shops;
    private List<Shop> filteredShops;
    private Position userPosition;

    public ShopFilterer(List<Shop> shops, Position userPosition) {
        this.userPosition = userPosition;
        this.shops = shops;
        this.filteredShops = new ArrayList<>();
    }

    /**
     * Filters shops
     *
     * @param filterData Contains all the information on which the shops are filtered
     */
    public List<Shop> filterShops(FilterData filterData) {
        this.filteredShops = new ArrayList<>(this.shops);
        if (filterData.getMaxDistance() != 0) {
            this.filterShopByDistance(filterData.getMaxDistance());
        }
        if (!(filterData.getName().equals(""))) {
            this.filterShopsByName(filterData.getName());
        }
        for (Day day : filterData.getOpeningDays()) {
            filterShopsByDay(day);
        }
        if (filterData.getOpeningTime() != null) {
            this.filterShopsByTime(filterData.getOpeningTime());
        }
        this.filterShopsByProducts(filterData.getProductList());
        return this.filteredShops;
    }


    /**
     * Removes the shops that are farther than a given distance
     *
     * @param distance The maximum distance
     */
    private void filterShopByDistance(int distance) {
        for (Shop shop : this.shops) {
            Position shopPosition = shop.getPosition();
            if (DistanceCalculator.getLinearDistance(shopPosition, this.userPosition) > distance) {
                this.filteredShops.remove(shop);
            }
        }
    }

    /**
     * Removes the shops if their name don't match a given name (non case sensitive)
     *
     * @param name Name on which the shops are filtered
     */
    private void filterShopsByName(String name) {
        for (Shop shop : this.shops) {
            String shopName = shop.getName();
            if (!(shopName.toLowerCase().contains(name.toLowerCase()))) {
                this.filteredShops.remove(shop);
            }
        }
    }

    /**
     * Removes the shops that aren't open on a given day
     *
     * @param day The day on which the shops must be open
     */
    private void filterShopsByDay(Day day) {
        for (Shop shop : this.shops) {
            Schedule schedule = shop.getSchedule();
            List<TimeSlot> timeSlotList = schedule.getDayTimeSlots(day);
            if (timeSlotList.isEmpty()) {
                this.filteredShops.remove(shop);
            }
        }
    }

    /**
     * Removes the shops that aren't open at the given time
     *
     * @param time The time at which the shops must be open
     */
    private void filterShopsByTime(Time time) {
        for (Shop shop : this.shops) {
            boolean isOpenAt = false;
            Schedule schedule = shop.getSchedule();
            for (DaySchedule daySchedule : schedule.getDaySchedules()) {
                for (TimeSlot timeSlot : daySchedule.getTimeSlots()) {
                    if (timeSlot.contains(time)) {
                        isOpenAt = true;
                    }
                }
            }
            if (!isOpenAt) {
                this.filteredShops.remove(shop);
            }
        }
    }

    /**
     * Removes the shops that don't contains all the productList
     *
     * @param productList The list of products that need to be in the shop
     */
    private void filterShopsByProducts(List<Product> productList) {
        try {
            for (Shop shop : this.shops) {
                for (Product product : productList) {
                    if (!shop.hasProduct(product)) {
                        this.filteredShops.remove(shop);
                        break;
                    }
                }
            }
        } catch (ClassCastException e) { // if the product doesn't exist no shop has it
            this.filteredShops.clear();
        }
    }

}
