package be.ac.ulb.infof307.g02.models;

import java.io.Serializable;
import java.util.Objects;

/**
 * Represents a period during which a store can be opened. A TimeSlot
 * is composed of 2 Times : Open time and Close time
 * minute
 */
public class TimeSlot implements Serializable {

    /**
     * Attributes a unique ID to the class to make serialization work. This ID is used
     * to make sure that the deserialized object is of this class.
     */
    private static final long serialVersionUID = 7474549248268499898L;
    private Time openTime, closeTime;

    public TimeSlot(Time openTime, Time closeTime) {
        if (openTime.isAfter(closeTime)) {
            throw new IllegalArgumentException("The open time must be before the close time");
        }
        this.openTime = openTime;
        this.closeTime = closeTime;
    }

    public Time getOpenTime() {
        return new Time(this.openTime);
    }

    public Time getCloseTime() {
        return new Time(this.closeTime);
    }

    /**
     * Checks if a given time is in the TimeSlot
     *
     * @param time
     * @return true if time is inside the TimeSlot, false if it isn't.
     */
    public boolean contains(Time time) {
        boolean isAfterOpeningHour = time.getHour() > this.openTime.getHour() || (time.getHour() == this.openTime.getHour()
                && time.getMinute() >= this.openTime.getMinute());
        boolean isBeforeCloseHour = time.getHour() < this.closeTime.getHour() || (time.getHour() == this.closeTime.getHour()
                && time.getMinute() <= this.closeTime.getMinute());
        return isAfterOpeningHour && isBeforeCloseHour;
    }

    /**
     * Checks it the TimeSlot overlaps another one.
     *
     * @param other Another TimeSlot
     * @return True if they overlap, false otherwise
     */
    public boolean isTimeSlotOverlapping(TimeSlot other) {
        if (this.contains(other.getOpenTime()))
            return true;
        if (this.contains(other.getCloseTime()))
            return true;
        if (other.contains(this.openTime))
            return true;
        return other.contains(this.closeTime);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        TimeSlot timeSlot = (TimeSlot) o;
        return Objects.equals(this.openTime, timeSlot.openTime) &&
                Objects.equals(this.closeTime, timeSlot.closeTime);
    }

    @Override
    public int hashCode() {

        return Objects.hash(this.openTime, this.closeTime);
    }

    @Override
    public String toString() {
        String openTimeString = this.twoDigits(this.openTime.getHour()) + ":" + this.twoDigits(this.openTime.getMinute());
        String closeTimeString = this.twoDigits(this.closeTime.getHour()) + ":" + this.twoDigits(this.closeTime.getMinute());
        return openTimeString + " - " + closeTimeString;
    }

    /**
     * Returns a two-digit version of the time.
     * Example: 6 -> "06", 23 -> "23"
     *
     * @param time
     * @return a two-digit version of the time.
     */
    private String twoDigits(int time) {
        return String.format("%02d", time);
    }

}
