package be.ac.ulb.infof307.g02.models;

import be.ac.ulb.infof307.g02.exceptions.model.InvalidPasswordException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.utils.HashPassword;
import be.ac.ulb.infof307.g02.utils.Utils;

/**
 * Represents a user. Stores their user name and their hashed password. We trim the
 * whitespace of the user's name. The trimmed user name and password must
 * respect validity conditions (see isPasswordValid).
 */
public class User {
    private static final int MINIMUM_PASSWORD_LENGTH = 8;
    /**
     * the userName
     */
    private String userName;
    private String hashedPassword;

    /**
     * Creates a User, trimming the whitespace at the start and end of the user
     * name. If the trimmed user name or the password is empty, throws an
     * IllegalArgumentException.
     *
     * @param userName
     * @param password
     * @throws UserCreationException
     */
    public User(String userName, String password) throws UserCreationException {
        String cleanedUserName = userName.trim();
        if (cleanedUserName.equals("") || !User.isPasswordValid(password)) {
            throw new UserCreationException();
        }
        this.userName = cleanedUserName;
        this.hashedPassword = HashPassword.hash(password);
    }

    /**
     * Creates a user using their name and the hashed password.
     *
     * @param userName
     * @param hashedPassword
     */
    public User(String userName, byte[] hashedPassword) {
        this.userName = userName;
        this.hashedPassword = new String(hashedPassword);
    }

    public User(User user) {
        this(user.getUserName(), user.getHashedPasswordBytes());
    }

    /**
     * Returns whether the password is valid. It must:
     * - Have at least 8 characters
     * - Have at least two types of characters (number, lower case, upper case, special character)
     *
     * @param password
     * @return whether the password is valid
     */
    public static boolean isPasswordValid(String password) {
        if (password.length() < MINIMUM_PASSWORD_LENGTH) {
            return false;
        }

        int numCharacterGroups = 0;
        String[] patterns = {
                ".*[0-9]+.*", // Has a number
                ".*[a-z]+.*", // Has a lower case letter
                ".*[A-Z]+.*", // Has an upper case letter
        };

        for (String pattern : patterns) {
            if (password.matches(pattern)) {
                numCharacterGroups++;
            }
        }

        if (Utils.hasSpecialChar(password)) {
            numCharacterGroups++;
        }

        return numCharacterGroups >= 2;
    }

    public String getUserName() {
        return this.userName;
    }

    /**
     * Sets the user name, trimming the whitespace at the start and the end. Throws
     * an IllegalArgumentException is the trimmed user name is empty.
     *
     * @param newUserName
     */
    public void setUserName(String newUserName) {
        String cleanedNewUserName = newUserName.trim();
        if (cleanedNewUserName.equals("")) {
            throw new IllegalArgumentException();
        }
        this.userName = cleanedNewUserName;
    }

    /**
     * Returns True if the given password matches the user's password.
     *
     * @param password
     */
    public boolean passwordMatches(String password) {
        return this.hashedPassword.equals(HashPassword.hash(password));
    }

    public boolean passwordMatches(User newUser) {
        return this.hashedPassword.equals(newUser.getHashedPassword());
    }

    public String getHashedPassword() {
        return this.hashedPassword;
    }

    public void setHashedPassword(String hashedPassword) {
        this.hashedPassword = hashedPassword;
    }

    public byte[] getHashedPasswordBytes() {
        return this.hashedPassword.getBytes();
    }

    /**
     * Updates the password. Throws an InvalidPasswordException if the new password is
     * invalid.
     *
     * @param newPassword
     */
    public void setPassword(String newPassword) throws InvalidPasswordException {
        if (!User.isPasswordValid(newPassword)) {
            throw new InvalidPasswordException();
        }
        this.hashedPassword = HashPassword.hash(newPassword);
    }
}
