package be.ac.ulb.infof307.g02.utils;

import javax.crypto.*;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

/**
 * Shows how to securely perform AES encryption in GCM mode, with 256 bits key size.
 */
public class AES {
    public static final String ALGORITHM_NAME = "AES";

    /**
     * The AES key size
     * We do not use 256 bit, because it makes all tests fail on Gitlab.
     * For it to we, JCE Unlimited Strength needs to be installed
     */
    private static final int AES_KEY_SIZE = 128;

    /**
     * Strings that describes all algorithm details (algorithm name + mode + padding)
     */
    private static final String ALGO_TRANSFORMATION_STRING = "AES/CBC/PKCS5Padding";

    public static SecretKey generateAESKey() {
        try {
            KeyGenerator keyGenerator = KeyGenerator.getInstance(ALGORITHM_NAME);
            keyGenerator.init(AES_KEY_SIZE);
            return keyGenerator.generateKey();
        } catch (NoSuchAlgorithmException e) {
            // Will never happen
        }
        return null;
    }

    /**
     * Returns the SecretKey representing from the AES key bytes
     *
     * @param aesKey
     */
    public static SecretKey createKey(byte[] aesKey) {
        return new SecretKeySpec(aesKey, AES.ALGORITHM_NAME);
    }

    /**
     * Build the initialization vector, this vector is used to prevent for bruteforce (Rainbow) attack,
     * which is required for our AES encryption scheme. We always use the same vector.
     */
    private static IvParameterSpec getIvParameterSpec() {
        // build the initialization vector.  This example is all zeros, but it
        // could be any value or generated using a random number generator.
        byte[] iv = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
        return new IvParameterSpec(iv);
    }


    /**
     * Encrypt a string message using AES algorithm
     *
     * @param message
     * @param aesKey
     */
    public static byte[] encrypt(String message, SecretKey aesKey) {
        return encrypt(message, aesKey, getIvParameterSpec());
    }

    /**
     * Encrypt a byte array message using AES algorithm
     *
     * @param message
     * @param aesKey
     */
    public static byte[] encrypt(byte[] message, SecretKey aesKey) {
        return encrypt(message, aesKey, getIvParameterSpec());
    }

    /**
     * Encrypt a string message using AES algorithm and IvParameters
     *
     * @param message
     * @param aesKey
     * @param ivParameterSpec
     */
    private static byte[] encrypt(String message, SecretKey aesKey, IvParameterSpec ivParameterSpec) {
        return encrypt(message.getBytes(), aesKey, ivParameterSpec);
    }

    /**
     * Encrypt a byte array message using AES algorithm and IvParameterSpec
     *
     * @param message
     * @param aesKey
     * @param ivParameterSpec
     */
    private static byte[] encrypt(byte[] message, SecretKey aesKey, IvParameterSpec ivParameterSpec) {
        try {
            Cipher cipher = Cipher.getInstance(ALGO_TRANSFORMATION_STRING);
            cipher.init(Cipher.ENCRYPT_MODE, aesKey, ivParameterSpec);
            byte[] cipherTextBytes = cipher.doFinal(message);
            return cipherTextBytes;
        } catch (NoSuchAlgorithmException | NoSuchPaddingException | BadPaddingException | InvalidKeyException | IllegalBlockSizeException | InvalidAlgorithmParameterException e) {
            // Will never appear
        }
        return null;
    }


    /**
     * Decrypt a byte array message using AES algorithm
     *
     * @param encryptedMessage
     * @param aesKey
     */
    public static byte[] decrypt(byte[] encryptedMessage, SecretKey aesKey) {
        return decrypt(encryptedMessage, aesKey, getIvParameterSpec());
    }

    /**
     * Decrypt a byte array message using AES algorithm and IvParameterSpec
     *
     * @param encryptedMessage
     * @param aesKey
     * @param ivParameterSpec
     */
    private static byte[] decrypt(byte[] encryptedMessage, SecretKey aesKey, IvParameterSpec ivParameterSpec) {
        try {
            Cipher cipher = Cipher.getInstance(ALGO_TRANSFORMATION_STRING);
            cipher.init(Cipher.DECRYPT_MODE, aesKey, ivParameterSpec);
            byte[] plainTextBytes = cipher.doFinal(encryptedMessage);
            return plainTextBytes;
        } catch (NoSuchPaddingException | NoSuchAlgorithmException | BadPaddingException | InvalidKeyException | IllegalBlockSizeException | InvalidAlgorithmParameterException e) {
            int a = 5;
        }
        return null;
    }
}