package be.ac.ulb.infof307.g02.utils;

import be.ac.ulb.infof307.g02.controllers.ArticleImportationObserver;
import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.database.ShopDB;
import be.ac.ulb.infof307.g02.exceptions.database.ArticleAlreadyInShopException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.ProductAlreadyExistsException;
import be.ac.ulb.infof307.g02.exceptions.import_export.CSVNotConsistentException;
import be.ac.ulb.infof307.g02.exceptions.import_export.HeadersNotConsistentException;
import be.ac.ulb.infof307.g02.exceptions.import_export.NoFileSelectedException;
import be.ac.ulb.infof307.g02.exceptions.import_export.RecordNotConsistentException;
import be.ac.ulb.infof307.g02.models.Article;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.Shop;
import be.ac.ulb.infof307.g02.views.FileSelectorView;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

// Correct Headers
enum Headers {
    Name, Unit, Price, Lipids, Proteins, Carbohydrates;

    public static String toCSV() {
        StringBuilder stringBuilder = new StringBuilder();
        for (Headers header : Headers.values()) {
            stringBuilder.append(header.name()).append(",");
        }
        // remove the last "," and replace it with a \n
        stringBuilder = new StringBuilder(stringBuilder.substring(0, stringBuilder.length() - 1) + "\n");

        return stringBuilder.toString();
    }
}

/**
 * Class handling the import of articles from CSV file.
 */
public class ArticleImporterFromFile {

    /**
     * The observer is informed when a new article is imported
     */
    private ArticleImportationObserver observer;

    public ArticleImporterFromFile(ArticleImportationObserver observer) {
        this.observer = observer;
    }

    /**
     * Displays a file chooser to select a CSV file to import and process the importation.
     *
     * @param targetShop Shop in which the articles are imported.
     * @throws DatabaseConnectionException
     * @throws CSVNotConsistentException
     */
    public void importFrom(Shop targetShop) throws DatabaseConnectionException, CSVNotConsistentException, NoFileSelectedException, ArticleAlreadyInShopException {

        FileSelectorView fileSelectorView = new FileSelectorView();
        File file = fileSelectorView.getSelectedFileToOpen();

        if (file != null) {
            if (!file.getName().contains(".csv")) {
                file = new File(file.getAbsolutePath() + ".csv");
            }

            this.importData(file, targetShop);
        } else {
            throw new NoFileSelectedException();
        }
    }

    /**
     * Checks the consistency of a CSV file and import all the records if it is consistent.
     *
     * @param file       File to import
     * @param targetShop Shop in which the articles are imported.
     * @throws CSVNotConsistentException
     * @throws DatabaseConnectionException
     */
    private void importData(File file, Shop targetShop) throws CSVNotConsistentException, DatabaseConnectionException, ArticleAlreadyInShopException {
        try {
            Reader in = new FileReader(file);
            CSVParser csvFileParser = CSVFormat.RFC4180.withFirstRecordAsHeader().parse(in);
            List<CSVRecord> records = csvFileParser.getRecords();
            List<Article> articlesToImport = checkFileConsistency(records);
            this.insertToDatabase(articlesToImport, targetShop);
        } catch (IOException e) {
            // Don't tell that to the user
        } catch (RecordNotConsistentException | HeadersNotConsistentException e) {
            throw new CSVNotConsistentException(e.getMessage());
        }
    }

    /**
     * Checks if the CSV file is consistent (headers + records).
     *
     * @param records Records of the imported CSV file
     * @return A list containing all the article and articles to import
     * @throws HeadersNotConsistentException
     * @throws RecordNotConsistentException
     */
    private static List<Article> checkFileConsistency(List<CSVRecord> records) throws HeadersNotConsistentException, RecordNotConsistentException {
        checkFileHeadersConsistency(records.get(0));
        List<Article> articlesToImport = checkContentConsistency(records);
        return articlesToImport;
    }

    /**
     * Inserts all the articles into the shop, by adding the products and articles in the database
     *
     * @param articlesToImport all the articles to import
     * @param shop             Shop in which the articles are imported.
     * @throws DatabaseConnectionException
     */
    private void insertToDatabase(List<Article> articlesToImport, Shop shop) throws DatabaseConnectionException,
            ArticleAlreadyInShopException {
        ProductDB productDB = new ProductDB();
        ShopDB shopDB = new ShopDB();
        int shopID = shop.getId();
        for (Article article : articlesToImport) {

            try {
                productDB.addProduct(article.getProduct());
                shopDB.insertArticleInShop(shopID, article);
            } catch (ProductAlreadyExistsException e) {
                shopDB.insertArticleInShop(shopID, article);
                this.observer.articleImported(article);

            } catch (ArticleAlreadyInShopException e) {
                // Never will occur because the product is new product in the ShopHasProductTable
                // and in the Product Table
            }
        }
    }

    /**
     * Checks if the headers are equal to Headers Enum(Name, Unit, Price)
     *
     * @param headerRecords Header of the imported CSV file
     * @throws HeadersNotConsistentException
     */
    private static void checkFileHeadersConsistency(CSVRecord headerRecords) throws HeadersNotConsistentException {
        if (!(headerRecords.isConsistent() && headerRecords.isMapped(Headers.Name.name())
                && headerRecords.isMapped(Headers.Unit.name())
                && headerRecords.isMapped(Headers.Price.name()))) {
            throw new HeadersNotConsistentException("Headers are not consistent");
        }
    }

    /**
     * Checks if all the records, except headers, of the CSV file are consistent.
     *
     * @param records Records of the imported CSV file
     * @return A list containing all the articles and articles to import
     * @throws RecordNotConsistentException
     */
    private static List<Article> checkContentConsistency(List<CSVRecord> records) throws RecordNotConsistentException {
        List<Article> articlesToImport = new ArrayList<>();
        for (CSVRecord record : records) {
            String name = record.get(Headers.Name).trim();
            Product.Unit unit = Product.Unit.create(record.get(Headers.Unit).trim());
            String price = record.get(Headers.Price).trim();
            String lipids = record.get(Headers.Lipids).trim();
            String proteins = record.get(Headers.Proteins).trim();
            String carbohydrates = record.get(Headers.Carbohydrates).trim();

            Product productToInsert;
            Article articleToInsert;
            try {
                productToInsert = new Product(
                        name,
                        unit,
                        Double.parseDouble(lipids),
                        Double.parseDouble(proteins),
                        Double.parseDouble(carbohydrates)
                );

                articleToInsert = new Article(productToInsert, Double.parseDouble(price));
                articlesToImport.add(articleToInsert);

            } catch (IllegalArgumentException e) {
                throw new RecordNotConsistentException(String.format(" %s %s %s", name, unit, price));
            }
        }
        return articlesToImport;
    }

}
