package be.ac.ulb.infof307.g02.utils;

import be.ac.ulb.infof307.g02.database.DatabaseConnection;
import be.ac.ulb.infof307.g02.database.KeyStorer;
import be.ac.ulb.infof307.g02.exceptions.crypto.NoKeyPairPresentException;
import be.ac.ulb.infof307.g02.exceptions.crypto.NoSuchKeysFoundForUserException;
import be.ac.ulb.infof307.g02.exceptions.crypto.UserAlreadyHasKeysException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.models.User;

import javax.crypto.SecretKey;
import java.security.KeyPair;
import java.security.PublicKey;

import static be.ac.ulb.infof307.g02.utils.Utils.toByteArray;


/**
 * Responsible for encrypting and decrypting data using RSA and AES.
 * Can also (re-)generate and store keys.
 */
public class Crypto {
    private final KeyStorer keyStorer;

    private KeyPair keyPair;
    private SecretKey aesKey;

    /**
     * Initializes cryptography for the given user, fetching their keys from the database and the keystore file.
     * If no keys exists for the user, we generate and store them.
     *
     * @param user
     * @param databaseConnection
     * @throws DatabaseException
     */
    public Crypto(User user, DatabaseConnection databaseConnection) throws DatabaseException {
        this.keyStorer = new KeyStorer(databaseConnection);

        try {
            this.keyPair = this.keyStorer.getKeyPair(user);
            this.aesKey = getAesKey(user);
        } catch (NoKeyPairPresentException | NoSuchKeysFoundForUserException e) {
            this.keyPair = RSA.generateKeyPair();
            this.aesKey = AES.generateAESKey();
            byte[] encryptedAesKey = this.getEncryptedAESKey();
            try {
                this.keyStorer.storeAllKeys(this.keyPair, encryptedAesKey, user);
            } catch (UserAlreadyHasKeysException e1) {
                // Will never happen
            }
        }
    }

    /**
     * Initializes cryptography with the given RSA keypair and generates a new AES key.
     * This constructor is useful for testing the class.
     *
     * @param keyPair
     */
    public Crypto(KeyPair keyPair) {
        this.keyStorer = null;
        this.keyPair = keyPair;
        this.aesKey = AES.generateAESKey();
    }

    /**
     * Returns the (unencrypted) AES key
     *
     * @param user
     * @throws DatabaseException
     * @throws NoSuchKeysFoundForUserException
     */
    private SecretKey getAesKey(User user) throws DatabaseException, NoSuchKeysFoundForUserException {
        byte[] encryptedAESKey = this.keyStorer.getEncryptedAESKey(user);
        byte[] decryptedAESKey = RSA.decrypt(encryptedAESKey, this.keyPair.getPrivate());
        return AES.createKey(decryptedAESKey);
    }


    /**
     * Decrypt text using AES.
     *
     * @param text encrypted text
     * @return plain text
     */
    public byte[] decrypt(byte[] text) {
        return AES.decrypt(text, this.aesKey);
    }

    /**
     * Encrypt the plain text using AES.
     *
     * @param text original plain text
     * @return Encrypted text
     */
    public byte[] encrypt(String text) {
        return AES.encrypt(text, this.aesKey);
    }

    /**
     * Encrypt int value using AES.
     *
     * @param value
     * @return
     */
    public byte[] encrypt(int value) {
        return AES.encrypt(toByteArray(value), this.aesKey);
    }

    /**
     * Encrypt double value using AES.
     *
     * @param value
     */
    public byte[] encrypt(double value) {
        return AES.encrypt(toByteArray(value), this.aesKey);
    }

    /**
     * Regenerate new RSA keys, re-encrypt the AES key using the new RSA Public Key then store all the keys.
     *
     * @param oldUser
     * @param updatedUser
     */
    public void regenerateKeys(User oldUser, User updatedUser) {
        KeyPair newKeyPair = RSA.generateKeyPair();
        byte[] encryptedAesKey = this.getEncryptedAESKey(this.aesKey.getEncoded(), newKeyPair.getPublic());
        this.keyStorer.updateAllKeys(newKeyPair, encryptedAesKey, oldUser, updatedUser);
    }


    /**
     * Return the RSA-encrypted AES key
     *
     * @return
     */
    private byte[] getEncryptedAESKey() {
        return RSA.encrypt(this.aesKey.getEncoded(), this.keyPair.getPublic());
    }

    /**
     * Return AES key after RSA encryption with the given RSA public key
     *
     * @param aesKey
     * @param publicKey
     */
    private byte[] getEncryptedAESKey(byte[] aesKey, PublicKey publicKey) {
        return RSA.encrypt(aesKey, publicKey);
    }


    public KeyPair getKeyPair() {
        return this.keyPair;
    }

    public SecretKey getAesKey() {
        return this.aesKey;
    }
}
