package be.ac.ulb.infof307.g02.utils;

import be.ac.ulb.infof307.g02.database.DatabaseConnection;
import be.ac.ulb.infof307.g02.database.DatabaseConnectionSingleton;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.models.User;

import java.security.KeyPair;


/**
 * Ensure that there is only one instance of crypto at a time.
 */
public class CryptoSingleton {
    private static Crypto crypto = null;

    /**
     * Returns the current Crypto instance
     */
    public static Crypto getCrypto() {
        return crypto;
    }

    /**
     * Creates a new Crypto instance when user logs in, which fetches all the required key information
     *
     * @param user
     */
    public static void setupCryptoForUser(User user) throws DatabaseException, DatabaseConnectionException {
        setupCryptoForUser(user, DatabaseConnectionSingleton.getDatabaseConnection());
    }

    /**
     * Creates a new Crypto instance when user logs in, which fetches all the required key information
     *
     * @param user
     * @param databaseConnection
     */
    public static void setupCryptoForUser(User user, DatabaseConnection databaseConnection) throws DatabaseException {
        synchronized (CryptoSingleton.class) {
            crypto = new Crypto(user, databaseConnection);
        }
    }

    /**
     * Resets the Crypto instance (when the user logs out)
     */
    public static void resetCrypto() {
        synchronized (CryptoSingleton.class) {
            crypto = null;
        }
    }

    /**
     * Creates a new Crypto instance used in tests
     */
    public static void setup(KeyPair keyPair) {
        synchronized (CryptoSingleton.class) {
            crypto = new Crypto(keyPair);
        }
    }

    /**
     * Regenerates the user's keys, stores the new information and creates a new Crypto instance
     * with the new keys.
     *
     * @param oldUser
     * @param updatedUser
     * @throws DatabaseException
     * @throws DatabaseConnectionException
     */
    public static void updateUserKeys(User oldUser, User updatedUser) throws DatabaseException, DatabaseConnectionException {
        updateUserKeys(oldUser, updatedUser, DatabaseConnectionSingleton.getDatabaseConnection());
    }

    /**
     * Regenerates the user's keys, stores the new information and creates a new Crypto instance
     * with the new keys.
     *
     * @param oldUser
     * @param updatedUser
     * @param databaseConnection
     * @throws DatabaseException
     */
    public static void updateUserKeys(User oldUser, User updatedUser, DatabaseConnection databaseConnection) throws DatabaseException {
        crypto.regenerateKeys(oldUser, updatedUser);
        setupCryptoForUser(updatedUser, databaseConnection);
    }
}
