package be.ac.ulb.infof307.g02.utils;

import be.ac.ulb.infof307.g02.models.Position;
import com.graphhopper.directions.api.client.ApiException;
import com.graphhopper.directions.api.client.api.MatrixApi;
import com.graphhopper.directions.api.client.model.MatrixResponse;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * This class is used to compute the distance between two points, passing by the road
 */
public class DistanceCalculator {
    /**
     * API Key for the GraphHopper API
     */
    private static final String API_KEY = "fc763912-1cbd-4970-8c5b-4d330929599a";
    /**
     * The vehicle used during the routing.
     */
    private static final String VEHICLE = "car";

    /**
     * Radius of the earth in Km
     */
    private static final int EARTH_RADIUS = 6371;

    private static final int KM_FACTOR = 1000;

    /**
     * Calculates the real route distance between two points in latitude and longitude using
     * GraphHopperAPI
     *
     * @param position1
     * @param position2
     * @return Distance (in km)
     */
    public static double getRealDistance(Position position1, Position position2) {
        MatrixApi api = new MatrixApi();

        List<String> allPoints = convertPositionsToStringList(position1, position2);
        List<String> requiredFields = Arrays.asList("distances");
        double distance;
        try {
            MatrixResponse response = api.matrixGet(API_KEY, allPoints, null, null, requiredFields, VEHICLE);
            distance = getRouteDistanceInKm(response.getDistances()); // cause distances are in meters
        } catch (ApiException e) {
            // If we can't reach the API, we use the straight line distance
            distance = getLinearDistance(position1, position2);
        }
        return distance;
    }

    private static double getRouteDistanceInKm(List<List<BigDecimal>> distances) {
        // We get the distance of the best route
        BigDecimal routeDistance = distances.get(0).get(1);
        return routeDistance.doubleValue() / KM_FACTOR;
    }

    /**
     * Converts each position to a String and returns a list with both strings
     *
     * @param position1
     * @param position2
     * @return list of string, each string represents a position
     */
    private static List<String> convertPositionsToStringList(Position position1, Position position2) {
        String latitude1 = String.valueOf(position1.getLatitude());
        String latitude2 = String.valueOf(position2.getLatitude());
        String longitude1 = String.valueOf(position1.getLongitude());
        String longitude2 = String.valueOf(position2.getLongitude());
        List<String> allPoints = new ArrayList<>();
        allPoints.add(latitude1 + "," + longitude1);
        allPoints.add(latitude2 + "," + longitude2);
        return allPoints;
    }

    /**
     * Calculates distance between two points in latitude and longitude using the
     * Haversine method.
     *
     * @param position1
     * @param position2
     * @return distance (in km)
     */
    public static double getLinearDistance(Position position1, Position position2) {
        double lat1 = position1.getLatitude();
        double lat2 = position2.getLatitude();

        double lon1 = position1.getLongitude();
        double lon2 = position2.getLongitude();


        double latDistance = Math.toRadians(lat2 - lat1);
        double lonDistance = Math.toRadians(lon2 - lon1);
        double a = Math.sin(latDistance / 2) * Math.sin(latDistance / 2) + Math.cos(Math.toRadians(lat1))
                * Math.cos(Math.toRadians(lat2)) * Math.sin(lonDistance / 2) * Math.sin(lonDistance / 2);
        double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
        double distance = EARTH_RADIUS * c;

        return distance;
    }
}
