package be.ac.ulb.infof307.g02.utils;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import java.security.*;

/**
 * Performs RSA encryption and decryption (in ECB mode) and generates RSA keypairs.
 */
public class RSA {
    /**
     * The size of the RSA keys.
     */
    private static final int RSA_KEY_LENGTH = 2048;
    private static final String ALGORITHM_NAME = "RSA";

    /**
     * The padding scheme used when encrypting with RSA, required to ensure good security
     */
    private static final String PADDING_SCHEME = "OAEPWITHSHA-512ANDMGF1PADDING";

    /**
     * The RSA mde of operation
     */
    private static final String MODE_OF_OPERATION = "ECB"; // This essentially means none behind the scene

    /**
     * String containing all the information for enryption (algorithm name, mode of operation, padding scheme)
     */
    private static final String ALGORITHM_STRING = ALGORITHM_NAME + "/" + MODE_OF_OPERATION + "/" + PADDING_SCHEME;

    public static KeyPair generateKeyPair() {
        try {
            KeyPairGenerator rsaKeyGen = KeyPairGenerator.getInstance(ALGORITHM_NAME);
            rsaKeyGen.initialize(RSA_KEY_LENGTH);
            return rsaKeyGen.generateKeyPair();
        } catch (NoSuchAlgorithmException e) {
            // Will never happen
        }
        return null;
    }

    /**
     * Encrypt a byte array using RSA algorithm
     *
     * @param bytes
     * @param publicKey
     */
    public static byte[] encrypt(byte[] bytes, PublicKey publicKey) {
        byte[] cipherTextArray = null;
        try {
            Cipher cipher = Cipher.getInstance(ALGORITHM_STRING);
            cipher.init(Cipher.ENCRYPT_MODE, publicKey);
            cipherTextArray = cipher.doFinal(bytes);
        } catch (NoSuchAlgorithmException | NoSuchPaddingException | IllegalBlockSizeException | InvalidKeyException | BadPaddingException e) {
            // Will never happen
        }
        return cipherTextArray;
    }


    /**
     * Decrypt a byte array using RSA algorithm
     *
     * @param encryptedMessage
     * @param privateKey
     */
    public static byte[] decrypt(byte[] encryptedMessage, PrivateKey privateKey) {
        byte[] plainText = null;
        try {
            Cipher cipher = Cipher.getInstance(ALGORITHM_STRING);
            cipher.init(Cipher.DECRYPT_MODE, privateKey);
            plainText = cipher.doFinal(encryptedMessage);
        } catch (NoSuchAlgorithmException | NoSuchPaddingException | IllegalBlockSizeException | InvalidKeyException | BadPaddingException e) {
            // Will never happen

        }
        return plainText;
    }
}