package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.controllers.CreateProductController;
import be.ac.ulb.infof307.g02.models.Product;
import javafx.beans.binding.Bindings;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.*;
import javafx.stage.Stage;
import javafx.stage.WindowEvent;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Allows the user to create a new product
 */
public class CreateProductView extends ViewHandlingDBExceptions {

    /**
     * The fxml resource for the view
     */
    private static final String ADD_PRODUCT_FXML = "createProduct.fxml";
    private static final int WIDTH_WINDOW = 500;
    private static final int HEIGHT_WINDOW = 450;
    private static final int MAX_SPINNER_VALUE = 100;
    private static final int MIN_SPINNER_VALUE = 0;
    private static final int INITIAL_VALUE = 0;

    @FXML
    private Button addProductButton;
    @FXML
    private TextField productNameField;
    @FXML
    private ComboBox<String> unitField;
    @FXML
    private Label errorLabel;
    @FXML
    private Spinner<Double> lipidsSpinner;
    @FXML
    private Spinner<Double> proteinsSpinner;
    @FXML
    private Spinner<Double> carbohydratesSpinner;

    private CreateProductController controller;


    /**
     * Constructor: initialize the Window, disable the "Add Product" button if all the required fields are empty
     */

    public CreateProductView(CreateProductController controller) {
        this.controller = controller;
        try {
            FXMLLoader loader = new FXMLLoader(getClass().getResource(Constants.FXML_FOLDER_PATH + ADD_PRODUCT_FXML));
            loader.setController(this);
            Parent root = loader.load();
            Stage stage = new Stage();
            stage.setTitle("Create Product");
            Scene addProductScene = new Scene(root, WIDTH_WINDOW, HEIGHT_WINDOW);
            stage.setScene(addProductScene);
            addProductScene.getStylesheets().add(Constants.CSS_FILE_PATH);
            stage.setResizable(false);
            stage.setAlwaysOnTop(true);
            stage.show();
            stage.setOnCloseRequest(event -> this.controller.notifyObserverWindowClosed());
            this.setupUI();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    /**
     * Sets up the user interface
     */
    private void setupUI() {
        this.addProductButton.disableProperty().bind(Bindings.isEmpty(productNameField.textProperty()));

        List<String> list = new ArrayList<>();
        for (Product.Unit unit : Product.Unit.values()) {
            list.add(unit.getUnitName());
        }
        ObservableList<String> items = FXCollections.observableList(list);
        this.unitField.getItems().clear();
        this.unitField.setItems(items);

        setupSpinner(this.lipidsSpinner);
        setupSpinner(this.proteinsSpinner);
        setupSpinner(this.carbohydratesSpinner);
    }

    private void setupSpinner(Spinner<Double> spinner) {
        spinner.setValueFactory(createSpinnerValueFactory());
        ViewUtils.fixSpinner(spinner);
    }

    /**
     * Creates the values that will be possible to set in the spinners
     */
    private SpinnerValueFactory<Double> createSpinnerValueFactory() {
        SpinnerValueFactory.DoubleSpinnerValueFactory doubleFactory = new SpinnerValueFactory.DoubleSpinnerValueFactory(MIN_SPINNER_VALUE, MAX_SPINNER_VALUE, INITIAL_VALUE);
        doubleFactory.setAmountToStepBy(0.1);
        return doubleFactory;
    }

    /**
     * Handles the click on the "Create Product" button. If the user
     * doesn't put the right credentials, an error label is displayed otherwise
     * it connects the user to the database.
     */
    @FXML
    public void onCreateProduct() {
        if (this.unitField.getSelectionModel().getSelectedItem() == null) {
            this.errorLabel.setText("Invalid Unit");
            return;
        }
        String productName = this.productNameField.getText().trim();
        Product.Unit productUnit = Product.Unit.create(this.unitField.getSelectionModel().getSelectedItem());
        double productLipids = this.lipidsSpinner.getValue();
        double productProteins = this.proteinsSpinner.getValue();
        double productCarbohydrates = this.carbohydratesSpinner.getValue();

        if (productLipids + productProteins + productCarbohydrates > 100) {
            this.errorLabel.setText("There can't be more than 100g of macros per 100g of product.");
        } else {
            this.controller.createProduct(productName, productUnit, productLipids, productProteins, productCarbohydrates);
        }
    }

    public void showProductAlreadyExistsError() {
        this.errorLabel.setText("Product Already Exists");
    }

    public void showSuccessPopup() {
        Popup.showPopup(AlertType.INFORMATION, "Product added", "Product added to the database !");
    }

    public void close() {
        Stage stage = (Stage) this.errorLabel.getScene().getWindow();
        stage.fireEvent(new WindowEvent(stage, WindowEvent.WINDOW_CLOSE_REQUEST));
    }

    public void showInvalidName() {
        this.errorLabel.setText("Invalid Product name");
    }
}
