package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.controllers.NutritionalValuesController;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.utils.ProductDoublePairConverter;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.control.Button;
import javafx.scene.control.Spinner;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.util.Pair;
import org.controlsfx.control.PrefixSelectionComboBox;

import java.util.ArrayList;
import java.util.List;

/**
 * A List Component is a FXML component composed of :
 * 1) A list of list items (i.e list of HBoxes containing a ComboBox, a Spinner and a delete Button).
 * 2) Two buttons (addButton and clearAllButton)
 * <p>
 * The ComboBoxes are used to select a product and the associated Spinner
 * is used to set the number associated with the selected Product.
 * <p>
 * The addButton is used to add an item in the ListComponent (i.e it add a new ComboBox and Spinner).
 * The clearAllButton clear the ListComponent (i.e it deletes all the comboBoxes and spinners).
 * <p>
 * It has also two optional settings :
 * - A button to check the nutritional values the products
 * - A button to check the availability of products in the shops
 */
public final class ListComponent<T> {

    private static final int BIG_BUTTON_WIDTH = 160;
    private static final int BIG_BUTTON_HEIGHT = 70;

    private static final int LITTLE_BUTTON_WIDTH = 250;
    private static final int LITTLE_BUTTON_HEIGHT = 35;
    private static final double MAX_SPINNER_VALUE = Double.MAX_VALUE;
    private static final double MIN_SPINNER_VALUE = 0.1;
    private static final double AMOUNT_TO_STEP_BY_IN_SPINNER = 0.1;
    private final ProductDoublePairConverter<T> converter;
    private final NutritionalValuesController nutritionalValuesController;
    private final boolean canShowNutritionalValues;

    @FXML
    private Button addButton;
    @FXML
    private Button clearButton;
    @FXML
    private VBox globalContainer;
    @FXML
    private VBox listContainer;
    @FXML
    private HBox buttonsContainer;
    @FXML
    private List<PrefixSelectionComboBox<Product>> comboBoxes;
    @FXML
    private Button nutritionalListButton;
    @FXML
    private List<Spinner<Double>> spinners;

    private AvailabilityChecker availabilityChecker;
    private boolean hasBeenModified = false;
    private ObservableList<Product> obsAllProductList;

    private boolean canCheckProductAvailability;

    public ListComponent(VBox globalContainer, List<Product> allProductsList, List<T> itemList, ProductDoublePairConverter<T> converter, NutritionalValuesController nutritionalValuesController) {
        this(globalContainer, allProductsList, itemList, converter, false, null, true, nutritionalValuesController);
    }

    /**
     * Constructs a ListComponent in a Container (VBox) and initializes it
     *
     * @param globalContainer The Container in which the listContainer will be set.
     * @param allProductsList All the product that will be selectable in the ComboBoxes.
     * @param itemList        The items with which the list must be initialized
     */
    public ListComponent(VBox globalContainer, List<Product> allProductsList, List<T> itemList, ProductDoublePairConverter<T> converter,
                         boolean canCheckProductAvailability, AvailabilityChecker availabilityChecker,
                         boolean canShowNutritionalValues, NutritionalValuesController nutritionalValuesController) {

        this.comboBoxes = new ArrayList<>();
        this.spinners = new ArrayList<>();
        this.availabilityChecker = availabilityChecker;
        this.nutritionalValuesController = nutritionalValuesController;
        this.canShowNutritionalValues = canShowNutritionalValues;
        this.converter = converter;

        this.initializeContainers(globalContainer);


        obsAllProductList = FXCollections.observableArrayList();
        obsAllProductList.addAll(allProductsList);

        this.canCheckProductAvailability = canCheckProductAvailability;


        for (T item : itemList) {
            Pair<Product, Double> pair = converter.toProductDoublePair(item);
            HBox newItem = this.createListItem(pair);
            this.listContainer.getChildren().add(newItem);
        }

        this.setUpButtons();
    }

    public ListComponent(VBox globalContainer, List<Product> productList, ProductDoublePairConverter<T> converter) {
        this(globalContainer, productList, new ArrayList<>(), converter, false, null, false, null);
    }

    public boolean hasBeenModified() {
        return this.hasBeenModified;
    }

    public void setHasBeenModified(boolean hasBeenModified) {
        this.hasBeenModified = hasBeenModified;
    }

    /**
     * Creates a listContainer and a buttonContainer inside the globalContainer
     *
     * @param globalContainer the container containing the list and button Containers.
     */
    private void initializeContainers(VBox globalContainer) {
        this.globalContainer = globalContainer;
        this.listContainer = new VBox();

        this.buttonsContainer = new HBox();
        this.buttonsContainer.setAlignment(Pos.BOTTOM_CENTER);

        this.globalContainer.getChildren().add(this.listContainer);
        this.globalContainer.getChildren().add(this.buttonsContainer);
    }

    /**
     * Sets up the Add new Item and Clear all buttons.
     */
    private void setUpButtons() {
        this.setUpAddItemButton();
        if (this.comboBoxes.isEmpty()) {
            this.maximizeAddButton();
        } else {
            this.setUpClearAllButton();
            if (this.canShowNutritionalValues) {
                this.setUpNutritionalListButton();
            }
        }
    }

    /**
     * Creates and sets up the addButton.
     */
    private void setUpAddItemButton() {
        this.addButton = new Button("Add item");
        this.addButton.setId("addButton");
        this.buttonsContainer.getChildren().add(this.addButton);
        this.addButton.setOnAction(this::onAddItem);
    }

    /**
     * Creates and sets up the clearAllButton.
     */
    private void setUpClearAllButton() {
        this.addButton.setPrefSize(LITTLE_BUTTON_WIDTH, LITTLE_BUTTON_HEIGHT);
        this.clearButton = new Button("Clear all");
        this.clearButton.setId("clearButton");
        this.clearButton.setPrefSize(LITTLE_BUTTON_WIDTH, LITTLE_BUTTON_HEIGHT);
        this.clearButton.setOnAction(this::onClearAll);

        buttonsContainer.getChildren().add(this.clearButton);
        HBox.setMargin(clearButton, new Insets(0, 0, 0, 10));
    }

    /**
     * Creates and sets up the button used to get the total nutritional value the list.
     */
    private void setUpNutritionalListButton() {
        this.addButton.setPrefSize(LITTLE_BUTTON_WIDTH, LITTLE_BUTTON_HEIGHT);
        this.nutritionalListButton = new Button("Total nutritional Value");
        this.nutritionalListButton.setId("nutritionalListButton");
        this.nutritionalListButton.setPrefSize(LITTLE_BUTTON_WIDTH, LITTLE_BUTTON_HEIGHT);
        this.nutritionalListButton.setOnAction(e -> this.nutritionalValuesController.computeAndShowNutritionalListView(
                (List<ProductAmount>) this.getItemList() // We will only use this with Product Amounts, so it will work
        ));

        buttonsContainer.getChildren().add(this.nutritionalListButton);
        HBox.setMargin(this.nutritionalListButton, new Insets(0, 0, 0, 10));
    }

    /**
     * Event Listener on the addItemButton.onAction
     *
     * @param event
     */
    @FXML
    private void onAddItem(ActionEvent event) {
        // If the list was empty the "clear all" button was not setUp.
        // --> Set it up
        this.hasBeenModified = true;
        if (this.comboBoxes.isEmpty()) {
            this.setUpClearAllButton();
            if (this.canShowNutritionalValues) {
                this.setUpNutritionalListButton();
            }
        }

        HBox newItem = this.createListItem(new Pair<>(this.obsAllProductList.get(0), 1.0));
        this.listContainer.getChildren().add(newItem);

    }

    /**
     * Maximize the size of the addButton.
     */
    private void maximizeAddButton() {
        this.addButton.setPrefSize(BIG_BUTTON_WIDTH, BIG_BUTTON_HEIGHT);
    }

    /**
     * Creates a list Item on the base of a Pair (Product, Double).
     *
     * @param productWithDouble The number we have to put in the Spinner
     * @return The HBox which represents the List Item
     */
    private HBox createListItem(Pair<Product, Double> productWithDouble) {
        HBox listItem = new HBox();
        listItem.setAlignment(Pos.CENTER);
        listItem.setPrefHeight(100);
        listItem.setPrefWidth(200);

        PrefixSelectionComboBox<Product> productComboBox = addProductComboBox(listItem, productWithDouble.getKey());
        Spinner<Double> spinner = addSpinner(listItem, productWithDouble.getValue());

        if (canShowNutritionalValues) {
            this.addNutritionalButton(listItem, productComboBox, spinner);
        }

        if (canCheckProductAvailability) {
            this.addAvailabilityButton(listItem, productComboBox, spinner);
        }
        this.addDeleteListItemButton(listItem, productComboBox, spinner);

        return listItem;
    }


    /**
     * Adds The Delete Button in a List Item
     *
     * @param listItem        The listItem it has to remove
     * @param productComboBox The ComboBox of the list item it has to remove
     * @param spinner         The Spinner of the list item it has to remove
     */
    private void addDeleteListItemButton(HBox listItem, PrefixSelectionComboBox<Product> productComboBox,
                                         Spinner<Double> spinner) {
        Button deleteButton = new Button();
        deleteButton.setText("X");
        deleteButton.getStyleClass().add("dangerButton");
        deleteButton.setPrefHeight(45);
        deleteButton.setOnAction((event) -> {
            this.hasBeenModified = true;
            // Button was clicked, do something...
            this.listContainer.getChildren().remove(listItem);
            this.comboBoxes.remove(productComboBox);
            this.spinners.remove(spinner);
            if (this.comboBoxes.isEmpty()) {
                this.removeClearButton();
                this.removeNutritionalListButton();
            }
        });

        listItem.getChildren().add(deleteButton);
        HBox.setMargin(deleteButton, new Insets(0, 25, 0, 0));
    }

    /**
     * Adds the Spinner used to select the number associated with a product in a List Item
     *
     * @param listItem The listItem that contains it
     * @param number   The initial value of the spinner
     * @return The Spinner itself
     */
    private Spinner<Double> addSpinner(HBox listItem, double number) {
        Spinner<Double> spinner = new Spinner<>(MIN_SPINNER_VALUE, MAX_SPINNER_VALUE, number, AMOUNT_TO_STEP_BY_IN_SPINNER);
        spinner.setEditable(true);
        spinner.setPrefHeight(45);
        spinner.setPrefWidth(100);

        spinner.valueProperty().addListener((obs, oldValue, newValue) -> this.hasBeenModified = true);

        ViewUtils.fixSpinner(spinner);
        this.spinners.add(spinner);
        listItem.getChildren().add(spinner);
        HBox.setMargin(spinner, new Insets(0, 10, 0, 10));

        return spinner;
    }

    /**
     * Adds the ComboBox used to select a product in a List Item
     *
     * @param listItem listItem that contains the ComboBox
     * @param product  The initial selection of the ComboBox
     * @return The ComboBox itself
     */
    private PrefixSelectionComboBox<Product> addProductComboBox(HBox listItem, Product product) {
        PrefixSelectionComboBox<Product> comboBox = new PrefixSelectionComboBox<>();
        comboBox.setPrefHeight(45);
        comboBox.setPrefWidth(200);
        comboBox.getSelectionModel().select(product);

        comboBox.getItems().addAll(this.obsAllProductList);
        comboBox.setOnAction((event) -> this.hasBeenModified = true);
        this.comboBoxes.add(comboBox);
        listItem.getChildren().add(comboBox);
        HBox.setMargin(comboBox, new Insets(0, 0, 0, 80));

        return comboBox;
    }


    /**
     * Adds the availability Button in the HBox representing an item of the list.
     *
     * @param hBox
     * @param productComboBox
     * @param spinner
     */
    private void addAvailabilityButton(HBox hBox, PrefixSelectionComboBox<Product> productComboBox,
                                       Spinner<Double> spinner) {
        Button availabilityButton = new Button();
        availabilityButton.setText("Check Availability");
        availabilityButton.wrapTextProperty().setValue(true);
        availabilityButton.getStyleClass().add("listItemButton");
        availabilityButton.setPrefHeight(45);
        availabilityButton.setPrefWidth(200);
        availabilityButton.setOnAction((ActionEvent event) -> {
            Product product = productComboBox.getSelectionModel().getSelectedItem();
            this.availabilityChecker.checkProductAvailability(new ProductAmount(product, spinner.getValue()));
        });
        hBox.getChildren().add(availabilityButton);
        HBox.setMargin(availabilityButton, new Insets(0, 25, 0, 20));
    }

    /**
     * Creates and adds a button to get the nutritional value of a certain product.
     *
     * @param listItem        Where the button is going to be added.
     * @param productComboBox In order to get the Item we want the information of (Event)
     * @param spinner         In order to get the quantity of the product we want (Event)
     */
    private void addNutritionalButton(HBox listItem, PrefixSelectionComboBox<Product> productComboBox, Spinner<Double> spinner) {
        Button nutritionalButton = new Button();
        nutritionalButton.setText("Nutritional Value");
        nutritionalButton.setPrefHeight(45);
        nutritionalButton.setPrefWidth(200);
        nutritionalButton.getStyleClass().add("listItemButton");
        nutritionalButton.setOnAction((event) -> {
            Product product = productComboBox.getSelectionModel().getSelectedItem();
            this.nutritionalValuesController.computerAndShowNutritionalView(new ProductAmount(product, spinner.getValue()));
        });

        listItem.getChildren().add(nutritionalButton);
        HBox.setMargin(nutritionalButton, new Insets(0, 5, 0, 0));
    }

    /**
     * Event Listener on clearButton.onAction
     *
     * @param event
     */
    @FXML
    private void onClearAll(ActionEvent event) {
        this.hasBeenModified = true;
        this.removeClearButton();
        this.removeNutritionalListButton();

        this.listContainer.getChildren().clear();
        this.comboBoxes = new ArrayList<>();
        this.spinners = new ArrayList<>();
    }

    /**
     * Removes the nutritional value button
     * (Used when the a product is removed from the list)
     */
    private void removeNutritionalListButton() {
        this.buttonsContainer.getChildren().remove(this.nutritionalListButton);
        this.maximizeAddButton();
    }

    /**
     * Removes the clearButton from the ListComponent
     */
    private void removeClearButton() {
        this.buttonsContainer.getChildren().remove(this.clearButton);
        this.maximizeAddButton();
    }

    /**
     * Adds a new product in all the comboBoxes
     *
     * @param newProduct
     */
    public void addNewProduct(Product newProduct) {
        this.obsAllProductList.add(newProduct);
        for (PrefixSelectionComboBox<Product> comboBox : this.comboBoxes) {
            comboBox.getItems().add(newProduct);
        }
    }

    /**
     * Returns only the number of the item at a given position in the list component
     *
     * @param position The position at which we want to get the number
     * @return The number
     */
    public double getNumberAtPosition(int position) {
        Spinner<Double> spinner = this.spinners.get(position);
        return spinner.getValue();
    }

    /**
     * Sets the number of a spinner at the a given position
     *
     * @param position  The position of the spinner we want to change
     * @param newNumber The new value we want to give to it
     */
    public void setNumberAtPosition(int position, double newNumber) {
        Spinner<Double> spinner = this.spinners.get(position);
        spinner.getValueFactory().setValue(newNumber);
    }

    /**
     * Taking a product as parameter, this function removes the product from the list component
     *
     * @param product
     */
    public void removeProduct(Product product) {
        for (PrefixSelectionComboBox comboBox : this.comboBoxes) {
            comboBox.getItems().remove(product);
        }
    }

    /**
     * Returns a list of Object (of type T), there is one object in the list of each item in the component
     *
     * @return
     */
    public List<T> getItemList() {
        List<T> itemList = new ArrayList<>();
        for (Pair<Product, Double> pair : this.getProductDoublePairs()) {
            T item = this.converter.toObject(pair);
            itemList.add(item);
        }
        return itemList;
    }

    /**
     * Multiplies all the quantities by multiplier.
     *
     * @param multiplier
     */
    public void updateAmount(double multiplier) {
        for (Spinner<Double> spinner : this.spinners) {
            double currentValue = spinner.getValue();
            spinner.getValueFactory().setValue(currentValue * multiplier);
        }
    }

    /**
     * Gets a list of (poduct, double) pairs, there is an item in the list for each element of the component
     */
    private List<Pair<Product, Double>> getProductDoublePairs() {
        List<Pair<Product, Double>> productDoublePairs = new ArrayList<>();
        int i = 0;
        for (PrefixSelectionComboBox comboBox : this.comboBoxes) {
            Spinner associatedSpinner = this.spinners.get(i);
            Pair<Product, Double> pair =
                    new Pair<>((Product) comboBox.getSelectionModel().getSelectedItem(), (Double) associatedSpinner.getValueFactory().getValue());
            productDoublePairs.add(pair);
            i++;
        }

        return productDoublePairs;
    }

}
