package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.models.DaySchedule;
import be.ac.ulb.infof307.g02.models.Position;
import be.ac.ulb.infof307.g02.models.Shop;
import be.ac.ulb.infof307.g02.models.TimeSlot;
import javafx.beans.value.ObservableValue;
import javafx.concurrent.Worker.State;
import javafx.fxml.FXML;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import netscape.javascript.JSObject;
import org.jtwig.JtwigModel;
import org.jtwig.JtwigTemplate;
import org.jtwig.environment.EnvironmentConfiguration;
import org.jtwig.environment.EnvironmentConfigurationBuilder;
import org.jtwig.json.JsonExtension;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is used to wrap all the JavaScript used for the map.
 * It uses a bridge to communicate between the Java and the JavaScript
 */
public class MapComponent {

    private final static String JAVA_APP = "javaApp";

    /**
     * The webView in which the map is created
     */
    @FXML
    private WebView webView;

    /**
     * The HTML and JavaScript code that runs in the WebView
     */
    private JtwigTemplate template;

    private Position userPosition;

    /**
     * The mapView owning the mapComponent
     */
    private MapView mapView;

    /**
     * The informations associated to the markers
     */
    private List<String> markersContent;

    /**
     * The positions of the markers displayed on the map
     */
    private List<Position> markerPositions;

    /**
     * The Bridge is used to allow the WebView to call java methods by itself.
     */
    private Bridge bridge;

    public MapComponent(MapView mapView) {
        this.mapView = mapView;
        this.webView = mapView.getWebView();

        this.userPosition = mapView.getUserPosition();
        this.markersContent = new ArrayList<>();
        this.markerPositions = new ArrayList<>();

        // To be a able to use JsonEncode in the twig template
        EnvironmentConfiguration configuration = EnvironmentConfigurationBuilder.configuration().extensions()
                .add(JsonExtension.defaultJsonExtension()).and().build();

        this.template = JtwigTemplate.classpathTemplate(Constants.TWIG_PATH, configuration);

        this.createBridgeToJS();
    }

    /**
     * Creates the bridge between the Java Application and the JavaScript code used to interact with the map.
     */
    private void createBridgeToJS() {
        this.bridge = new Bridge();
        final WebEngine webEngine = this.webView.getEngine();
        webEngine.getLoadWorker().stateProperty()
                .addListener((ObservableValue<? extends State> ov, State oldState, State newState) -> {
                    if (newState == State.SUCCEEDED) {
                        // Gets the JavaScript running in the webView and adds the Bridge to his members.
                        // Then, we can call java code from the JavaScript.
                        JSObject win = (JSObject) webEngine.executeScript("window");
                        win.setMember(JAVA_APP, bridge);
                    }
                });
    }

    /**
     * Adds a marker to the list that will be given to the JavaScript
     *
     * @param position      The position of the marker
     * @param markerContent The content displayed in the marker
     */
    public void addMarker(Position position, String markerContent) {
        this.markerPositions.add(position);
        this.markersContent.add(markerContent);
    }

    /**
     * Creates the JavaScript model then show it in the webView
     */
    public void showMapInWebView() {
        JtwigModel model = JtwigModel.newModel().with("markerPositions", this.markerPositions)
                .with("markersContent", this.markersContent).with("userPosition", this.userPosition);

        String completeTemplate = this.template.render(model);
        this.webView.getEngine().loadContent(completeTemplate);
    }

    /**
     * Asks the JavaScript to show a route between a list of position
     *
     * @param listOfPos
     */
    public void showRoute(List<ArrayList<Double>> listOfPos) {
        this.webView.getEngine().executeScript("showRoute(" + listOfPos.toString() + ")");

    }

    /**
     * Asks the JavaScript to remove all the elements of a best tour from the map
     */
    public void onClosePane() {
        this.webView.getEngine().executeScript("removeBestRouteElements()");
    }

    /**
     * Asks the JavaScript to remove the start and end point of a best tour from the map
     */
    public void removeBestTourPositions() {
        this.webView.getEngine().executeScript("removeBestTourPositions()");
    }


    /**
     * Creates an html elements with an onclick js function on it
     *
     * @param type    The type of the element (e.g Button)
     * @param text    The text displayed in the element
     * @param onClick The method bound to the element
     * @return A string representing the html element
     */
    public String createHTMLElement(String type, String text, String onClick) {
        String element = "<" + type + " onclick = \"" + onClick + "\">" + text +
                "</" + type + ">";
        return element;
    }

    /**
     * Creates an html elements of a specified class with an onclick js function on it
     *
     * @param type    The type of the element (e.g Button)
     * @param text    The text displayed in the element
     * @param onClick The method bound to the element
     * @param _class  The class of the element
     * @return A string representing the html element
     */
    public String createHTMLElement(String type, String text, String onClick, String _class) {
        String element = "<" + type + " onclick = \"" + onClick + "\" class=\"" + _class + "\">" + text +
                "</" + type + ">";
        return element;
    }

    /**
     * Creates a simple html element
     *
     * @param type The type of the element (e.g Button)
     * @param text The text displayed in the element
     * @return A string representing the html element
     */
    public String createHTMLElement(String type, String text) {
        String element = "<" + type + ">" + text + "</" +
                type + ">";
        return element;
    }

    /**
     * Creates an html schedule
     *
     * @param daySchedules
     * @return A string representing the html schedule
     */
    public String createHTMLSchedule(List<DaySchedule> daySchedules) {
        StringBuilder builder = new StringBuilder();
        builder.append("<div class='leftAligned'>");
        for (DaySchedule daySchedule : daySchedules) {
            builder.append("<br>");
            builder.append(this.createDayScheduleString(daySchedule));
        }
        builder.append("</div>");

        return builder.toString();
    }

    /**
     * Create a string representing the schedule for a day. <br>
     * Example output (open on that day): <strong>Monday</strong>: 08:00 - 12:00
     * 15:00 - 18:00 <br>
     * Example output (closed on that day): <strong>Monday</strong>: Closed
     *
     * @param daySchedule
     */
    private String createDayScheduleString(DaySchedule daySchedule) {
        StringBuilder builder = new StringBuilder();

        // Put day name
        builder.append("<strong>");
        builder.append(daySchedule.getDayName());
        builder.append("</strong>: ");
        if (daySchedule.isEmpty()) {
            builder.append("Closed");
        } else {
            for (TimeSlot timeSlot : daySchedule.getTimeSlots()) {
                builder.append(timeSlot.toString());
                builder.append(" ");
            }
        }

        return builder.toString();
    }

    /**
     * Asks the JavaScript to hide some shops
     *
     * @param shopsToKeep
     */
    public void hideFilteredShops(List<Shop> shopsToKeep) {
        List<Position> positionsToKeep = new ArrayList<>();
        for (Shop shop : shopsToKeep) {
            positionsToKeep.add(shop.getPosition());
        }

        // Gets the Javascript running in the webView and asks to hide some shops
        JSObject jsobj = (JSObject) this.webView.getEngine().executeScript("window");
        jsobj.call("hideFilteredShops", positionsToKeep);
    }

    /**
     * Tells the JavaScript that it can redisplay all the Shops
     */
    public void clearFilters() {
        this.webView.getEngine().executeScript("clearFilters()");
    }


    /**
     * The Bridge is used to allow the WebView to call java methods by itself.
     */
    public class Bridge {
        /**
         * Allows to update the user position when it was changed in the map.
         *
         * @param lat
         * @param lng
         */
        public void updateUserPosition(double lat, double lng) {
            mapView.updateUserPosition(lat, lng);

        }

        /**
         * Allows to export all the products of a shop when a button is clicked on the map
         */
        public void exportShopArticles(int shopId) {
            mapView.exportShopArticles(shopId);
        }

        public void triggerTourReady(double startLat, double startLng, double endLat, double endLng) {
            Position startingPosition = new Position(startLat, startLng);
            Position endingPosition = new Position(endLat, endLng);
            mapView.acknowledgeControllerTourReady(startingPosition, endingPosition);
        }

        /**
         * Allows to import products of a shop when a button is clicked on the map
         */
        public void importShopArticles(int shopId) {
            mapView.importShopArticles(shopId);
        }


    }
}
