package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.controllers.MapController;
import be.ac.ulb.infof307.g02.models.*;
import com.jfoenix.controls.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.scene.web.WebView;
import org.controlsfx.control.textfield.TextFields;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * View showing the map.
 */
public class MapView extends ViewWithMenu {

    /**
     * The fxml resource for the view
     */
    private static final String MAP_FXML = "map.fxml";

    /**
     * The fxml resource for the filter menu
     */
    private static final String FILTER_MENU_FXML = "filterMenu.fxml";
    private final static int STACK_PANE_WIDTH = 200;
    private final static int CLOSE_BUTTON_HEIGHT = 30;
    private MapController mapController;

    @FXML
    private AnchorPane anchorpane;
    @FXML
    private Button filterMenuButton;
    @FXML
    private JFXDrawer drawer;
    @FXML
    private Button filterButton;
    @FXML
    private Button clearFiltersButton;
    @FXML
    private VBox filterMenuBox;
    @FXML
    private JFXTextField shopNameField;
    @FXML
    private JFXChipView<Product> stockContainer;
    @FXML
    private JFXSlider distanceSlider;
    @FXML
    private JFXTimePicker timePicker;
    @FXML
    private CheckBox mondayCheckBox;
    @FXML
    private CheckBox tuesdayCheckBox;
    @FXML
    private CheckBox wednesdayCheckBox;
    @FXML
    private CheckBox thursdayCheckBox;
    @FXML
    private CheckBox fridayCheckBox;
    @FXML
    private CheckBox saturdayCheckBox;
    @FXML
    private CheckBox sundayCheckBox;
    @FXML
    private CheckBox publicHolidayCheckBox;

    /**
     * The webView containing the map
     */
    @FXML
    private WebView webView;
    /**
     * The view container
     */
    @FXML
    private AnchorPane anchorPane;

    private MapComponent mapComponent;
    private List<Shop> shops;


    /**
     * Initializes the view, creates the map Component and the filter menu
     *
     * @param controller The MapController
     * @param scene      The Scene in which the view will be created
     */
    public MapView(MapController controller, Scene scene) {
        super(controller);
        this.mapController = controller;
        this.shops = this.mapController.getShops();


        try {
            FXMLLoader loader = new FXMLLoader(getClass().getResource(Constants.FXML_FOLDER_PATH + MAP_FXML));
            loader.setController(this);
            scene.setRoot(loader.load());

            FXMLLoader filterMenuloader = new FXMLLoader(getClass().getResource(Constants.FXML_FOLDER_PATH + FILTER_MENU_FXML));
            filterMenuloader.setController(this);
            this.filterMenuBox = filterMenuloader.load();
            this.createMenu();

            this.mapComponent = new MapComponent(this);
            this.setupMap();

        } catch (IOException e) {
            // Will never happen
        }

    }


    /**
     * Sets up the map: it creates all the markers for the shops, adds them to
     * the map and then shows the map.
     */
    private void setupMap() {
        this.setupFilterMenu();


        for (Shop shop : this.shops) {
            this.mapComponent.addMarker(shop.getPosition(), this.createShopMarkerContent(shop));
        }
        this.mapComponent.showMapInWebView();
    }

    /**
     * Sets up all the UI element for the filter menu
     */
    private void setupFilterMenu() {
        this.drawer.toBack();
        TextFields.bindAutoCompletion(shopNameField, this.shops);
        this.stockContainer.getSuggestions().addAll(this.mapController.getProducts());
        this.distanceSlider.setValue(0);


        filterButton.setOnAction((event -> {
            String shopName = this.shopNameField.getText().trim();
            int distance = (int) this.distanceSlider.getValue();
            List<Day> daysList = this.getSelectedDays();
            Time time = this.getTimeFromPicker();
            List<Product> products = this.stockContainer.getChips();

            FilterData filterData = new FilterData(shopName, distance, daysList, time, products);
            this.mapController.filterShops(filterData);

        }));

        this.clearFiltersButton.setOnAction(event -> {

            this.clearAllFiltersFields();
            clearFiltersOnMap();
            mapController.clearFilters(); // reset the list of shops to all the shops contained in the db
        });

        this.drawer.setSidePane(filterMenuBox);
        this.filterMenuButton.setOnAction((event) -> {
            if (this.drawer.isOpened()) {
                this.hideFilterMenu();
            } else {
                this.drawer.toFront();
                this.drawer.open();

            }
        });

    }

    /**
     * Creates the marker content for the shop, e.g. create a String containing
     * HTML that shows the shop name and its schedule.
     *
     * @param shop
     */
    private String createShopMarkerContent(Shop shop) {
        List<DaySchedule> daySchedules = shop.getDaySchedules();

        StringBuilder builder = new StringBuilder();

        // Put shop name as marker title
        builder.append(this.mapComponent.createHTMLElement("h3", shop.getName()));

        // Add button for direction
        double shopLatitude = shop.getPosition().getLatitude();
        double shopLongitude = shop.getPosition().getLongitude();
        String args = shopLatitude + "," + shopLongitude;
        String navigation = "navigation(" + args + ")";
        builder.append(this.mapComponent.createHTMLElement("button", "Get Direction", navigation, "direction"));

        // Add export button
        String exportShopArticles = "exportShopArticles(" + shop.getId() + ")";
        builder.append(this.mapComponent.createHTMLElement("button", "Export Articles", exportShopArticles));

        // Add import button
        String importShopArticles = "importShopArticles(" + shop.getId() + ")";
        builder.append(this.mapComponent.createHTMLElement("button", "Import Articles", importShopArticles));


        // Add schedule
        builder.append(this.mapComponent.createHTMLSchedule(daySchedules));

        return builder.toString();
    }

    /**
     * Gets all the days that the user has selected
     *
     * @return The selected days
     */
    private List<Day> getSelectedDays() {
        List<Day> daysList = new ArrayList<>();
        if (this.mondayCheckBox.isSelected()) {
            daysList.add(Day.MONDAY);
        }
        if (this.tuesdayCheckBox.isSelected()) {
            daysList.add(Day.TUESDAY);
        }
        if (this.wednesdayCheckBox.isSelected()) {
            daysList.add(Day.WEDNESDAY);
        }
        if (this.thursdayCheckBox.isSelected()) {
            daysList.add(Day.THURSDAY);
        }
        if (this.fridayCheckBox.isSelected()) {
            daysList.add(Day.FRIDAY);
        }
        if (this.saturdayCheckBox.isSelected()) {
            daysList.add(Day.SATURDAY);
        }
        if (this.sundayCheckBox.isSelected()) {
            daysList.add(Day.SUNDAY);
        }
        if (this.publicHolidayCheckBox.isSelected()) {
            daysList.add(Day.PUBLIC_HOLIDAY);
        }
        return daysList;
    }

    /**
     * Gets the time that the user has selected
     *
     * @return The selected time
     */
    private Time getTimeFromPicker() {
        Time time = null;
        java.time.LocalTime localTime = timePicker.getValue();
        if (localTime != null) {
            time = new Time(localTime.getHour(), localTime.getMinute());
        }
        return time;
    }

    /**
     * Resets all filters to default
     */
    private void clearAllFiltersFields() {
        this.shopNameField.setText("");
        this.distanceSlider.setValue(0);
        this.mondayCheckBox.setSelected(false);
        this.tuesdayCheckBox.setSelected(false);
        this.wednesdayCheckBox.setSelected(false);
        this.thursdayCheckBox.setSelected(false);
        this.fridayCheckBox.setSelected(false);
        this.saturdayCheckBox.setSelected(false);
        this.sundayCheckBox.setSelected(false);
        this.publicHolidayCheckBox.setSelected(false);
        this.timePicker.setValue(null);
        for (int i = this.stockContainer.getChips().size() - 1; i > -1; i--) {
            this.stockContainer.getChips().remove(i);
        }
    }

    /**
     * Updates the current position of the user
     *
     * @param latitude  His new latitude
     * @param longitude His new longitude
     */
    public void updateUserPosition(double latitude, double longitude) {
        this.mapController.updateUserPosition(new Position(latitude, longitude));
    }

    /**
     * Allows the user to export the articles of a shop
     *
     * @param shopId The shop from which the articles should be exported
     */
    public void exportShopArticles(int shopId) {
        this.mapController.onExportArticles(shopId);
    }

    /**
     * Allows the user to import articles in a shop
     *
     * @param shopId The shop in which the articles should be imported
     */
    public void importShopArticles(int shopId) {
        this.mapController.importShopArticles(shopId);
    }

    /**
     * Show a best tour on the map
     *
     * @param bestTour The best tour that has to be shown
     */
    public void showBestTour(BestTour bestTour) {
        this.showBestTourPane(bestTour.getGroceryLists());
        this.showBestTourRoute(bestTour.getTour());
    }

    /**
     * Creates the pane used to show the user which article of a grocery list
     * he has to buy in each shop
     *
     * @param groceryLists
     */
    private void showBestTourPane(List<GroceryList> groceryLists) {
        StackPane stackPane = new StackPane();
        Accordion accordion = new Accordion();
        Button closeButton = new Button("Close");

        closeButton.setOnAction((event) -> {
            // Button was clicked, do something...
            this.anchorPane.getChildren().remove(stackPane);
            this.mapComponent.onClosePane();
        });

        // Fill the pane with the grocery list for each shop
        boolean first = true;
        for (GroceryList groceryList : groceryLists) {
            ObservableList<String> productList = FXCollections.observableList(groceryList.toList());
            TitledPane pane = new TitledPane(groceryList.getGroceryListName(), new ListView<>(productList));
            if (first) {
                accordion.setExpandedPane(pane);
                first = false;
            }
            accordion.getPanes().add(pane);
        }

        this.setBestTourPaneStyle(stackPane, accordion, closeButton);
        stackPane.getChildren().add(accordion);
        stackPane.getChildren().add(closeButton);
        this.anchorPane.getChildren().add(stackPane);
    }

    /**
     * Shows the User the route he has to take for a best tour
     *
     * @param path
     */
    private void showBestTourRoute(List<Position> path) {
        ArrayList<ArrayList<Double>> listOfPos = new ArrayList<>();
        for (Position position : path) {
            ArrayList<Double> posTuple = new ArrayList<>();
            posTuple.add(position.getLatitude());
            posTuple.add(position.getLongitude());
            listOfPos.add(posTuple);
        }
        this.mapComponent.showRoute(listOfPos);
    }

    /**
     * Sets the style of the pane of the best tour.
     *
     * @param stackPane
     * @param accordion
     * @param closeButton
     */
    private void setBestTourPaneStyle(StackPane stackPane, Accordion accordion, Button closeButton) {
        accordion.getStyleClass().add("accordion");
        stackPane.setPrefWidth(STACK_PANE_WIDTH);
        stackPane.setMaxSize(Double.MAX_VALUE, Double.MAX_VALUE);
        closeButton.setMaxWidth(Double.MAX_VALUE);
        closeButton.setPrefHeight(CLOSE_BUTTON_HEIGHT);
        closeButton.getStyleClass().add("closeButton");
        stackPane.setAlignment(Pos.BOTTOM_LEFT);
        stackPane.getStyleClass().add("stackPane");
        AnchorPane.setBottomAnchor(stackPane, 0.0);
        AnchorPane.setTopAnchor(stackPane, 0.0);
    }

    /**
     * Gets all the shops displayed in the view
     */
    public List<Shop> getShops() {
        return new ArrayList<>(this.shops);
    }

    /**
     * Gets the webView in which the map is displayed
     */
    public WebView getWebView() {
        return this.webView;
    }


    public Position getUserPosition() {
        return this.mapController.getUserPosition();
    }

    /**
     * Tells the Controller that the map is ready to display the Best Tour
     *
     * @param start The start position of the best tour
     * @param end   The end position of the best tour
     */
    public void acknowledgeControllerTourReady(Position start, Position end) {
        this.mapController.onPositionsForTourReady(start, end);

    }

    public void showNoBestTourError() {
        Popup.showPopupNoTourError();
    }

    public void showUnexpectedError() {
        Popup.showUnexpectedError();

    }

    /**
     * Removes the start and end markers for a best tour
     */
    public void removeBestTourPositions() {
        this.mapComponent.removeBestTourPositions();
    }

    /**
     * Filters the shops markers on the map
     *
     * @param shopsToKeep The shops that should remain on the map
     */
    public void filterShops(List<Shop> shopsToKeep) {
        this.mapComponent.hideFilteredShops(shopsToKeep);
    }

    public void hideFilterMenu() {
        this.drawer.close();
        this.drawer.toBack();
    }

    /**
     * Re-displays all the shops on the map
     */
    public void clearFiltersOnMap() {
        mapComponent.clearFilters();
    }
}
