package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.controllers.NutritionalValuesController;
import be.ac.ulb.infof307.g02.controllers.RecipeController;
import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.Recipe;
import be.ac.ulb.infof307.g02.utils.ProductAmountConverter;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Spinner;
import javafx.scene.control.SpinnerValueFactory;
import javafx.scene.control.TextArea;
import javafx.scene.layout.VBox;
import javafx.scene.text.Text;

import java.io.IOException;
import java.util.List;

/**
 * View of the recipe scene.
 */
public class RecipeView extends ViewWithMenu {

    /**
     * The fxml resource for the view
     */
    private static final String RECIPE_FXML = "recipe.fxml";
    @FXML
    private Spinner<Integer> numPeopleSpinner;
    @FXML
    private Text title;
    @FXML
    private TextArea instructionTextArea;
    @FXML
    private VBox listViewContainer;

    private ListComponent<ProductAmount> listComponent;

    private RecipeController recipeController;

    public RecipeView(RecipeController controller, Scene scene, Recipe recipe,
                      List<Product> allProductsList) {

        super(controller);
        this.recipeController = controller;

        try {
            FXMLLoader loader = new FXMLLoader(getClass().getResource(Constants.FXML_FOLDER_PATH + RECIPE_FXML));
            loader.setController(this);
            scene.setRoot(loader.load());
            this.setUpSpinner(this.recipeController.getNumberPeople());
            this.setUpInstructionsTextArea(this.recipeController.getInstructions());

            this.title.setText(this.recipeController.getRecipeName());

            this.listComponent = new ListComponent<>(listViewContainer, allProductsList, recipe.getProductAmountList(), new ProductAmountConverter(), new NutritionalValuesController());
            this.createMenu();

        } catch (IOException e) {
            // Will never happen
        }
    }

    /**
     * Writes the instructions string given as parameter in the text area.
     * Add a listener so that the recipeController is notified whenever a change is made.
     *
     * @param instructions
     */
    private void setUpInstructionsTextArea(String instructions) {
        this.instructionTextArea.setText(instructions);
        this.instructionTextArea.textProperty().addListener((obs, oldValue, newValue) -> this.recipeController.acknowledgeChange());
    }

    /**
     * Sets the spinner to the value given as parameter.
     * The spinner value represents the number os people for which the recipe was conceived for.
     * Therefore, the value must be an integer non negative. In fact, it must be defined
     * between 1 and 1000.
     * <p>
     * Adds a listener so that the recipeController gets notified whenever a change is made.
     *
     * @param numPeople
     */
    private void setUpSpinner(int numPeople) {
        SpinnerValueFactory<Integer> dblFactory = new SpinnerValueFactory.IntegerSpinnerValueFactory(1, 1000, numPeople);
        this.numPeopleSpinner.setValueFactory(dblFactory);
        this.numPeopleSpinner.focusedProperty().addListener((observable, oldValue, newValue) -> {
            if (!newValue) {
                this.numPeopleSpinner.increment(0);
            }
        });
        this.numPeopleSpinner.valueProperty().addListener((obs, oldNumPeople, newNumPeople) -> {
            this.recipeController.acknowledgeChange();
            this.listComponent.updateAmount(((double) newNumPeople) / oldNumPeople);
        });
        // update value when the focus is lost so we don't have to type enter
        // https://stackoverflow.com/a/39380146
        this.numPeopleSpinner.focusedProperty().addListener((observable, oldValue, newValue) -> {
            if (!newValue) {
                this.numPeopleSpinner.increment(0); // won't change value, but will commit
                // editor
            }
        });
    }

    /**
     * Called whenever the user clicks on the save button.
     * It tries to save the recipe. If a problem occurs during this operation
     * a popup showing the error is displayed on the screen.
     */
    @FXML
    public void onSaveChanges() {
        try {
            this.recipeController.saveRecipe();
        } catch (RecipeCreationException e) {
            this.showErrorSavingRecipe();
        }
    }

    /**
     * Called whenever the user clicks on the generate
     * grocery list button. It asks the recipeController to generate a list
     */
    @FXML
    public void onGenerateGroceryList() {
        this.recipeController.generateGroceryList();
    }

    /**
     * Returns the amount of a certain product located at a given
     * position in the recipe.
     *
     * @param position
     * @return amount
     */
    public double getNumberAtPosition(int position) {
        return this.listComponent.getNumberAtPosition(position);
    }

    /**
     * Allows to change the amount value of a certain product at
     * a given position in the recipe.
     *
     * @param position
     * @param newAmount
     */
    public void setAmountOfItemAtPosition(int position, double newAmount) {
        this.listComponent.setNumberAtPosition(position, newAmount);
    }

    /**
     * Shows a popup stating that the recipe was successfully saved.
     */
    public void showSuccessPopup() {
        Popup.showPopupRecipeSaved();
    }

    /**
     * Opens a popup asking the user for a confirmation on whether ir not he/she
     * wants to save the recipe. The boolean answer is then returned.
     *
     * @return
     */
    public boolean showSaveMessageBox() {
        return Popup.showPopupConfirmSaveRecipe();
    }

    /**
     * Opens a popup informing the user that an error has occurred while saving the recipe.
     */
    public void showErrorSavingRecipe() {
        Popup.showPopupErrorSavingRecipe();
    }

    /**
     * Returns the value present in the spinner containing the information of how
     * many people was the recipe conceived for.
     *
     * @return
     */
    public int getNumPeopleValue() {
        return this.numPeopleSpinner.getValue();
    }

    /**
     * Returns the string present in the instructions text area
     *
     * @return
     */
    public String getInstructionsText() {
        return this.instructionTextArea.getText().trim();
    }

    /**
     * Opens a popup asking the user to confirm the overwrite
     * The boolean answer is then returned
     *
     * @return
     */
    public boolean askUserOverWriteConfirmation() {
        return Popup.showPopupOverwriteConfirmation();
    }


    public void setIsListModified(boolean isListModified) {
        this.listComponent.setHasBeenModified(isListModified);
    }

    public boolean isListModified() {
        return this.listComponent.hasBeenModified();
    }

    public List<ProductAmount> getProductAmountsList() {
        return this.listComponent.getItemList();
    }

    public void showErrorGettingRecipe() {
        Popup.showPopupErrorGettingRecipe();
    }
}
