package be.ac.ulb.infof307.g02.views;

import be.ac.ulb.infof307.g02.constants.Constants;
import be.ac.ulb.infof307.g02.controllers.SelectionController;
import javafx.beans.binding.Bindings;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Insets;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.TextField;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.scene.text.Font;
import javafx.scene.text.Text;
import javafx.scene.text.TextFlow;

import java.io.IOException;

/**
 * Used to display all the grocery lists and recipes of a user.
 */
public class SelectionView extends ViewWithMenu {

    /**
     * The fxml resource for the view
     */
    private static final String SELECTION_FXML = "selection.fxml";
    @FXML
    private Text title;
    @FXML
    private VBox allItemsVBox;
    @FXML
    private Text errorMessageText;
    @FXML
    private TextField newItemTextField;
    @FXML
    private Button createItemButton;
    private SelectionController selectionController;

    public SelectionView(SelectionController controller, Scene scene, String title) {
        super(controller);
        this.selectionController = controller;
        try {
            FXMLLoader loader = new FXMLLoader(getClass().getResource(Constants.FXML_FOLDER_PATH + SELECTION_FXML));
            loader.setController(this);
            scene.setRoot(loader.load());
            this.createMenu();

            this.title.setText(title);
        } catch (IOException e) {
            // Don't need to tell that to the user
        }

    }

    /**
     * Initializes the Window, disable the create list button while
     * newGroceryListTextField is empty field.
     */
    @FXML
    private void initialize() {
        // Disable "Create List" button if the list name field is empty
        this.createItemButton.disableProperty().bind(Bindings.isEmpty(newItemTextField.textProperty()));
    }

    private String getNewItemName() {
        return this.newItemTextField.getText();
    }

    public void setErrorText(String message) {
        this.errorMessageText.setText(message);
    }

    /**
     * Creates a new HBox with the help of createListItem() on the view and adds
     * it to the VBox
     *
     * @param listName
     */
    public void createNewEntry(String listName) {
        HBox newItem = this.createListItem(listName);
        this.allItemsVBox.getChildren().add(newItem);

    }

    /**
     * Creates the HBox with the TextView as
     * well as the delete button inside
     *
     * @param listName
     * @return
     */
    private HBox createListItem(String listName) {
        HBox hbox = this.createHBox();
        TextFlow textFlow = this.createItemName(listName, hbox);
        Button deleteButton = this.createDeleteButton(listName, hbox);

        hbox.getChildren().add(textFlow);
        hbox.getChildren().add(deleteButton);
        return hbox;
    }

    /**
     * Creates the delete button for an item
     *
     * @param listName
     * @param hBox
     * @return
     */
    private Button createDeleteButton(String listName, HBox hBox) {
        // Adding delete button
        Button deleteButton = new Button();
        deleteButton.setText("X");
        deleteButton.setMinHeight(25);
        deleteButton.setMinWidth(25);
        deleteButton.getStyleClass().add("dangerButton");
        deleteButton.setOnAction((event) -> {
            this.allItemsVBox.getChildren().remove(hBox);
            this.setErrorText("");
            this.selectionController.deleteItem(listName);
        });
        return deleteButton;
    }

    /**
     * Creates the name of an item.
     * Takes all the available space except for the delete button)
     * the text must be wrapped in a textFlow in order to take all the available space.
     *
     * @param listName
     * @param hBox
     * @return
     */
    private TextFlow createItemName(String listName, HBox hBox) {
        TextFlow textFlow = new TextFlow();
        Text itemName = new Text();
        itemName.setText(listName);
        itemName.setFont(new Font("Roboto Mono Light for Powerline", 26));
        textFlow.getChildren().add(itemName);
        hBox.setOnMouseClicked(((event) -> {
            // Show the list that was clicked
            this.selectionController.onSelection(listName);
        }));
        textFlow.setMaxWidth(Double.MAX_VALUE);
        textFlow.setPrefWidth(javafx.scene.control.Control.USE_COMPUTED_SIZE);
        textFlow.setMinWidth(javafx.scene.control.Control.USE_COMPUTED_SIZE);
        HBox.setHgrow(textFlow, Priority.ALWAYS);
        textFlow.setPadding(new Insets(5, 0, 0, 0));

        return textFlow;
    }

    /**
     * Creates the HBox the name and the delete button of an item.
     *
     * @return
     */
    private HBox createHBox() {
        HBox hbox = new HBox();
        hbox.setStyle("-fx-cursor: hand; " + "-fx-border-color: transparent transparent #dddddd transparent;"
                + "-fx-border-width: 1px;");
        hbox.setPadding(new Insets(5, 10, 5, 10));

        return hbox;
    }

    public void onCreateNewItem() {
        this.selectionController.createNewItem(this.getNewItemName());
    }


}