package tests.database;

import be.ac.ulb.infof307.g02.database.*;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.GroceryListInsertionException;
import be.ac.ulb.infof307.g02.exceptions.database.GroceryListItemInsertException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.GroceryList;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.User;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import tests.BaseTest;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class GroceryListDBTest extends BaseTest {

    private GroceryListDB groceryListDB;
    private static List<ProductAmount> itemList;
    private static User user;

    @Before
    public void setUp() throws Exception {
        DatabaseConnection databaseConnection = DatabaseConnectionSingleton.getDatabaseConnection();
        groceryListDB = new GroceryListDB(databaseConnection);
        groceryListDB.emptyAllTables();
        // Need a user in the DB for the tests.
        user = new User("azerty", "pedro1234");
        UserDB userDB = new UserDB(databaseConnection);
        userDB.addUser(user);

        // Need products in the DB for the tests.
        Product prod1 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);
        Product prod2 = new Product("salt", Product.Unit.KG, 0, 0, 0);
        Product prod3 = new Product("kiwi", Product.Unit.UNIT, 0.5, 1.1, 15);

        ProductDB prodDB = new ProductDB(databaseConnection);
        prodDB.addProduct(prod1);
        prodDB.addProduct(prod2);
        prodDB.addProduct(prod3);

        // itemList used for the tests
        ProductAmount item1 = new ProductAmount(prod1, 3);
        ProductAmount item2 = new ProductAmount(prod2, 1);

        itemList = new ArrayList<ProductAmount>() {
            {
                add(item1);
                add(item2);
            }
        };
    }

    @After
    public void tearDown() throws Exception {
        groceryListDB.emptyAllTables();
    }

    @Test
    public void getGroceryList_validUserAndName_success() throws GroceryListInsertionException, GroceryListItemInsertException, DatabaseException {
        GroceryList groceryList = new GroceryList("Super list", itemList);
        groceryListDB.addGroceryList(user, groceryList);

        GroceryList listFromDB = groceryListDB.getGroceryList(user, "Super list");

        assertEquals(listFromDB.getGroceryListName(), groceryList.getGroceryListName());
    }

    @Test
    public void getGroceryList_invalidUser_fail() throws UserCreationException, DatabaseException {
        GroceryList groceryList = new GroceryList("Super list");
        GroceryList listFromDB = groceryListDB.getGroceryList(new User("pedro", "pedro1234"), "Super list");
        assertEquals(groceryList, listFromDB);
    }

    @Test
    public void getGroceryList_invalidListName_fail() throws DatabaseException {
        GroceryList listFromDB = groceryListDB.getGroceryList(user, "blabla");
        assertTrue(listFromDB.getProductAmountList().isEmpty());
    }

    @Test
    public void addGroceryList_validList_success() throws GroceryListInsertionException, DatabaseException, GroceryListItemInsertException {
        GroceryList groceryList = new GroceryList("Super list", itemList);

        groceryListDB.addGroceryList(user, groceryList);

        GroceryList listFromDB = groceryListDB.getGroceryList(user, "Super list");
        assertEquals(listFromDB.getGroceryListName(), groceryList.getGroceryListName());
    }

    @Test(expected = GroceryListInsertionException.class)
    public void addGroceryList_duplicateList_fail() throws GroceryListInsertionException, GroceryListItemInsertException {
        GroceryList groceryList = new GroceryList("Super list", itemList);

        groceryListDB.addGroceryList(user, groceryList);
        groceryListDB.addGroceryList(user, groceryList);

    }

    @Test
    public void deleteGroceryList_ValidUserAndName_success() throws GroceryListInsertionException, GroceryListItemInsertException, DatabaseException {
        GroceryList groceryList = new GroceryList("Super list", itemList);
        groceryListDB.addGroceryList(user, groceryList);
        groceryListDB.deleteGroceryList(user, "Super list");
        GroceryList listFromDB = groceryListDB.getGroceryList(user, "Super list");
        assertTrue(listFromDB.getProductAmountList().isEmpty());
    }

    @Test
    public void addItemToGroceryList_validArguments_itemAdded() throws Exception {
        GroceryList groceryList = new GroceryList("Super list", itemList);
        groceryListDB.addGroceryList(user, groceryList);

        groceryListDB.addItemToGroceryList(user, "Super list", new ProductAmount(new Product("kiwi", Product.Unit.UNIT, 0.5, 1.1, 15), 3));

        GroceryList listFromDB = groceryListDB.getGroceryList(user, "Super list");
        boolean itemFound = false;
        for (ProductAmount item : listFromDB.getProductAmountList()) {
            if (item.getProduct().getName().equals("kiwi")) {
                itemFound = true;
            }
        }
        assertTrue(itemFound);
    }

}
