package tests.database;

import be.ac.ulb.infof307.g02.database.DatabaseConnectionSingleton;
import be.ac.ulb.infof307.g02.database.ProductDB;
import be.ac.ulb.infof307.g02.database.RecipeDB;
import be.ac.ulb.infof307.g02.database.UserDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.NoSuchRecipeException;
import be.ac.ulb.infof307.g02.exceptions.database.RecipeInsertionException;
import be.ac.ulb.infof307.g02.exceptions.database.RecipeItemInsertException;
import be.ac.ulb.infof307.g02.exceptions.model.RecipeCreationException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.Product;
import be.ac.ulb.infof307.g02.models.ProductAmount;
import be.ac.ulb.infof307.g02.models.Recipe;
import be.ac.ulb.infof307.g02.models.User;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import tests.BaseTest;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class RecipeDBTest extends BaseTest {

    private RecipeDB db;
    private List<ProductAmount> itemList;
    private User user;

    @Before
    public void setUp() throws Exception {
        db = new RecipeDB(DatabaseConnectionSingleton.getDatabaseConnection());
        db.emptyAllTables();
        // Need a user in the DB for the tests.
        user = new User("azerty", "pedro1234");
        UserDB userDB = new UserDB(DatabaseConnectionSingleton.getDatabaseConnection());
        userDB.addUser(user);

        // Need products in the DB for the tests.
        Product prod1 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);
        Product prod2 = new Product("sugar", Product.Unit.KG, 0, 0, 100);
        Product prod3 = new Product("kiwi", Product.Unit.UNIT, 0.5, 1.1, 15);

        ProductDB prodDB = new ProductDB(DatabaseConnectionSingleton.getDatabaseConnection());
        prodDB.addProduct(prod1);
        prodDB.addProduct(prod2);
        prodDB.addProduct(prod3);

        // itemList used for the tests
        ProductAmount item1 = new ProductAmount(prod1, 3);
        ProductAmount item2 = new ProductAmount(prod2, 1);

        itemList = new ArrayList<ProductAmount>() {
            {
                add(item1);
                add(item2);
            }
        };
    }

    @After
    public void tearDown() throws SQLException {
        db.emptyAllTables();
    }

    @Test
    public void createRecipe_validArguments_success() throws RecipeCreationException {
        Recipe recipe = new Recipe("Super recipe", 3);
    }

    @Test(expected = RecipeCreationException.class)
    public void createRecipe_emptyName_fail() throws RecipeCreationException {
        Recipe recipe = new Recipe("", 3);
    }

    @Test(expected = RecipeCreationException.class)
    public void createRecipe_InvalidNumPeople_fail() throws RecipeCreationException {
        Recipe recipe = new Recipe("Super recipe", -2);
    }


    @Test
    public void getRecipe_validUserAndName_success() throws DatabaseException, RecipeCreationException,
            RecipeInsertionException, RecipeItemInsertException, NoSuchRecipeException {
        Recipe recipe = new Recipe("Super recipe", 3);
        recipe.setUnderlyingProductAmountList(itemList);
        db.addRecipe(user, recipe);

        Recipe recipeFromDB = db.getRecipe(user, "Super recipe");

        assertEquals(recipeFromDB, recipe);
    }

    @Test(expected = NoSuchRecipeException.class)
    public void getRecipe_invalidUser_fail()
            throws DatabaseException, NoSuchRecipeException, UserCreationException, RecipeCreationException {
        db.getRecipe(new User("pedro", "pedro1234"), "Super recipe");
    }

    @Test(expected = NoSuchRecipeException.class)
    public void getRecipe_invalidName_fail() throws DatabaseException, NoSuchRecipeException, RecipeCreationException {
        Recipe recipeFromDB = db.getRecipe(user, "blabla");
    }

    @Test
    public void addRecipe_validRecipe_success() throws Exception {
        Recipe recipe = new Recipe("Super recipe", 8, itemList, "Mix all");

        db.addRecipe(user, recipe);

        Recipe recipeFromDB = db.getRecipe(user, "Super recipe");

        assertEquals(recipeFromDB, recipe);
    }

    @Test(expected = RecipeInsertionException.class)
    public void addRecipe_duplicateRecipe_fail()
            throws RecipeCreationException, RecipeInsertionException, RecipeItemInsertException {
        Recipe recipe = new Recipe("Super recipe", 8);
        db.addRecipe(user, recipe);
        db.addRecipe(user, recipe);
    }

    @Test(expected = NoSuchRecipeException.class)
    public void deleteRecipe_ValidUserAndName_success() throws DatabaseException, RecipeCreationException,
            RecipeInsertionException, RecipeItemInsertException, NoSuchRecipeException {
        Recipe recipe = new Recipe("Super recipe", 8);
        db.addRecipe(user, recipe);
        db.deleteRecipe(user, "Super recipe");
        Recipe recipeFromDB = db.getRecipe(user, "Super recipe");
    }

    @Test
    public void addItemToRecipe_validArguments_itemAdded() throws Exception {
        Recipe recipe = new Recipe("Super recipe", 8);
        db.addRecipe(user, recipe);

        db.addProductAmountToRecipe(user, "Super recipe", new ProductAmount(new Product("kiwi", Product.Unit.UNIT, 0.5, 1.1, 15), 3));

        Recipe recipeFromDB = db.getRecipe(user, "Super recipe");
        boolean itemFound = false;
        for (ProductAmount item : recipeFromDB.getProductAmountList()) {
            if (item.getProduct().getName().equals("kiwi")) {
                itemFound = true;
            }
        }
        assertTrue(itemFound);
    }

}
