package tests.models;

import be.ac.ulb.infof307.g02.exceptions.database.NoShopHasThatProductException;
import be.ac.ulb.infof307.g02.models.*;
import org.hamcrest.CoreMatchers;
import org.junit.BeforeClass;
import org.junit.Test;
import tests.BaseTest;

import java.util.*;

import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.*;

public class BestTourFinderTest extends BaseTest {

    private static final Product apple = new Product("Apple", Product.Unit.UNIT, 0.2, 0.3, 14);
    private static final Product banana = new Product("Banana", Product.Unit.UNIT, 0.3, 1.1, 23);
    private static final Product pineapple = new Product("Pineapple", Product.Unit.UNIT, 0.1, 0.5, 13);
    private static final Position home = new Position(50.8467, 4.3524); // Grand-Place
    private static final Position montDesArts = new Position(50.8444, 4.3559);
    private static final Position beijing = new Position(39.9042, 116.4074);
    private static final Position losAngeles = new Position(34.0522, -118.2437);
    /*
     * @formatter:off
     * The approximate conversions are:
     *     Latitude:  1 deg = 110.574 km
     *     Longitude: 1 deg = 111.320 * cos(latitude) km
     *
     * We'll use latitude very close to 0, so 1 deg longitude = 111.320 km
     * @formatter:on
     */
    private static final double KM_IN_DEGREES_LAT = 1 / 110.574;
    private static final double KM_IN_DEGREES_LONG = 1 / 110.320;
    private static Schedule schedule;

    @BeforeClass
    public static void setUpBeforeClass() {
        schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(1, 2), new Time(3, 4)), Day.TUESDAY);
    }

    /**
     * We test the simplest case: there is only one shop and it has all the
     * products.
     *
     * Shops: - Aldi: Apple 5, Banana 10
     *
     * Grocery List: Apple 5, Banana 5 FindShopsResult: Aldi
     */
    @Test
    public void getTour_simplestCase_success() throws NoShopHasThatProductException {
        // Create shops
        Shop aldi = new Shop(0, "Aldi", schedule, montDesArts);
        aldi.addToStock(new Article(apple, (float) 5.0));
        aldi.addToStock(new Article(banana, (float) 10.0));

        List<Shop> shops = new ArrayList<>();
        shops.add(aldi);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));
        groceryList.addProductOrIncreaseAmount(new ProductAmount(banana, 5));

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        assertTourIsCorrect(shops, tourFinder);
    }

    /**
     * We test a simple: there is a very cheap and a very expensive shop at the same
     * place. Obviously, we should use the cheap shop.
     *
     * Shops: - Aldi: Apple 5, Banana 10 - LuxeShop: Apple 500, Banana 1000
     *
     * Grocery List: Apple 5, Banana 5 FindShopsResult: Aldi
     */
    @Test
    public void getTour_cheapAndExpensiveShop_useCheapShop() throws NoShopHasThatProductException {
        // Create shops
        Shop aldi = new Shop(0, "Aldi", schedule, montDesArts);
        aldi.addToStock(new Article(apple, (float) 5.0));
        aldi.addToStock(new Article(banana, (float) 10.0));

        Shop luxeShop = new Shop(0, "LuxeShop", schedule, montDesArts);
        luxeShop.addToStock(new Article(apple, (float) 500.0));
        luxeShop.addToStock(new Article(banana, (float) 1000.0));

        List<Shop> shops = new ArrayList<>();
        shops.add(aldi);
        shops.add(luxeShop);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));
        groceryList.addProductOrIncreaseAmount(new ProductAmount(banana, 5));

        // expected tour
        List<Shop> expectedTour = new ArrayList<>();
        expectedTour.add(aldi);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        assertTourIsCorrect(expectedTour, tourFinder);
    }

    /**
     * We test a more complex case: we want apples and bananas. All shops are at the
     * same place. Bananas are really cheap at Aldi and expensive elsewhere and
     * apples are really cheap at Colruyt and expensive elsewhere. We should go to
     * Aldi and Colruyt only.
     *
     * Shops: - Aldi: Apple 5, Banana 1000 - Colruyt: Apple 500, Banana 10 -
     * LuxeShop: Apple 50000, Banana 10000
     *
     * Grocery List: Apple 5, Banana 5 FindShopsResult: Aldi, Colruyt
     */
    @Test
    public void getTour_TwoGoodShops_useBothShops() throws NoShopHasThatProductException {
        // Create shops
        Shop aldi = new Shop(0, "Aldi", schedule, montDesArts);
        aldi.addToStock(new Article(apple, (float) 5.0));
        aldi.addToStock(new Article(banana, (float) 10000.0));

        Shop colruyt = new Shop(0, "Colruyt", schedule, montDesArts);
        colruyt.addToStock(new Article(apple, (float) 50000.0));
        colruyt.addToStock(new Article(banana, (float) 10.0));

        Shop luxeShop = new Shop(0, "LuxeShop", schedule, montDesArts);
        luxeShop.addToStock(new Article(apple, (float) 50000.0));
        luxeShop.addToStock(new Article(banana, (float) 10000.0));

        List<Shop> shops = new ArrayList<>();
        shops.add(aldi);
        shops.add(colruyt);
        shops.add(luxeShop);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));
        groceryList.addProductOrIncreaseAmount(new ProductAmount(banana, 5));

        // expected tour
        List<Shop> expectedTour = new ArrayList<>();
        expectedTour.add(aldi);
        expectedTour.add(colruyt);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        assertTourIsCorrect(expectedTour, tourFinder);
    }

    /**
     * We test a simple case: we want apples and bananas. Colruyt has only bananas
     * and Aldi has only apples. We should go to both to get all products, even if
     * it's very expensive and they're very far apart. There's also a PineappleShop
     * with selling Pineapples which we don't visit.
     *
     * Shops: - Aldi: Apple 50000 - Colruyt: Banana 10000 - PineappleShop: PineApple
     * 1
     *
     * Grocery List: Apple 5, Banana 5 FindShopsResult: Aldi, Colruyt
     */
    @Test
    public void getTour_OneProductPerShop_GoToAllRequiredShops() throws NoShopHasThatProductException {
        // Create shops
        Shop aldi = new Shop(0, "Aldi", schedule, beijing);
        aldi.addToStock(new Article(apple, (float) 50000.0));

        Shop colruyt = new Shop(0, "Colruyt", schedule, losAngeles);
        colruyt.addToStock(new Article(banana, (float) 100000.0));

        Shop pineappleShop = new Shop(0, "PineappleShop", schedule, montDesArts);
        pineappleShop.addToStock(new Article(pineapple, (float) 1.0));

        List<Shop> shops = new ArrayList<>();
        shops.add(aldi);
        shops.add(colruyt);
        shops.add(pineappleShop);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));
        groceryList.addProductOrIncreaseAmount(new ProductAmount(banana, 5));

        // expected tour
        List<Shop> expectedTour = new ArrayList<>();
        expectedTour.add(aldi);
        expectedTour.add(colruyt);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        assertTourIsCorrect(expectedTour, tourFinder);
    }

    /**
     * We test a simple case: two shops sell apples and we want apples. One is a bit
     * cheaper but really far. We should go to the closest shop.
     *
     * Shops: - Aldi (Mont des Arts): Apple 100 - Colruyt (Beijing): Banana 90
     *
     * Grocery List: Apple 5 FindShopsResult: Aldi
     */
    @Test
    public void getTour_shopCheaperButTooFar_GetClosestShop() throws NoShopHasThatProductException {
        // Create shops
        Shop aldi = new Shop(0, "Aldi", schedule, montDesArts);
        aldi.addToStock(new Article(apple, (float) 100.0));

        Shop colruyt = new Shop(0, "Colruyt", schedule, beijing);
        colruyt.addToStock(new Article(apple, (float) 90.0));

        List<Shop> shops = new ArrayList<>();
        shops.add(aldi);
        shops.add(colruyt);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));

        // expected tour
        List<Shop> expectedTour = new ArrayList<>();
        expectedTour.add(aldi);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        assertTourIsCorrect(expectedTour, tourFinder);
    }

    /**
     *
     * @formatter:off
     * We test a complicated situation to make sure it works. There are four kinds of products:
     * p1, p2, p3 and p4.
     * Here is the layout of the map (the products are between parentheses):
     *
     *        A (p1: 10) x x x x x x x
     *           x                    x x x x x x
     *           x                                x
     *           x            B (p4:50, p2:50)     x
     *           x                                  C (p3: 5, p2: 20)
     *      F (p4: 5) x                             x
     *                  x                          x
     *                    x E(p2: 5)              x
     *                              x            x
     *                               x          x                                    D (p4: 3)
     *                                x        x
     *                                 x     x
     *                                  x   x
     *                              StartPoint
     *
     * Grocery List: p1, p2, p3, p4
     * FindShopsResult: E, F, A and C.
     * Tour: startPoint -> E -> F -> A -> C -> startPoint
     * @formatter:on
     */
    @Test
    public void getTour_ComplicatedMap_FindsBestShops() throws NoShopHasThatProductException {
        // Create products
        Product p1 = new Product("p1", Product.Unit.UNIT, 0, 0, 0);
        Product p2 = new Product("p2", Product.Unit.UNIT, 0, 0, 0);
        Product p3 = new Product("p3", Product.Unit.UNIT, 0, 0, 0);
        Product p4 = new Product("p4", Product.Unit.UNIT, 0, 0, 0);

        // Create positions for shops and start
        // @formatter:off
        Position startPosition = new Position(0, 0);
        Position posA = new Position(-10 * KM_IN_DEGREES_LAT, 10  * KM_IN_DEGREES_LONG);
        Position posB = new Position(-4  * KM_IN_DEGREES_LAT, 7   * KM_IN_DEGREES_LONG);
        Position posC = new Position(5   * KM_IN_DEGREES_LAT, 6   * KM_IN_DEGREES_LONG);
        Position posD = new Position(15 *  KM_IN_DEGREES_LAT, 400 * KM_IN_DEGREES_LONG);
        Position posE = new Position(-5  * KM_IN_DEGREES_LAT, 4   * KM_IN_DEGREES_LONG);
        Position posF = new Position(-10 * KM_IN_DEGREES_LAT, 5   * KM_IN_DEGREES_LONG);
        // @formatter:on

        // Create shops
        Shop shopA = new Shop(0, "shopA", schedule, posA);
        shopA.addToStock(new Article(p1, (float) 10.0));

        Shop shopB = new Shop(0, "shopB", schedule, posB);
        shopB.addToStock(new Article(p4, (float) 50.0));
        shopB.addToStock(new Article(p2, (float) 50.0));

        Shop shopC = new Shop(0, "shopC", schedule, posC);
        shopC.addToStock(new Article(p2, (float) 20.0));
        shopC.addToStock(new Article(p3, (float) 5.0));

        Shop shopD = new Shop(0, "shopD", schedule, posD);
        shopD.addToStock(new Article(p4, (float) 3.0));

        Shop shopE = new Shop(0, "shopE", schedule, posE);
        shopE.addToStock(new Article(p2, (float) 10.0));

        Shop shopF = new Shop(0, "shopF", schedule, posF);
        shopF.addToStock(new Article(p4, (float) 5.0));

        Shop[] shopsArray = {shopA, shopB, shopC, shopD, shopE, shopF};
        List<Shop> shops = toList(shopsArray);

        // Create grocery list
        ProductAmount[] productsWanted = {new ProductAmount(p1, 5), new ProductAmount(p2, 5), new ProductAmount(p3, 5),
                new ProductAmount(p4, 5)};

        GroceryList groceryList = new GroceryList("my simple list", toList(productsWanted));

        // Expected tour
        Shop[] expectedShops = {shopE, shopF, shopA, shopC};
        List<Shop> expectedTour = toList(expectedShops);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(startPosition, startPosition, shops, groceryList);
        assertTourIsCorrect(expectedTour, tourFinder);
    }

    /**
     * We'll create a tour with many shops, each of which has a unique product we
     * want. There will be some useless shops (no product) and some useless products
     * (we don't want them). The resulting tour should have all shops with the
     * products that we want.
     */
    @Test
    public void getTour_manyShopsWithUniqueProduct_GetAllShop() throws NoShopHasThatProductException {
        // Create products and product amounts
        List<Product> products = new ArrayList<>();
        List<ProductAmount> productAmounts = new ArrayList<>();
        Product currentProduct;
        for (int i = 0; i < 100; i++) {
            currentProduct = new Product("Product " + Integer.toString(i + 1), Product.Unit.KG, 0, 0, 0);
            products.add(currentProduct);
            productAmounts.add(new ProductAmount(currentProduct, 5));
        }

        // Add some useless products (not wanted)
        for (int i = 100; i < 200; i++) {
            currentProduct = new Product("Product " + Integer.toString(i + 1), Product.Unit.KG, 0, 0, 0);
            products.add(currentProduct);
        }

        // Create shops
        List<Shop> shops = new ArrayList<>();
        List<Shop> usedShops = new ArrayList<>();
        Shop currentShop;
        for (int i = 0; i < 100; i++) {
            currentShop = new Shop(0, "Shop" + Integer.toString(i + 1), schedule, montDesArts);
            currentShop.addToStock(new Article(products.get(i), (float) 10.0));
            shops.add(currentShop);
            usedShops.add(currentShop);
        }

        // Add some useless shops (no products)
        for (int i = 100; i < 200; i++) {
            currentShop = new Shop(0, "Shop" + Integer.toString(i + 1), schedule, montDesArts);
            shops.add(currentShop);
        }

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list", productAmounts);

        // expected tour
        List<Shop> expectedTour = new ArrayList<>(usedShops);

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
        List<Shop> bestTour = tourFinder.getBestTourShops();

        // We don't test the order, it doesn't matter here
        assertToursHaveSameShops(bestTour, expectedTour);
    }


    /**
     * When all shops are empty, an exception should be raised.
     */
    @Test(expected = NoShopHasThatProductException.class)
    public void getTour_shopsAreEmpty_ThrowsException() throws NoShopHasThatProductException {
        // Create empty shop
        Shop shop = new Shop(0, "Shop", schedule, montDesArts);
        List<Shop> shops = toList(shop);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
    }

    /**
     * When all shops are empty, an exception should be raised.
     */
    @Test(expected = NoShopHasThatProductException.class)
    public void getTour_shopsDontHaveAllProducts_ThrowsException() throws NoShopHasThatProductException {
        // Create empty shop
        Shop shop = new Shop(0, "Shop", schedule, montDesArts);
        shop.addToStock(new Article(apple, (float) 5.0));
        List<Shop> shops = toList(shop);

        // Create grocery list
        GroceryList groceryList = new GroceryList("my simple list");
        groceryList.addProductOrIncreaseAmount(new ProductAmount(apple, 5));
        groceryList.addProductOrIncreaseAmount(new ProductAmount(banana, 5));

        // Test result
        BestTourFinder tourFinder = new BestTourFinder(home, home, shops, groceryList);
    }


    private void assertToursHaveSameShops(List<Shop> expectedTour, List<Shop> actualTour) {
        assertThat(new HashSet<>(expectedTour), is(new HashSet<>(actualTour)));
    }

    private void assertTourIsCorrect(List<Shop> expectedTour, BestTourFinder finder) {
        assertBestTourShopsAreCorrect(expectedTour, finder);
        assertBestTourPositionsAreCorrect(expectedTour, finder);
    }

    private void assertBestTourPositionsAreCorrect(List<Shop> expectedTour, BestTourFinder finder) {
        List<Position> positions = finder.getBestTourPositions();
        List<Shop> bestTourShops = finder.getBestTourShops();
        List<Shop> reversedTour = reverseTour(bestTourShops);

        assertEquals(positions.get(0), finder.getStartPos());
        assertEquals(positions.get(positions.size() - 1), finder.getEndPos());

        int size = bestTourShops.size();
        if (finder.getStartPos().equals(finder.getEndPos())) {
            for (int i = 0; i < size; i++) {
                assertTrue(
                        positions.get(i + 1).equals(bestTourShops.get(i).getPosition()) ||
                                positions.get(i + 1).equals(reversedTour.get(i).getPosition())
                );
            }
        } else {
            for (int i = 0; i < size; i++) {
                assertEquals(positions.get(i + 1), bestTourShops.get(i).getPosition());
            }
        }
    }

    private void assertBestTourShopsAreCorrect(List<Shop> expectedTour, BestTourFinder finder) {
        List<Shop> bestTourShops = finder.getBestTourShops();
        if (finder.getStartPos().equals(finder.getEndPos())) {
            // If the order doesn't matter, the shop tour can go in either direction (a-b-c or c-b-a)
            List<Shop> reversedTour = reverseTour(bestTourShops);
            assertThat(
                    expectedTour,
                    CoreMatchers
                            .either(CoreMatchers.is(bestTourShops))
                            .or(CoreMatchers.is(reversedTour))

            );
        } else {
            assertEquals(expectedTour, bestTourShops);
        }
    }

    private List<Shop> reverseTour(List<Shop> actualTour) {
        List<Shop> reversedTour = new ArrayList<>(actualTour);
        Collections.reverse(reversedTour);
        return reversedTour;
    }

    private <T> List<T> toList(T[] array) {
        return new ArrayList<>(Arrays.asList(array));
    }

    private <T> List<T> toList(T element) {
        ArrayList<T> list = new ArrayList<>();
        list.add(element);
        return list;
    }
}
