package tests.models;

import be.ac.ulb.infof307.g02.models.*;
import be.ac.ulb.infof307.g02.utils.Utils;
import org.junit.Test;
import tests.BaseTest;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.*;

public class ShopTest extends BaseTest {

    // Constructors

    @Test
    public void constructor_validValues_correctCreation() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        new Shop(1, "Carrefour", schedule, new Position(1, 1));
    }

    @Test(expected = IllegalArgumentException.class)
    public void constructor_negativeId_ThrowsException() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        new Shop(-3, "Carrefour", schedule, new Position(1, 1));
    }

    @Test(expected = IllegalArgumentException.class)
    public void constructor_emptyName_ThrowsException() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        new Shop(1, "", schedule, new Position(1, 1));
    }

    @Test(expected = IllegalArgumentException.class)
    public void constructor_emptySchedule_ThrowsException() {
        new Shop(1, "", new Schedule(), new Position(1, 1));
    }

    // Setters

    @Test
    public void setId_validValues_idChanged() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1));
        shop.setId(2);
        int id = shop.getId();
        assertEquals(id, 2);
    }

    @Test
    public void setName_validValues_nameChanged() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1));
        shop.setName("Delhaize");
        String name = shop.getName();
        assertEquals(name, "Delhaize");
    }

    @Test
    public void setSchedule_validValues_scheduleChanged() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1));
        Schedule newSchedule = new Schedule();
        newSchedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(12, 0)), Day.MONDAY);
        shop.setSchedule(newSchedule);
        assertEquals(shop.getSchedule(), newSchedule);
    }

    @Test(expected = IllegalArgumentException.class)
    public void setId_negativeValue_ThrowsException() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Delhaize", schedule, new Position(1, 1));
        shop.setId(-2);
    }

    @Test(expected = IllegalArgumentException.class)
    public void setName_emptyName_ThrowsException() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Delhaize", schedule, new Position(1, 1));
        shop.setName("");
    }

    @Test(expected = IllegalArgumentException.class)
    public void setSchedule_emptySchedule_ThrowsException() {
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Delhaize", schedule, new Position(1, 1));
        shop.setSchedule(new Schedule());
    }

    @Test
    public void containsProducts_validValues_returnTrue() {
        Product prod1 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);
        Product prod2 = new Product("water", Product.Unit.LITER, 0, 0, 0);
        Article art1 = new Article(prod1, 2);
        Article art2 = new Article(prod2, 1);
        Map<Product, Article> stock = new HashMap<>();
        stock.put(art1.getProduct(), art1);
        stock.put(art2.getProduct(), art2);
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1), stock);
        ProductAmount item1 = new ProductAmount(prod1, 2);
        ProductAmount item2 = new ProductAmount(prod2, 1);
        List<ProductAmount> itemList = Arrays.asList(item1, item2);
        GroceryList list = new GroceryList("mylist", itemList);

        boolean res = shop.sellAllProducts(list.getProductsList());
        assertTrue(res);

    }

    @Test
    public void containsProducts_invalidValues_returnFalse() {
        Product prod1 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);
        Product prod2 = new Product("water", Product.Unit.LITER, 0, 0, 0);
        Product prod3 = new Product("boeuf", Product.Unit.KG, 15, 26, 0);
        Article art1 = new Article(prod1, 2);
        Article art2 = new Article(prod2, 1);
        Map<Product, Article> stock = new HashMap<>();
        stock.put(art1.getProduct(), art1);
        stock.put(art2.getProduct(), art2);
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1), stock);
        ProductAmount item1 = new ProductAmount(prod1, 2);
        ProductAmount item2 = new ProductAmount(prod3, 1);
        List<ProductAmount> itemList = Arrays.asList(item1, item2);
        GroceryList list = new GroceryList("mylist", itemList);

        boolean res = shop.sellAllProducts(list.getProductsList());
        assertFalse(res);

    }

    @Test
    public void getPriceOfList_validValues_returnCorrectPrice() {
        Product prod1 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);
        Product prod2 = new Product("water", Product.Unit.LITER, 0, 0, 0);
        Product prod3 = new Product("banana", Product.Unit.KG, 0.3, 1.1, 23);// so we are sure that its not the same object in memory
        Product prod4 = new Product("water", Product.Unit.LITER, 0, 0, 0);
        Article art1 = new Article(prod1, 6);
        Article art2 = new Article(prod2, 1);
        Map<Product, Article> stock = new HashMap<>();
        stock.put(art1.getProduct(), art1);
        stock.put(art2.getProduct(), art2);
        Schedule schedule = new Schedule();
        schedule.addTimeSlot(new TimeSlot(new Time(8, 0), new Time(10, 0)), Day.MONDAY);
        Shop shop = new Shop(1, "Carrefour", schedule, new Position(1, 1), stock);
        ProductAmount item1 = new ProductAmount(prod3, 2);
        ProductAmount item2 = new ProductAmount(prod4, 1);
        List<ProductAmount> itemList = Arrays.asList(item1, item2);
        GroceryList list = new GroceryList("mylist", itemList);

        double result = shop.getPriceOfList(list);

        assertTrue(Utils.equalsDouble(result, 13.0));

    }

}
