package tests.models;

import be.ac.ulb.infof307.g02.exceptions.model.InvalidPasswordException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.User;
import be.ac.ulb.infof307.g02.utils.HashPassword;
import org.junit.Test;
import tests.BaseTest;

import static org.junit.Assert.*;

public class UserTest extends BaseTest {

    // Constructor

    @Test
    public void constructor_ValidValues_CorrectCreation() throws UserCreationException {
        User user = new User("Zazu", "aaaaaa1994");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_EmptyUsername_ThrowsException() throws UserCreationException {
        User user = new User("", "qwerty");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_UsernameOnlyWhitespace_ThrowsException() throws UserCreationException {
        User user = new User("  \t  \n ", "1998");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_EmptyPassword_ThrowsException() throws UserCreationException {
        User user = new User("Nala", "");
    }

    // Getters

    @Test
    public void getUserName_NameMushu_ReturnsMushu() throws UserCreationException {
        User user = new User("Mushu", "aaaaaa1998");
        assertEquals("Mushu", user.getUserName());
    }

    @Test
    public void getHashedPassword_PassPumba_ReturnsHashedPumba() throws UserCreationException {
        User user = new User("Timon", "aaaaaaPumba");
        assertEquals(HashPassword.hash("aaaaaaPumba"), user.getHashedPassword());
    }

    // Setters

    @Test
    public void setUserName_NewNameScrat_NameHasChanged() throws UserCreationException {
        User user = new User("Mushu", "aaaaaa1998");
        assertEquals("Mushu", user.getUserName());
        user.setUserName("Scrat");
        assertEquals("Scrat", user.getUserName());
    }

    @Test(expected = IllegalArgumentException.class)
    public void setUserName_NewNameOnlyWhitespace_ThrowsException() throws UserCreationException {
        User user = new User("Bob", "aaaaaa1998");
        user.setUserName("  \t  \n  ");
    }

    @Test(expected = IllegalArgumentException.class)
    public void setUserName_NewNameEmpty_ThrowsException() throws UserCreationException {
        User user = new User("Esmeralda", "aaaaaa1998");
        user.setUserName("");
    }

    @Test
    public void isPasswordValid_PasswordTooShort_ReturnsFalse() {
        assertFalse(User.isPasswordValid("a")); // 1 char
        assertFalse(User.isPasswordValid("aaa")); // 3 char
        assertFalse(User.isPasswordValid("aaaaa")); // 5 char
        assertFalse(User.isPasswordValid("aaaaaa")); // 6 char
        assertFalse(User.isPasswordValid("aaaaaa")); // 7 char
        assertFalse(User.isPasswordValid("aaBB22*")); // 7 char, all char groups
    }

    @Test
    public void isPasswordValid_NotEnoughCharacterGroups_ReturnsFalse() {
        assertFalse(User.isPasswordValid("abcdefghi")); // only lowercase
        assertFalse(User.isPasswordValid("ABCDEFGHI")); // only uppercase
        assertFalse(User.isPasswordValid("123456789")); // only numbers
        assertFalse(User.isPasswordValid("@@**//++--")); // only special chars
    }

    @Test
    public void isPasswordValid_LongAndEnoughCharacterGroups_ReturnsTrue() {
        assertTrue(User.isPasswordValid("abcdefghI")); // lowercase + upper
        assertTrue(User.isPasswordValid("abcdefgh0")); // lowercase + number
        // Test a sample of the allowed special chars
        assertTrue(User.isPasswordValid("abcdefgh+")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh-")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh*")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh/")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh@")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh#")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh$")); // lowercase + special
        assertTrue(User.isPasswordValid("abcdefgh%")); // lowercase + special

        assertTrue(User.isPasswordValid("ABCDEFGH0")); // uppercase + number
        assertTrue(User.isPasswordValid("ABCDEFGH+")); // uppercase + special

        assertTrue(User.isPasswordValid("12345678*")); // number + special
    }

    @Test(expected = UserCreationException.class)
    public void constructor_OnlyLowercaseLettersInPassword_ThrowsException() throws UserCreationException {
        User user = new User("Pocahontas", "aaaaaaaaaa");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_OnlyUppercaseLettersInPasword_ThrowsException() throws UserCreationException {
        User user = new User("Pocahontas", "AAAAAAAAAAAA");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_OnlyNumbersInPassword_ThrowsException() throws UserCreationException {
        User user = new User("Pocahontas", "00000000001111");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_OnlySpecialCharsInPassword_ThrowsException() throws UserCreationException {
        User user = new User("Pocahontas", "@@####++--**//");
    }

    @Test(expected = UserCreationException.class)
    public void constructor_PasswordTooShort_ThrowsException() throws UserCreationException {
        User user = new User("Pocahontas", "aA+4567");
    }

    @Test
    public void constructor_PasswordLongAndWithTwoCharTypes_Success() throws UserCreationException {
        User user = new User("Pocahontas", "aaaaaaaaA");
        user = new User("Pocahontas", "aaaaaaaa0");
        user = new User("Pocahontas", "aaaaaaaa+");

        user = new User("Pocahontas", "AAAAAAAAA0");
        user = new User("Pocahontas", "AAAAAAAAA+");

        user = new User("Pocahontas", "0123456789++");
    }

    @Test
    public void setPassword_NewPasswordaaaaaaRapumzel_HashedPasswordHasChanged()
            throws UserCreationException, InvalidPasswordException {
        User user = new User("Jasmine", "aaaaaa1998");
        assertEquals(HashPassword.hash("aaaaaa1998"), user.getHashedPassword());
        user.setPassword("aaaaaaRapumzel");
        assertEquals(HashPassword.hash("aaaaaaRapumzel"), user.getHashedPassword());
    }

    @Test
    public void setPassword_NewHashedPasswordHashedaaaaaaRapumzel_HashedPasswordHasChanged()
            throws UserCreationException {
        User user = new User("Jasmine", "aaaaaa1998");
        user.setHashedPassword(HashPassword.hash("aaaaaaRapumzel"));
        assertEquals(HashPassword.hash("aaaaaaRapumzel"), user.getHashedPassword());
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_NewPasswordEmpty_ThrowsException() throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("");
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_NewPasswordTooShort_ThrowsException()
            throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("aaaaAAA");
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_OnlyNumbers_ThrowsException() throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("123456789");
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_OnlyLowerCaseLetters_ThrowsException()
            throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("abcdefghijk");
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_OnlyUpperCaseLetters_ThrowsException()
            throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("ABCDEFGHIJKL");
    }

    @Test(expected = InvalidPasswordException.class)
    public void setPassword_OnlySpecialChars_ThrowsException() throws UserCreationException, InvalidPasswordException {
        User user = new User("Pocahontas", "aaaaaa1995");
        user.setPassword("@+-*/...!!");
    }

}
