package tests.utils;

import be.ac.ulb.infof307.g02.database.DatabaseConnection;
import be.ac.ulb.infof307.g02.database.DatabaseConnectionSingleton;
import be.ac.ulb.infof307.g02.database.KeyDB;
import be.ac.ulb.infof307.g02.database.UserDB;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseConnectionException;
import be.ac.ulb.infof307.g02.exceptions.database.DatabaseException;
import be.ac.ulb.infof307.g02.exceptions.database.UserNameTakenException;
import be.ac.ulb.infof307.g02.exceptions.model.UserCreationException;
import be.ac.ulb.infof307.g02.models.User;
import be.ac.ulb.infof307.g02.utils.AES;
import be.ac.ulb.infof307.g02.utils.Crypto;
import be.ac.ulb.infof307.g02.utils.CryptoSingleton;
import be.ac.ulb.infof307.g02.utils.RSA;
import org.hamcrest.core.IsEqual;
import org.hamcrest.core.IsNot;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import tests.BaseTest;

import javax.crypto.SecretKey;
import java.security.KeyPair;
import java.sql.SQLException;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;

public class CryptoTest extends BaseTest {

    private static DatabaseConnection testDB;
    private static KeyDB keyDB;
    private static UserDB userDB;

    @BeforeClass
    public static void setUpBeforeClass() throws Exception {
        testDB = DatabaseConnectionSingleton.getDatabaseConnection();
        keyDB = new KeyDB(testDB);
        userDB = new UserDB(testDB);
    }

    @AfterClass
    public static void tearDownAfterClass() throws Exception {
        keyDB.emptyAllTables();
    }

    @Before
    public void setup() throws SQLException {
        userDB.emptyAllTables();
    }

    @Test
    public void rsa_EncryptAndDecrypt_Success() {
        KeyPair keyPair = RSA.generateKeyPair();

        String message = "Hello! Is it me you're looking for?";

        byte[] encryptedMessage = RSA.encrypt(message.getBytes(), keyPair.getPublic());
        byte[] decryptedMessage = RSA.decrypt(encryptedMessage, keyPair.getPrivate());

        assertEquals(message, new String(decryptedMessage));
    }

    @Test
    public void aes_EncryptAndDecrypt_Success() {
        SecretKey secretKey = AES.generateAESKey();

        String message = "I can see it in your eyes. I can see it in your smile.";

        byte[] encryptedMessage = AES.encrypt(message.getBytes(), secretKey);
        byte[] decryptedMessage = AES.decrypt(encryptedMessage, secretKey);

        assertEquals(message, new String(decryptedMessage));
    }

    @Test
    public void generateKeys_UserChangesPassword_RSAKeysChangeButAESKeyDoesnt() throws UserCreationException, UserNameTakenException, DatabaseException, DatabaseConnectionException {
        User user = new User("lionel", "hello123");
        userDB.addUser(user);
        CryptoSingleton.setupCryptoForUser(user, testDB);

        Crypto crypto = CryptoSingleton.getCrypto();
        KeyPair keyPair = crypto.getKeyPair();
        SecretKey aesKey = crypto.getAesKey();

        User updatedUser = new User("lionel", "goodbye123");
        userDB.updatePassword(updatedUser.getUserName(), updatedUser.getHashedPassword());
        CryptoSingleton.updateUserKeys(user, updatedUser, testDB);
        crypto = CryptoSingleton.getCrypto();

        KeyPair newKeyPair = crypto.getKeyPair();
        SecretKey newAesKey = crypto.getAesKey();

        assertArrayEquals(aesKey.getEncoded(), newAesKey.getEncoded());
        assertArraysAreDifferent(keyPair.getPublic().getEncoded(), newKeyPair.getPublic().getEncoded());
        assertArraysAreDifferent(keyPair.getPrivate().getEncoded(), newKeyPair.getPrivate().getEncoded());
    }

    @Test
    public void generateKeys_UserChangesUsername_RSAKeysChangeButAESKeyDoesnt() throws UserCreationException, UserNameTakenException, DatabaseException, DatabaseConnectionException {
        User user = new User("lionel", "hello123");
        userDB.addUser(user);
        CryptoSingleton.setupCryptoForUser(user, testDB);

        Crypto crypto = CryptoSingleton.getCrypto();
        KeyPair keyPair = crypto.getKeyPair();
        SecretKey aesKey = crypto.getAesKey();

        User updatedUser = new User("bob", "hello123");
        userDB.updateUserName("lionel", "bob");
        CryptoSingleton.updateUserKeys(user, updatedUser, testDB);
        crypto = CryptoSingleton.getCrypto();

        KeyPair newKeyPair = crypto.getKeyPair();
        SecretKey newAesKey = crypto.getAesKey();

        assertArrayEquals(aesKey.getEncoded(), newAesKey.getEncoded());
        assertArraysAreDifferent(keyPair.getPublic().getEncoded(), newKeyPair.getPublic().getEncoded());
        assertArraysAreDifferent(keyPair.getPrivate().getEncoded(), newKeyPair.getPrivate().getEncoded());
    }

    public void assertArraysAreDifferent(byte[] a, byte[] b) {
        assertThat(a, IsNot.not(IsEqual.equalTo(b)));
    }
}
